//  $Id: prom_grid.hh,v 1.4 2011-01-22 20:13:38 madams Exp $
//  Author:             Mark F. Adams
// Copyright (c) 2000 by Mark F. Adams 
//  Filename:           prom_grid.hh
//
#ifndef  __PROM_GRID_H__
#define  __PROM_GRID_H__

#include "prom_elem.hh"

//! class PromBank: memory managment for common objects
/*!
  class PromBank: memory managment for common objects
*/
class PromBank 
{
  friend class PromAggNode;
  friend class PromGeoNode;
public:
  //! constructor
  PromBank( int bsz = 1024 );
  //! destructor
  virtual ~PromBank();
  // 
  PromAggNode  *newAggNode();
  PromGeoNode  *newGeoNode();
  PromDTet     *newTet(PromNode *a, PromNode *b, PromNode *c, PromNode *d);
  PromFacet    *newFacet(PromNode* a,PromNode* b,PromNode*c,PromNode* d=NULL);
private:
  PromDTet     *newSymbData();
  PromFacet    *newGeoData();
  shpfloat     *newP( int sz );
public:
  int           DeleteAllCaches();
  int           DeleteFacets();
  int           DeleteTets();
  int           DeleteSymb();
  int           DeleteP();
  int           Kill( PromDTet* );
  int           Kill( PromFacet* );
  int           Kill( PromNode* );
  // special Tet helpers
  int           getNumTets();
private:
  int           Deallocate( PromDTet* const e );
  int           Deallocate( PromAggNode* const e );
  int           Deallocate( PromGeoNode* const e );
  int           Deallocate( PromFacet * const facet );
  int           Deallocate( shpfloat* const, int sz );
protected:
  // allocate data
  int           GetFreeTet( PromDTet** );
  int           GetFreeFacet( PromFacet** );
  int           GetFreeAggNode( PromAggNode** );
  int           GetFreeGeoNode( PromGeoNode** );
  int           GetFreeP( shpfloat**, int sz );
  // data  
  PromList<PromDTet*>   *tetBlocks_;
  PromDTet     *currTet_;
  PromDTet     *freeTet_;

  PromList<PromFacet*>   *facetBlocks_;
  PromFacet    *currFacet_;
  PromFacet    *freeFacet_;
  
  PromList<PromAggNode*> *aggNodeBlocks_;
  PromAggNode  *currAggNode_;
  PromAggNode  *freeAggNode_;

  PromList<PromGeoNode*> *geoNodeBlocks_;
  PromGeoNode  *currGeoNode_;
  PromGeoNode  *freeGeoNode_;

  PromList<PromSymbGeoData*>   *symbGeoDataBlocks_;
  PromList<PromSymbData*>   *symbDataBlocks_;

  PromList<PromNode**>   *symbNodeArrBlocks_;

  PromList<shpfloat*> *PBlocks_[10];
  shpfloat     *currP_[10];
  PromList<shpfloat*> freeP_[10];
  int           Psizes_[10];
  int           nPSizes_;

  const int     blocksize_;
  int           numTetFree_;
  int           numActiveTets_;
  int           numFacetFree_;
  int           numNodesFree_; // covers all nodes and symb data
  int           numPFree_[10];
  bool          clearSymbCalled;
};

/***********************************************************************
 *                              PromCRVector
 **********************************************************************/
class PromCRVector : public PromVector
{ 
public:
  //!
  PromCRVector( const PromMap &map, const PromMap *pmap = NULL ) : 
    PromVector( map ), p_(NULL){
    if( pmap != NULL ) p_ = new PromVector( *pmap );
  }
  PromCRVector( const PromCRVector *xx ) : 
    PromVector( xx->map_ ), p_(NULL){
    if( xx->p_ != NULL ) p_ = new PromVector( xx->p_->map_ );
  }
  //!
  virtual ~PromCRVector(){
    if( p_ != NULL /*&& malled_*/ ) delete p_;
  }
  //#define PROM_TRACE_NONCR
  //#define PROM_TRACE_CR
  //!
  int Set( double alpha ){
    assert(this!=NULL);
    int ierr = PromVector::Set( alpha );                    CHKERRQ(ierr);
    if( p_ != NULL ){
      p_->Set( alpha ); 
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);
#endif
    }
    else {
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif
    }
    return 0;
  }
  int Scale( double alpha ){
    assert(this!=NULL);
    int ierr = PromVector::Scale( alpha );                    CHKERRQ(ierr);
    if( p_ != NULL ){
      p_->Scale( alpha ); 
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method alpha=%e\n",__LINE__,alpha);
#endif
    }
    else {
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif
    }
    return 0;
  }
  //!
  int AYPX( double alpha, const PromVector_base * const xxB ){
    assert(this!=NULL); int ierr;
    // Computes y = x + alpha y. 
    ierr = PromVector::AYPX( alpha, xxB );             CHKERRQ(ierr);
    const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xxB);
    if( xx == NULL || p_ == NULL ) {
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif
    }
    else {
      assert( xx->p_ != NULL );
      ierr = p_->AYPX( alpha, xx->p_ );                    CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);
#endif
    }
    return 0;
  }

  //!
  int Dot(const PromVector_base * const xxB, double * const dot)const{
    assert(this!=NULL); assert(xxB!=NULL);int ierr;
    double temp;
    ierr = PromVector::Dot( xxB, dot );                 CHKERRQ(ierr);
    const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xxB);
    if( xx == NULL  || p_ == NULL ) {
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif      
    }
    else {
      assert( xx->p_ != NULL );
      ierr = p_->Dot( xx->p_, &temp );                     CHKERRQ(ierr);
      *dot += temp;
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif    
    }
    return 0;  
  }
  //!
  int Norm2( double * const norm ) const{
    assert(this!=NULL);
    double temp, temp2;
    int ierr = PromVector::Norm2( &temp );                 CHKERRQ(ierr);
    if( p_ != NULL ){ 
      ierr = p_->Norm2( &temp2 );                           CHKERRQ(ierr);
      *norm = sqrt( temp2 * temp2 + temp * temp );    
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif      
    }
    else {
      *norm = temp;
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif  
    }
    return 0;  
  }
  //!
  int NormMax( double * const norm ) const{
    assert(this!=NULL);
    double temp, temp2;
    int ierr = PromVector::NormMax( &temp );                 CHKERRQ(ierr);
    if( p_ != NULL ){ 
      ierr = p_->NormMax( &temp2 );                           CHKERRQ(ierr);
      if( temp2 > temp ) *norm = temp2;
      else *norm = temp;
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif      
    }
    else {
      *norm = temp;
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif  
    }
    return 0;  
  }
  //!
  int Copy( const PromVector_base * const xxB ){
    assert(this!=NULL); int ierr;
    PromVector::Copy( xxB );
    const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xxB);
    if( xx == NULL || p_ == NULL ) {
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif   
    }
    else {
      if( p_->getn() != xx->p_->getn() ) {
	SETERRQ2(PETSC_COMM_SELF,1,"inc. lengths %d %d",p_->getn(),xx->p_->getn());
      }
      ierr = p_->Copy( xx->p_ ); CHKERRQ(ierr); 
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif
    }
    return 0;
  }
  //!
  int AXPY( double alpha, const PromVector_base * const xxB ) {
    assert(this!=NULL); assert(xxB!=NULL); int ierr;
    // Computes y = alpha x + y. 
    ierr = PromVector::AXPY( alpha, xxB );              CHKERRQ(ierr);
    const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xxB);
    if( xx == NULL || p_ == NULL  ) {
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif
    }
    else {
      assert( xx->p_ != NULL );
      ierr = p_->AXPY( alpha, xx->p_ );                    CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif
    }
    return 0;
  }
  //!
  int AXPBY( double a, double b, const PromVector_base *const xxB ){
    assert(this!=NULL); int ierr;
    ierr = PromVector::AXPBY( a, b, xxB );                    CHKERRQ(ierr);
    const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xxB);
    if( xx == NULL  || p_ == NULL ) {
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif   
    }
    else {
      assert( xx->p_ != NULL );
      ierr = p_->AXPBY( a, b, xx->p_ );      CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif
    }
    return 0;
  }
  //! this(w) = alpha x + y
  int WAXPY( double alpha, const PromVector_base * const xxB, 
	     const PromVector_base * const yyB ){
    assert(this!=NULL); int ierr;
    ierr = PromVector::WAXPY( alpha, xxB, yyB );            CHKERRQ(ierr);
    const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xxB);
    const PromCRVector *const yy = dynamic_cast<const PromCRVector*>(yyB);
    if( xx == NULL || yy == NULL ||  p_ == NULL ) {
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif
    }
    else {
      assert(xx->p_ != NULL && yy->p_ != NULL);
      ierr = p_->WAXPY(alpha, xx->p_, yy->p_); CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif 
    }
    return 0;
  }
  //!
  int MAXPY( const int nv, double alpha[], const PromVector_base *const*Xarr){
    assert(this!=NULL); int i,ierr; assert(nv<1000); assert(Xarr!=NULL);
    //! this = this + alpha[j] X[j]
    ierr = PromVector::MAXPY( nv, alpha, Xarr ); CHKERRQ(ierr);
    if( p_ != NULL ) {
      const PromVector_base *arg[1000];
      for( i = 0 ; i < nv ; i++ ) {
	const PromCRVector *vv = dynamic_cast<const PromCRVector*>(Xarr[i]);
	assert(vv != NULL && vv->p_ != NULL); arg[i] = vv->p_;
      }
      p_->PromVector::MAXPY( nv, alpha, arg ); CHKERRQ(ierr); 
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif 
    }
    else {
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif
    }
    return 0;
  }
  //!
  int MDot( const int nv, const PromVector_base *const*Xarr, double alpha[])const{
    assert(this!=NULL); int ii,ierr; assert(nv<1000); assert(Xarr!=NULL);
    //! alpha[i] <- this * xarr[i]
    ierr = PromVector::MDot( nv, Xarr, alpha );  CHKERRQ(ierr);
    if( p_ != NULL ){ 
      const PromVector_base *arg[1000]; 
      for(ii=0;ii<nv;ii++) {
	const PromCRVector *vv = dynamic_cast<const PromCRVector*>(Xarr[ii]);
	assert(vv != NULL && vv->p_ != NULL); arg[ii] = vv->p_;
      }
      double alpha2[1000]; 
      PetscLogFlops( 2*nv*vec_->map->n ); 
      ierr = p_->MDot( nv, arg, alpha2 );  CHKERRQ(ierr);
      for(ii=0;ii<nv;ii++) alpha[ii] += alpha2[ii];
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif   
    }
    else {
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif  
    }
    return 0;
  }
  int PointwiseMult( const PromVector_base * const X,
		     const PromVector_base * const Y){
    SETERRQ(PETSC_COMM_SELF,-1,"PromCRVector::PointwiseMult");
#if defined(PARCH_win32)
    return 0;
#endif
  }
  int PointwiseDiv( const PromVector_base * const X,
		    const PromVector_base * const Y){
    SETERRQ(PETSC_COMM_SELF,-1,"PromCRVector::PointwiseDiv");
#if defined(PARCH_win32)
    return 0;
#endif
  }
  int Assembly(){
    assert(this!=NULL);
    int ierr = VecAssemblyBegin( vec_ ); CHKERRQ(ierr);
    ierr = VecAssemblyEnd( vec_ ); CHKERRQ(ierr);
    if( p_ != NULL ){ 
      ierr = p_->Assembly();                           CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif      
    }
    else {
#if defined(PROM_TRACE_NONCR)
      PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR meth\n",__LINE__);
#endif  
    }    
    return 0;
  }
  // data 
  PromVector *p_; 
};

/***********************************************************************
 *                           PromCRMatrix
 **********************************************************************/
class PromCRMatrix : public PromMatrix
{
public:
  //! constructor
  PromCRMatrix( const PromGrid *grid, int nA,int *pnA,int nB,int *pnB );
  //!
  PromCRMatrix( const PromMap &map, int nA,int *pnA,int nB,int *pnB,
		int lcols=-1, int gcols=-1) : 
    PromMatrix(map, nA, pnA, nB, pnB, lcols, gcols),D_(NULL), C_(NULL), Ct_(NULL), 
    CDinvCt_(NULL) {} 
  //!
  PromCRMatrix( const PromMap map, Mat Amat ) : 
    PromMatrix( map, Amat ),D_(NULL), C_(NULL), Ct_(NULL), CDinvCt_(NULL)
  {
    dirty_data_ = TRUE; // user matrix is dirty w/o me adding to it
  } 
  //! Copy constructor
  PromCRMatrix( const PromCRMatrix* AA ) : 
    PromMatrix( AA ), D_(NULL), C_(NULL), Ct_(NULL), CDinvCt_(NULL) {
    if( AA->C_ != NULL ) C_ = new PromMatrix( AA->C_ );
    if( AA->D_ != NULL ) D_ = new PromMatrix( AA->D_ );
    if( AA->Ct_ != NULL ) Ct_ = new PromMatrix( AA->Ct_ );
  }
  virtual ~PromCRMatrix(){
    if( C_ != NULL ) delete C_; if( Ct_ != NULL ) delete Ct_; 
    if( D_ != NULL ) delete D_;
    if( CDinvCt_ != NULL ) delete CDinvCt_;
  }
  //! MultTransposeAdd: v3 = v2 + A' * v1
  virtual int MultTransposeAdd( const PromVector_base * const xB,
				const PromVector_base * const yB, 
				PromVector_base * const wB ) const {
    assert(this!=NULL);
    int ierr;
    ierr = PromMatrix::MultTransposeAdd( xB, yB, wB );  CHKERRQ(ierr); 
    if( C_ != NULL ) {
      const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xB);
      const PromCRVector *const yy = dynamic_cast<const PromCRVector*>(yB);
      PromCRVector *const ww = dynamic_cast<PromCRVector*>(wB); 
      if( xx != NULL && yy != NULL && ww != NULL ) {
	// v3 = v2 + A' * v1 
	if( Ct_ != NULL ) {
	  ierr = Ct_->MultAdd( xx->p_, yy, ww );      CHKERRQ(ierr);
	}
	else {
	  ierr = C_->MultTransposeAdd( xx->p_, yy, ww );      CHKERRQ(ierr);
	}
	ierr = C_->MultAdd( xx, yy->p_, ww->p_ );           CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
	PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);
#endif	    
      }
      else {
#if defined(PROM_TRACE_NONCR)
	PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR met\n",__LINE__);
#endif
      }
    }
    if( D_ != NULL ) {
      const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xB);
      const PromCRVector *const yy = dynamic_cast<const PromCRVector*>(yB);
      PromCRVector *const ww = dynamic_cast<PromCRVector*>(wB); 
      if( xx != NULL && yy != NULL && ww != NULL ) {
	ierr = D_->MultTransposeAdd(xx->p_, yy->p_, ww->p_); CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
	PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);
#endif	    
      }
      else {
#if defined(PROM_TRACE_NONCR)
	PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR met\n",__LINE__);
#endif
      }
    }
    return 0;
  }
  //! MultTranspose:  v2 = A' * v1
  int MultTranspose( const PromVector_base * const xB,
		     PromVector_base * const yB ) const {
    assert(this!=NULL);
    int ierr;
    ierr = PromMatrix::MultTranspose( xB, yB );  CHKERRQ(ierr); 
    if( C_ != NULL ) {
      const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xB);
      PromCRVector *const yy = dynamic_cast<PromCRVector*>(yB);
      if( xx != NULL && yy != NULL ){   
	// v3 = v2 + A' * v1
	if( Ct_ != NULL ) {
	  ierr = Ct_->MultAdd( xx->p_, yy, yy ); CHKERRQ(ierr);
	}
	else {
	  ierr = C_->MultTransposeAdd( xx->p_, yy, yy ); CHKERRQ(ierr);
	}
	ierr = C_->Mult( xx, yy->p_ );                 CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
	PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);
#endif	
      }
      else {
#if defined(PROM_TRACE_NONCR)
	PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR met\n",__LINE__);
#endif
      }
    }
    else if( D_ != NULL ) {
      const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xB);
      PromCRVector *const yy = dynamic_cast<PromCRVector*>(yB);
      if( xx != NULL && yy != NULL ) {
	ierr = D_->MultTranspose( xx->p_, yy->p_ );       CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
	PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);
#endif	
      }
      else {
#if defined(PROM_TRACE_NONCR)
	PetscPrintf(PETSC_COMM_WORLD,"\t\t[0] line %d non-CR met\n",__LINE__);
#endif
      }
    }
    return 0;
  }
  //! Mult: v2 = A * v1
  int Mult(const PromVector_base* const xB, PromVector_base* const yB) const {
    assert(this!=NULL);
    int ierr; 

    ierr = PromMatrix::Mult( xB, yB );  CHKERRQ(ierr); 
    if( C_ != NULL ) {
      const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xB);
      PromCRVector *const yy = dynamic_cast<PromCRVector*>(yB);
      if( xx == NULL || yy == NULL) { 
#if defined(PROM_TRACE_NONCR)
	PetscPrintf( PETSC_COMM_WORLD, "\t[0] PromCRMatrix::Mult NO CR %p %p\n",
		     xx != NULL, yy != NULL);
#endif
      }
      else {
	// v3 = v2 + A' * v1.
	if( Ct_ != NULL ) {
	  ierr = Ct_->MultAdd( xx->p_, yy, yy ); CHKERRQ(ierr);
	}
	else {
	  ierr = C_->MultTransposeAdd( xx->p_, yy, yy ); CHKERRQ(ierr);
	}
	ierr = C_->Mult( xx, yy->p_ );                 CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
	PetscPrintf( PETSC_COMM_WORLD, "\t[0] PromCRMatrix::Mult C[%d,%d]\n",
		     C_->getM(),C_->getN() );
#endif	
      }
    }
    else if( D_ != NULL ) {
      assert(C_ == NULL); // not now
      const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xB);
      const PromCRVector *const yy = dynamic_cast<const PromCRVector*>(yB);
      if( xx != NULL && yy != NULL ) {   
	ierr = D_->Mult( xx->p_, yy->p_ );                  CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
	PetscPrintf( PETSC_COMM_WORLD, "\t[0] PromCRMatrix::Mult D[%d,%d]\n",
		     D_->getM(),D_->getN() );
#endif
      }
      else {
#if defined(PROM_TRACE_NONCR)
	PetscPrintf( PETSC_COMM_WORLD, "\t[0] CRMatrix::Mult(D) NO CR %p %p\n",
		     xx != NULL, yy != NULL);
#endif
      }
    }
    return 0;
  }  
  //! Mult: v2 = A * v1 -- not tested
  int Mult(const PromMatrix_base* const xB, PromMatrix **out ) const {
    assert(this!=NULL && 0); // not tested
    int ierr;  

    PromMatrix *yy;
    const PromCRMatrix *const xx = dynamic_cast<const PromCRMatrix*>(xB);
    ierr = PromMatrix::Mult( xx, &yy );  CHKERRQ(ierr);
    *out = yy;
    if( C_ != NULL ) {
      if( xx == NULL || yy == NULL) { 
#if defined(PROM_TRACE_NONCR)
	PetscPrintf( PETSC_COMM_WORLD, "\t[0] PromCRMatrix::Mult NO CR %p %p\n",
		     xx != NULL, yy != NULL);
#endif
      }
      else {
	// v3 = v2 + A' * v1.
	if( Ct_ != NULL ) {
	  //ierr = Ct_->MultAdd( xx->D_, yy, yy ); CHKERRQ(ierr);
	}
	else {
	  //ierr = C_->MultTransposeAdd( xx->D_, yy, yy ); CHKERRQ(ierr);
	}
	//ierr = C_->Mult( xx, yy->p_ );                 CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
	PetscPrintf( PETSC_COMM_WORLD, "\t[0] PromCRMatrix::Mult C[%d,%d]\n",
		     C_->getM(),C_->getN() );
#endif	
      }
    }
    else if( D_ != NULL ) {
      assert(C_ == NULL); // not now
      if( xx != NULL && yy != NULL ) {   
	//ierr = D_->Mult( xx->D_, yy->p_ );                  CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
	PetscPrintf( PETSC_COMM_WORLD, "\t[0] PromCRMatrix::Mult D[%d,%d]\n",
		     D_->getM(),D_->getN() );
#endif
      }
      else {
#if defined(PROM_TRACE_NONCR)
	PetscPrintf( PETSC_COMM_WORLD, "\t[0] CRMatrix::Mult(D) NO CR %p %p\n",
		     xx != NULL, yy != NULL);
#endif
      }
    }
    return 0;
  }  

  //! MultAdd
  int MultAdd(const PromVector_base* const xB,const PromVector_base* const yB,
	      PromVector_base* const wB) const {
    assert(this!=NULL);
    int ierr;
    ierr = PromMatrix::MultAdd( xB, yB, wB );  CHKERRQ(ierr); 
    if( C_ != NULL ) {
      const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xB);
      const PromCRVector *const yy = dynamic_cast<const PromCRVector*>(yB);
      PromCRVector *const ww = dynamic_cast<PromCRVector*>(wB); 
      if( xx != NULL && yy != NULL && ww != NULL ) {
	if( ww->p_ == NULL ) SETERRQ(PETSC_COMM_SELF,-1,"!w->p_"); 
	// v3 = v2 + A' * v1.
	if( Ct_ != NULL ) {
	  ierr = Ct_->MultAdd( xx->p_, yy, ww ); CHKERRQ(ierr);
	}
	else {
	  ierr = C_->MultTransposeAdd( xx->p_, yy, ww ); CHKERRQ(ierr);
	}
	ierr = C_->MultAdd( xx, yy->p_, ww->p_ );      CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
	PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);
#endif	    
      }
    }
    else if( D_ != NULL ) {
      const PromCRVector *const xx = dynamic_cast<const PromCRVector*>(xB);
      const PromCRVector *const yy = dynamic_cast<const PromCRVector*>(yB);
      PromCRVector *const ww = dynamic_cast<PromCRVector*>(wB); 
      if( xx != NULL && yy != NULL && ww != NULL ){
	ierr = D_->MultAdd( xx->p_, yy->p_, ww->p_ );         CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
	PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);
#endif	
      }
    }
    return 0;
  }
  //!
  int SetCValues( const int nrow, int *pgeqi, const int ncol, int *pcol,
		  double *vals, const int add_type ){
    assert(this!=NULL);
    assert( C_ != NULL );
    return C_->SetValues(nrow,pgeqi,ncol,pcol,vals,add_type);
  }
  //!
  int SetCtransValues( const int nrow, int *pgeqi, const int ncol, int *pcol,
		       double *vals, const int add_type ){
    assert(this!=NULL); assert( Ct_ != NULL );
    return Ct_->SetValues(nrow,pgeqi,ncol,pcol,vals,add_type);
  }
  //!
  int SetDValues( const int nrow, int *pgeqi, const int ncol, int *pcol,
		  double *vals, const int add_type ){
    assert(this!=NULL);
    assert(D_!=NULL);
    return D_->SetValues(nrow,pgeqi,ncol,pcol,vals,add_type);
  }
  //!
  int ZeroEntries(){
    assert(this!=NULL);
    int ierr = PromMatrix::ZeroEntries(); CHKERRQ(ierr); 
    if( C_ != NULL ) { 
      ierr = C_->ZeroEntries(); CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif    
    }
    if( Ct_ != NULL ) { 
      ierr = Ct_->ZeroEntries(); CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif    
    }
    if( D_ != NULL ) { 
      ierr = D_->ZeroEntries(); CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif    
    }
    return 0;
  }
  //!
  int Assembly(){
    assert(this!=NULL);
    int ierr = MatAssemblyBegin( mat_, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
    ierr = MatAssemblyEnd( mat_, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
    if( C_ != NULL ) { 
      ierr = MatAssemblyBegin(C_->mat_, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
      ierr = MatAssemblyEnd(C_->mat_, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif    
    }
    if( Ct_ != NULL ) { 
      ierr = MatAssemblyBegin(Ct_->mat_, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
      ierr = MatAssemblyEnd(Ct_->mat_, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif    
    }
    if( D_ != NULL ) { 
      ierr = MatAssemblyBegin(D_->mat_, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
      ierr = MatAssemblyEnd(D_->mat_, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif    
    }
    return 0;
  } 
  //!
  int FlushAssembly(){
    assert(this!=NULL);
    int ierr = MatAssemblyBegin( mat_, MAT_FLUSH_ASSEMBLY ); CHKERRQ(ierr);
    ierr = MatAssemblyEnd( mat_, MAT_FLUSH_ASSEMBLY ); CHKERRQ(ierr);
        if( C_ != NULL ) { 
      ierr = MatAssemblyBegin(C_->mat_, MAT_FLUSH_ASSEMBLY ); CHKERRQ(ierr);
      ierr = MatAssemblyEnd(C_->mat_, MAT_FLUSH_ASSEMBLY ); CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif    
    }
    if( Ct_ != NULL ) { 
      ierr = MatAssemblyBegin(Ct_->mat_, MAT_FLUSH_ASSEMBLY ); CHKERRQ(ierr);
      ierr = MatAssemblyEnd(Ct_->mat_, MAT_FLUSH_ASSEMBLY ); CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif    
    }
    if( D_ != NULL ) { 
      ierr = MatAssemblyBegin(D_->mat_, MAT_FLUSH_ASSEMBLY ); CHKERRQ(ierr);
      ierr = MatAssemblyEnd(D_->mat_, MAT_FLUSH_ASSEMBLY ); CHKERRQ(ierr);
#if defined(PROM_TRACE_CR)
      PetscPrintf(PETSC_COMM_WORLD,"\t[0] line %d CR method\n",__LINE__);    
#endif    
    }
    return 0;
  } 
  // all-at-once MG helpers
  PromMatrix *D_;
  PromMatrix *C_;
  PromMatrix *Ct_;
  PromMatrix *CDinvCt_;
};

#if defined(PROM_HAVE_COMPLEX)
#include <complex>
typedef std::complex<double> PromComplex;
#endif 

/***********************************************************************
 *                              PromComplexVector                      *
 ***********************************************************************/
class PromComplexVector : public PromVector
{ 
public:
  //!
  PromComplexVector( const PromMap &map ) : PromVector(map, 2){
#if !defined(PROM_HAVE_COMPLEX)
    PetscPrintf(PETSC_COMM_WORLD,"[0]PromComplexVector: not supported ERROR\n"); 
#endif
  }
  PromComplexVector(const PromVector *xx) : PromVector(xx->map_, 2){
#if !defined(PROM_HAVE_COMPLEX)
    PetscPrintf(PETSC_COMM_WORLD,"[0]PromComplexVector: not supported ERROR\n"); 
#endif  
  }
  virtual int Sqrt();
  virtual int SetValues( const int nrow, int *rowp, const double *vals,
			 const int add_type);
  virtual int SetValues_i( const int nrow, int *rowp, const double *vals,
			   const int add_type );
  virtual int Norm2( double * const norm ) const;
  virtual int NormMax( double * const norm ) const;
  virtual int Normalize( double * const norm ) {
    assert(this!=NULL); 
    int ierr = Norm2( norm ); CHKERRQ(ierr); 
    double a = *norm;
    if( a == 0.0 ) {
#if defined( PETSC_USE_BOPT_g )
      PetscPrintf(PETSC_COMM_WORLD,"[0]%s line %d ERROR: |x| = 0.0",__FILE__,__LINE__);
#endif
    } 
    else {
      a = 1. / a;
#if defined(PROM_HAVE_COMPLEX) 
      PromComplex aa(a,0.0);
      ierr = cScale( aa ); CHKERRQ(ierr);
#endif
    } 
    return 0;
  } 
  virtual int getN()const{assert(this && vec_); return vec_->map->N/2;}
  virtual int getn()const{assert(this && vec_); return vec_->map->n/2; }
#if defined(PROM_HAVE_COMPLEX) 
  int cAYPX( const PromComplex &a, const PromVector_base *X);
  int cWAXPY( const PromComplex &a, const PromVector_base *X, const PromVector_base *Y);
  int cAXPY( const PromComplex &a, const PromVector_base *X);
  int cMAXPY(const int nv, PromComplex a[],const PromVector_base*const*X);
  int cDot(const PromVector_base*const b, PromComplex &dot )const;
  int cMDot(const int nv, const PromVector_base*const*Xarr, PromComplex arr[])const;
  int cScale( PromComplex &a );
  int cAXPBY( PromComplex &a, PromComplex &b, const PromVector_base *const X );
#endif
  virtual int AYPX( double a, const PromVector_base *const X){
    assert(this!=NULL); assert(a==1. || a==-1.);
    int ierr = PromVector::AYPX( a, X ); CHKERRQ(ierr);
    return 0;
  }
  virtual int AXPY(double a, const PromVector_base *const X){
    assert(this!=NULL); assert(a==1. || a==-1.);
    int ierr = PromVector::AXPY( a, X ); CHKERRQ(ierr);
    return 0;
  }
  virtual int Scale( double a ){
    int ierr = PromVector::Scale( a ); CHKERRQ(ierr);
    return 0; 
  }
  virtual int MAXPY(const int nv, double a[],const PromVector_base*const*X){
    SETERRQ(PETSC_COMM_SELF,1," real method called for Complex");
  }
  virtual int Dot(const PromVector_base*const b, double *const dot )const{
    SETERRQ(PETSC_COMM_SELF,1," real method called for Complex");
  }
  virtual int MDot(const int nv, const PromVector_base*const*Xarr, 
		   double arr[] )const{
    SETERRQ(PETSC_COMM_SELF,1," real method called for Complex");
  }
  virtual int AXPBY( double a, double b, const PromVector_base *const X ) {
    SETERRQ(PETSC_COMM_SELF,1," real method called for Complex");
  }
  virtual int PointwiseMult( const PromVector_base *const X, 
			     const PromVector_base *const Y );
  virtual int PointwiseDiv( const PromVector_base *const X,
			    const PromVector_base *const Y );
  virtual int Reciprocal();
  virtual int Print(char *str=NULL, FILE *file=stderr, int limit=100000)const{
    double *arr; int n = 2*getn();
    GetArray( &arr );
    PromContext::printArr( file, arr, (limit > n) ? n : limit,
			   (str==NULL) ? "vector[%d]" : str, 2 );
    RestoreArray( &arr );
    return 0;
  }
  virtual bool isComplex()const{return TRUE;}
};

/***********************************************************************
 *                           PromComplexMatrix                         *
 ***********************************************************************/
class PromComplexMatrix : public PromMatrix
{
public:
  //! constructor
  PromComplexMatrix(const PromGrid *grid, int nA,int *pnA,int nB,int *pnB);
  //!
  PromComplexMatrix( const PromMap &map, int nA,int *pnA,int nB,int *pnB,
		     int lcols=-1, int gcols=-1);
  //! Copy constructor
  PromComplexMatrix( const PromComplexMatrix* AA ) : PromMatrix( AA ) {}
  virtual int GetOwnershipRange(int *first, int *end)const;
  virtual int getN()const{assert(this && mat_); return mat_->cmap->N/2;}
  virtual int getn()const{assert(this && mat_); return mat_->cmap->n/2; }
  virtual int GetLocalNodeRow( const int brow, const int len, int &ncols,
			       int adjacs[], double **Aij = NULL ) const;
  virtual int GetRow(int row, int *ncols,int **colpp=NULL,double **arrdat=NULL)const;
  virtual int GetRow_i(int row, int *ncols,int **colpp=NULL,double **arrdat=NULL)const;
  virtual int RestoreRow(int row,int*ncols,int**colpp=NULL,double**arrdat=NULL)const;
  virtual int SetValues( const int nrow, int *pgeqi, const int ncol, int *pcol,
			 double *vals, const int add_type );
  virtual int SetValuesBlocked( const int nrow, int *pgidi, const int ncol,
				int *pcol, double *vals, const int add_type);
  virtual int ZeroRows( PromIS is, double diag = 0.0 );
  int SetValues_i( const int nrow, int *pgeqi, const int ncol,
		   int *colp, double *vals, const int add_type );
  int SetValuesBlocked_i( const int nrow, int *pgidi, const int ncol,
			  int *pcol, double *vs, const int add_type);
  virtual int Scale( double val ){
    SETERRQ(PETSC_COMM_SELF,1," Scale() Not Supproted for Complex");
#if defined(PARCH_win32)
    return 0;
#endif
  }
  virtual int Scale( const PromVector_base * const l, 
		     const PromVector_base * const r ){
    return PromMatrix::Scale(l,r);
  }
  virtual int NormFrob( double *const norm ) const{
    SETERRQ(PETSC_COMM_SELF,1," NormFrob() Not Supproted for Complex");
#if defined(PARCH_win32)
    return 0;
#endif
  }
  virtual bool isComplex()const{return TRUE;}
private:
  int SetValuesBl_private(const int nrow, int *prow, const int ncol, int *pcol,
			  double *vals, const int add_t, const bool iscomplex);
};

//! class PromGridBase: Base (virtual) grid class for PromGrid.
/*!
  Data for PromGrid - used for common constructor and simple methods.
*/
class PromGridBase
{
public:
  //! PromGridBase constructor
  PromGridBase( const PromOptions &opts, 
		PromPerfMonitor &perf,
		PromParentBank * const bank, const int nloc, 
		const int fact, 
		const PromComm &wcomm );
  virtual ~PromGridBase();
  int BuildComunicator(const PromComm &wcomm);
  int getNewTag()const{ 
    assert(this!= NULL);
    int tag; 
    if( gComm_ != MPI_COMM_NULL ) PetscCommGetNewTag( gComm_, &tag );
    else tag = -3;
#if defined(PETSC_USE_BOPT_g)
    int mype;
    if(gComm_ != MPI_COMM_NULL) {
      MPI_Allreduce( &tag, &mype, 1, MPI_INT, MPI_MAX, gComm_); 
      assert(mype==tag);
    }
#endif
    return tag;
  }
  // data
public:
  int           factor_;   // = 1
  int           nLocalNd_; // num. local in 'nNodes_' (used for inactive grids)
  int           nNodes_;   // size of nodes array
  // solver data structures
  int           nBlocks_;
  PromIS       *blockISs_;
  // list of Patent_Arr <np, np*<PARENT> >
  PromTable<PromParentArr_extra*> *gid_extraParents_;
  // blocks to put in one asm
  //! data
  PromMatrix *stiffness_;          //
  PromMatrix *prolongator_;        // 
  PromMap      *scalarMap_;     // scalar map for this
  PromMap      *CRmap_;              // cache for constraint map for MG-KKT
  //!
  const PromOptions &options_;
  PromPerfMonitor &perf_mon_;
  PromParentBank * const    parBank_;
  PromTable<PromNode*>      ghosts_; // table of node pointers, keyed to global ids+1
  PromIS                   *CRAggs_; // cache for CR aggs
  int                       nCRAggs_;
  // comms
  MPI_Group     group_;
  MPI_Group     mpi_group_world_; 
  MPI_Comm      gComm_;
};

//! class PromGrid: Base (virtual) grid class for Prometheus.
/*!
  Common Graph methods for Prometheus - pure virtual.
*/
class PromGrid : public PromMap, public PromBank, public PromGridBase 
{
public:
  //! fine grid, normal and dummy grid constructor (const dof per node)
  PromGrid(const PromOptions &opts, PromPerfMonitor &perf, 
	   PromParentBank * const bank, const int nloc, 
	   const PromComm &wcomm, const int bsize );
  //! reduced coarse grid (const dof per node)
  PromGrid( const PromOptions &opts, PromPerfMonitor &perf, 
	    PromParentBank * const bank, const int partloc, 
	    const int factor, const PromComm &wcomm, const int bsize );
  //! fine grid and normal grid constructor
  PromGrid( const PromOptions &opts, PromPerfMonitor &perf, 
	    PromParentBank *const bank, const int nloc, 
	    const PromComm &wcomm, const int lid_ndof[] );
  //! reduced coarse grid
  PromGrid( const PromOptions &opts, PromPerfMonitor &perf, 
	    PromParentBank * const bank, const int partloc, const int factor,
	    const int nLocPartEq, const PromComm &wcomm );
  //! PromGrid deconstructor. 
  virtual ~PromGrid();
  // matrix triple product routines
  int          CreateProlongator( PromGrid* const lastgrid,
				  const PromComm &comm );
  int          MakeNextMat( PromGrid *nnext );
  int          MakeScalarNextMat( PromGrid *lastGrid ); 
  int          SmoothScalarP( PromGrid *nextG ) const ;
 private:
  int          MakeNextMat_private( PromGrid *nnext );
 public:
  //!  Archive grid (read/write or just print). 
  virtual int  Archive(FILE*file=stderr,const PromArchiveType type=PROM_PRINT);
  // pure virtual 
  virtual int  FineGridInitWithA() = 0;
  virtual int  SetCoords( const double coords[], const int ndm = 0,
			  const double * const constoffsets = NULL ) const = 0;
  virtual int  SetVects( const double vects[], const int nvecs ) = 0;
  virtual int  PrepForSolve();
  virtual int  MakeInitAdjacs()=0;
  virtual int  CreateCoarseGridSpace( PromGrid *lastgrid )=0;
  virtual PromNode* newNode( const int lid )=0;
  virtual PromMGType getMGType()=0;
  virtual int  Coarsen()=0;
  //
  virtual int  Repartition( PromGrid *lgrid );
  // fine grid methods
  int          SetIndex( const PromTable<int> &gid_ghostLid );
  int          GetGhostData( const bool fulldata = TRUE );
  // general grid methods
  int          GetProc( const int gid, int &proc ) const;
  bool         isActive()const;
  bool         isActive( int proc )const;
  bool         isBase()const{ assert(this!=NULL); return (prevgrid_==NULL); }
  bool         isTop()const{ assert(this!=NULL); return (nextgrid_ == NULL); }
  int          getLevel()const;
  int          getNumLevels()const;
  int          NumSelected( int &nloc, int &nnodes, int &nneq ) const ; 
  // addghosts
  int          addGhost( PromNode *&currj, const int procj, const int globalID,
			 const int realID, const int gtag  );
  int          ReceiveAddGhost( int todo, const int gtag,
				const PromTable<int> &procTable );
  // others
  int          SetNodes( PromGrid *last, const int buffsz, int buff[],
			 const int t2,  
			 PromTable<int> &ghost_gid1_lid1, 
			 PromList<PromMPINode*> &ghosts );
  int          ExchangeAdjacs( const int lastfac, MPI_Comm lcomm );
  int          RepartitionTop( const PromComm &comm );
  int          SendStatesToSlaves();
  int          ReceiveStates();
  static int   FindProc(const int *p_g,const int np, const int gid,int &proc);
  int          AddExtraGhostsParents( PromGrid * const nextg,
				      const PromMatrix *const CC = NULL );
  // repartitioning
  int          Repartition_private( PromGrid *lgrid );
  //  solve stuff
  int          MakeASMBlocks();
  int          MakeMatrix(bool complex=FALSE, bool useCRmat=FALSE, 
			  const int extras=0, PromTable<PromCR*> *CR_id=NULL, 
			  void *mat = NULL);
  // assembly helpers
  int          AssembleTable( PromGrid *lastG );
  int          AddValueTable( const int I, const int J, const int ndfI,
			      const int ndfJ,double *cache, const int proci,
			      const bool iscomplex = false );
  int          MakeProlongator(const PromGrid*const lastgrid, PromMatrix*prol=0,
			       const bool clearDAP=TRUE,int *nnz=0,int *onnz=0);
  //! allocate nodes.
  int          InitNodes( const int nnodes ); // alloc nodes
  //! fine grid init.
  int          FineInit( const int lid_ndof[], int bsize,
			 const PromTable<int> &gid_ghostLid,
			 const PromList<int*> *elemList );
  //! general helpers
  int          masterProc()const;
  int          masterProc( const int p )const;
  int          getEndProc()const;
  int          getEndProc( const int p )const;
  bool         isSameGroup( int proc1, int proc2)const;
  // write file 
  int         WriteFEAP( int lev, MPI_Comm comm, 
			 PromList<PromPoly*> * elist,
			 PromList<PromFacet*> * flist = NULL )const;
  int         WriteFacets(FILE *file, PromList<PromFacet*> *facets ) const;
  int         WriteFEAP( int lev, PromTable<PromPoly*> *elems )const;
  int         WriteMATE( FILE *file )const;
private:
  int          GlobalizeAdjac(const int mynew0, PromList<int> *newLocAdj_ogid,
			      PromList<int> *newGhsAdj_oldgid,
			      PromTable<int> &oldgid_newgid,
			      PromTable<int> &ghostoldgid_newlid );
  int          DistributeGrid( PromGrid *prevG, PromTable<int> &oldgid_newgid );
  int          CreateNodes( const int lid_ndof[], int bsize );
protected:
  //! normal coarse init.
  int          InitNormal( PromGrid *lastgrid );
  //! dummy coarse init.
  int          InitDummy( PromGrid *lastgrid );
  //! reduced coarse init
  int          InitReduced( PromGrid *lastgrid, const int factor,
			    const int buffsz, int *buffer, 
			    const int rtag, PromTable<int> &gid1_lid, 
			    PromList<PromMPINode*> &ghostlist );
  //! coarse constructor helpers.
  int          CoarsePreInit( int nloc, PromGrid *lastgrid );
  int          CoarseNormalInit( PromGrid *lastgrid );
  int          CoarseReduceInit(PromGrid *last, const int buffsz, int nldata[],
				const int tag, PromTable<int> &gid1_lid1, 
				PromList<PromMPINode*> &ghosts );
  int          CoarsePostInit( PromGrid *lastG );
  
  int          ReceiveGhost( const int proc, int *&pb );
  int          ReceiveDep( const int rtag, const int sz, const int part, 
			   PromList<int*> &sendlist, 
			   PromList<int*> &sendlist2 );
  int          ReceiveAdjac( const int tag, const int tag2, int &todo );
  int          CollectNewDeps( int todo2, int nrecv, const int rtag, 
			       const int stag );
  int          ReceiveGhostDep( const int rtag, const int stag1,
				const int pr, const int sz);
  int          AddDep( PromNode *curri, const int procj, const int locidj, 
		       PromList<int*> &sendlist2 );
protected:
  int          FixLastGrid( PromTable<int> &oldgid_newgid, 
			    const int oldproc_gnode[],
			    const int lid_part[], PromGrid *lgrid );
  int          Repartition( int *&lid_part, int *&part_c, int *&part_count );
  int          MoveGrid( PromTable<int> &oldgid_newgid, 
			 PromTable<int> &oldghostgids_newlid, 
			 const int newproc_gnode[], 
			 const int lid_part[], const int part_c[],
			 PromList<int> *&newLadjs, PromList<int> *&newGadjs);
  int          CollectGhostData(PromTable<int> &oldgid_newgid, 
				PromTable<int> &oldghostgids_newlid,
				const int oldproc_gnode[], const int lid_part[],
				const PromList<int> oldLadjs[],
				PromList<int> newGadjs[] );
  int          ReduceCounts(int lid_part[], int *&part_c, int *&newproc_gnode);
  // resoolve ghosts
  int          ResolveGhostData( const PromGrid* );
  int          Resolve_phase1(const PromGrid * lastG,PromTable<int> &procTable,int&todo);
  int          Resolve_phase2( PromTable<int> &procTable,
			       int todo, int *newghosts );
  int          Resolve_phase3( const PromGrid *lastG, 
			       const PromTable<int> &procTable,
			       const int newghosts );
  int          Resolve_phase3_act( const PromGrid *lastG, 
				   const PromTable<int> &procTable, 
				   const int newghosts, const int tag3 );
  int          Resolve_phase3_ina( const PromGrid *lastG, const int tag3 );
  int          LocalGhostDepSwapp( const PromGrid *lastG,
				   PromTable<int> &procTable,
				   const int nrealb, const int ndeps, int &todo);
  //
  int          CreateMIS( PromList<PromNode*> &list,
			  int is_del( const PromNode* const, 
				      const PromNode* const, bool& ),
			  const bool virginMIS = FALSE );
  int          SquareGraph( int **orig_adj_out );
  int          UnSquareGraph( int **orig_adj_out );  
  int          ReceiveState_NC( PromTable<PromNode*> &gid_ghost, const int sz, 
				int &ndone, int &todo2, int &todo, 
				const int rtag, const int stag, const int part,
				MPI_Comm comm );
  int          ReadMISFile( FILE *file, int &flag /*out*/);
  static inline double* GetCacheEntry( double cache[],const int nParI, 
				       const int I_ceq[], int J_ceq[],
				       int &nCorsCols, 
				       bool dirtyCache[],
				       int parGIDcache[], const int J, 
				       const int Jndf, const int II );
  int          AddCache(double cache[], const int nParI,
			const int I_ceq[], const int J_ceq[], 
			const int nCorsCols, const PromParent * const * parI,
			const bool dirtyCache[],
			const int parGIDcache[], const bool iscomplex = false);
private:  
  int  MakeNodeBlocks( const int overlap = 0 );
  int  PartitionBlocksNodes( const int nblocks, int pnode_part[] );
public: 
  int          SetParents( PromGrid* );
  int          SetBoundNodes( PromGrid *symbG );
  //! Allocate new array of pointers to nodes and copy old node pointers in
  int          AddNNodePointers( int nn );
  
  // data
  PromNode    **nodes_;  // pointer to array of pointers to nodes
  PromGrid     *prevgrid_; 
  PromGrid     *nextgrid_;

  int         WriteElems( FILE *file, FILE *efile, 
			  const PromList<PromPoly*> * const elems, 
			  MPI_Comm comm)const;
};

//! class PromGeoGrid: Geometric MG grid.
/*!
  Geometric MG grid.
*/
class PromGeoGrid : public PromGrid
{
  //! PromGeoGrid private general constructor. 
  int InitEnvNodes();
public:
  //! fine PromGeoGrid constructor (const dof per node)
  PromGeoGrid( const PromOptions &opts, PromPerfMonitor &perf, 
	       PromParentBank *const bank, const int nloc, 
	       const PromTable<int> &gid_ghostLid,
	       const PromList<int*> *elemList,
	       const PromComm &wcomm, int bsize );
  //! normal coarse grid constructor (const dof per node)
  PromGeoGrid( const PromOptions &opts, PromPerfMonitor &perf,
	       PromGeoGrid *lastgrid, const int nloc, const int nnodes, 
	       const PromComm &wcomm, int bsize );
  //! dummy coarse grid constructor (const dof per node)
  PromGeoGrid( const PromOptions &opts, PromPerfMonitor &perf,
	       PromGeoGrid *lastgrid, const int fact, const PromComm &wcomm,
	       int bsize );
  //! Reduced coarse grid constructor (const dof per node)
  PromGeoGrid( const PromOptions &opts, PromPerfMonitor &perf, 
	       PromGeoGrid *lastgrid, const int factor, const int buffsz, 
	       int *buffer, const int tag, PromTable<int> &gid1_lid, 
	       PromList<PromMPINode*> &ghostlist, const int partloc,
	       const int nnodes, const PromComm &wcomm, int bsize );
  //! PromGeoGrid deconstructor. 
  virtual ~PromGeoGrid(); 
  virtual int  FineGridInitWithA();
  virtual int  SetCoords( const double coords[], const int ndm = 0,
			  const double*const offsets=NULL)const;
  virtual int  SetVects( const double vects[], const int nvecs );
  int BuildElemList( const PromTable<int> &gid_ghostLid, 
		     const PromList<int*> *elemList );
  int MakeElemAdjacs();
  // virtual methods
  virtual int CreateCoarseGridSpace( PromGrid *lastgrid );

  // coarse grid
  virtual int MakeInitAdjacs();
  virtual int Repartition( PromGrid *lgrid ){
    int ierr = PromGrid::Repartition( lgrid );   CHKERRQ(ierr); 
    ierr = DeleteElemList();  CHKERRQ(ierr); 
    ierr = DeleteTets();  CHKERRQ(ierr);
    ierr = DeleteFacets();  CHKERRQ(ierr);
    return 0;
  }
  virtual int PrepForSolve();
  virtual PromNode *newNode( const int lid );
  virtual PromMGType getMGType();
  // geo methods
  int         SetNodeTypes( const int tag = -1 );
  virtual int Coarsen();
  int         SetParents();
  int         CreateGeometricCoarseSpace( PromGeoGrid * const lastGrid );
protected: 
  // mesh helpers 
  int         Triangulate();
  int         AddToMesh( PromGeoNode *node, PromDTet *elem );
  int         GetCavity( PromDTet*, PromGeoNode*, PromTable<PromDTet*>*,
			 gfloat *vol );
  int         OffsetParts( PromGeoGrid * const lastGrid,
			   const PromCoord &factor, PromCoord &minCoord, 
			   PromCoord &maxCoord, const bool coordsIN );
  int         findNodeElems ( int &lost ) const;
  // FixMesh methods  
  int         FixMesh( PromList<PromPoly*> *elemList );
  int         RemoveDisconnectedPartElems(int &,PromList<PromPoly*> *elemList);
  int         RemoveUnconnectedElems_easy(PromList<PromPoly*> *elemList);
  int         RemoveEnvElems( int &, PromList<PromPoly*> *elemList );
  int         RemoveOutsideSlivers( int &, PromList<PromPoly*> *elemList );
  int         RemoveOutsides( int &, PromList<PromPoly*> *elemList );
  int         MoveNodes( int &nunfound, const int, gfloat &maxoff, 
			 PromList<PromPoly*> *elemList  );
  int         ResolveNodes( int numUnfound, gfloat &maxoff, 
			    PromList<PromPoly*> *elemList );
public:       // real adjacs 
  int         MakeExactAdjac( PromGrid *lastG );
  int         DeleteElemList(); // deletes all elements out of elemList
protected: 
  int         MakeExactAdjac_private( PromGrid *lastG ); 
  int         MakeAdjacData( PromGrid *lastG, const int gtag, int &todo, 
			     PromTable<PromTable<PromTable<int>*>*> &proci_tab,
			     PromTable<int> &procTable);
  int         SendAdjacData( const int tag, PromTable<PromTable<PromTable<int>*>*> &proci_t_g,
			     MPI_Request sreqs[], int *sbuffs[]);
  // 'MoveNodes' helpers
  int         GetListOfFacets( PromGeoNode*, PromGeoNode *[][2], int &, 
			       PromList<PromStack*> edgeLists[], 
			       PromList<PromStack*> &, PromList<PromStack*> &, 
			       PromCoord &avNorm, gfloat &scale, bool &noop);
  int         MoveNode( PromGeoNode *curr, const PromCoord &avNorm, 
			gfloat &scale, PromList<PromPoly*> *elemList ); 
  int         ResetUnfound_nodes(PromGeoNode*, PromGeoNode*[][2], const int , 
				 PromList<PromStack*> edgeLists[], 
				 PromList<PromStack*> &, 
				 PromList<PromStack*> &,
				 int &, const bool, const bool );
  int         GetFaces( PromGeoNode *curr, PromGeoNode *edges[][2], 
			const int numfacets, int surf[][5], int topfacets[5],
			int &);
  int         getNeighborChildren( const PromGeoNode *curr,
				   PromList<PromStack*> edgeL[],
				   PromGeoNode *es[][2], const int );  
  int         GetScaleFactor( PromGeoNode *curr, const PromCoord &avNorm,
			      PromList<PromStack*> *unfound, 
			      PromList<PromStack*> *ed,
			      PromGeoNode *[][2], int, gfloat & );
  int         ResetChildElems( PromList<PromPoly*> *elems,
			       PromList<PromStack*> *, 
			       PromList<PromPoly*> *elemList, 
			       bool *killed = NULL,
			       const bool kill = FALSE );
  int         CheckForUpdate( bool respond, PromList<PromPoly*> *elemList );
  int         EmptyChildren( PromGeoNode *curr );
  // SetNodeTypes helpers
  int         MakeFacets( PromList<PromFacet*> *facets );
  int         SetFaceIDs( PromList<PromFacet*> &facets, 
			  PromList<PromFacet*> flists[], 
			  int *nfaces=NULL, 
			  PromList<PromFacet*> *seedFacets=NULL, 
			  PromTable<int> *lwid_hgid_alias=NULL )const;
  int         CollectSeeds( const int tag, PromList<int> &higherProcs, 
			    PromList<PromFacet*> &facets, 
			    PromList<PromFacet*> &seedFacets )const;
  int         SendSeeds( const int tag, PromList<int> &lowerProcs, 
			 PromList<PromFacet*> &facets )const;
  int         GlobalizeFacets( PromList<PromFacet*> &facets, 
			       PromTable<int> &lwid_hgid_alias )
    const;
  int         SetTopoTypes( PromList<PromFacet*> flists[] );
  int         MakeNodeList( PromList<PromNode*> &list, const bool print )const;
  // mesh helpers
  int         evacuateChildren( PromDTet *);
  int         retireElement( PromDTet *deadElem );
  //
  int         CleanElemList(); // cleans deleted elements out of elemList
  //data 
  PromGeoNode *env_[PROM_NUMENV]; // envelope nodes
  PromList<PromPoly*>  elemList_; // list of active elements on current level
};

//! class PromAggGrid: Aggregation grid base class (virtual).
/*!
  Aggregation grid base class (virtual).
*/
class PromAggGrid : public PromGrid
{
  int SetAggMethod(); // fine
public:
  //! fine PromAggGrid constructor (const dof per node) - algebraic
  PromAggGrid( const PromOptions &opts, PromPerfMonitor &perf, 
	       PromParentBank *const bank, const int nloc, 
	       const PromTable<int> &gid_ghostLid, 
	       const PromList<int*> *elemList,
	       const PromComm &wcomm, const int ndf );
  //! normal coarse grid (const dof per node)
  PromAggGrid( const PromOptions &opts, PromPerfMonitor &perf,
	       PromAggGrid *lastgrid, const int nloc, const int nnodes, 
	       const PromComm &wcomm, const int ndf );
  //! dummy coarse grid (const dof per node)
  PromAggGrid( const PromOptions &opts, PromPerfMonitor &perf, 
	       PromAggGrid *lastgrid, const int fact, const PromComm &wcomm,
	       const int bsize );
  //! reduced coarse grid (const dof per node)
  PromAggGrid( const PromOptions &opts, PromPerfMonitor &perf, 
	       PromAggGrid *lastgrid, const int factor, const int buffsz,
	       int *buffer, const int tag, PromTable<int> &gid1_lid, 
	       PromList<PromMPINode*> &ghostlist, const int partloc, 
	       const int nnodes, const PromComm &wcomm, const int ndf );
  //! fine PromAggGrid constructor
  PromAggGrid( const PromOptions &opts, PromPerfMonitor &perf, 
	       PromParentBank *const bank, const int nloc, 
	       const PromTable<int> &gid_ghostLid, 
	       const PromList<int*> *elemList,
	       const PromComm &wcomm, const int lid_ndof[] ); 
  //! normal coarse grid
  PromAggGrid( const PromOptions &opts, PromPerfMonitor &perf,
	       PromAggGrid *lastgrid, const int nloc, const int nnodes, 
	       const PromComm &wcomm, const int lid_ndof[]);
  //! dummy coarse grid
  PromAggGrid( const PromOptions &opts, PromPerfMonitor &perf, 
	       PromAggGrid *lastgrid, const int fact, const PromComm &wcomm);
  //! reduced coarse grid
  PromAggGrid( const PromOptions &opts, PromPerfMonitor &perf, 
	       PromAggGrid *lastgrid, const int factor, const int buffsz,
	       int *buffer, const int tag, PromTable<int> &gid1_lid, 
	       PromList<PromMPINode*> &ghostlist, const int partloc, 
	       const int nnodes, const int nLocPartEq, const PromComm &wcomm );
  //! destructor
  virtual     ~PromAggGrid();
  //! methods
  virtual int  FineGridInitWithA();
  virtual int  SetCoords(const double coords[], const int ndm = 0,
			 const double*const constoffsets=NULL)const;
  virtual int  SetVects( const double vects[], const int nvecs );
  virtual int  ScaleRBM( PromVector *right )const;
  // amg methods
  virtual int PrepForSolve();
  virtual PromNode *newNode( const int lid );
  virtual PromMGType getMGType();
  // coarse grid
  virtual int MakeInitAdjacs();
  virtual int CreateCoarseGridSpace( PromGrid *lastgrid );
  // agg graph 
  int         MakeFirstParentLayer();
  int         AddNextAdjacs( PromAggGrid *nnext );
  int         AllocPar0( PromAggGrid *nnext );
  virtual int Coarsen();
  //
protected: 
  int         ConstructAggs();
  int         PartitionAggs();
  int         MakeWeightMatrix( const PromMap &map, int ndwts[],
				idx_t **padj, bool mis2 = false );
  int         MakeWeightMatrix( int ndwts[], idx_t **padj );
  int         SendAggUpdates( PromTable<PromList<PromNode*>*> &proc_list,
			      PromList<MPI_Request*> &smessages, 
			      const int tag, const int kk );
  int         CapAndRecvUpdates( PromTable<int> &procTable, const int tag, const int gtag, int &todo, 
				 PromList<PromAggNode*> &agglist, const int kk );
  int         AddDepsToList( PromTable<int> &procTable );
  int         SetSelCaches( int &nselected, const bool global_agg,
			    PromTable<int> *proc_nsel );
  int         SetAMGEdgeWeights( PromMatrix *A, idx_t *adjncy,
				 const int numadjacTOT, int *egwtsTOT );
  int         PickSubdomains( const int pnode_part[] );
public:  
  //
  int         SmoothP()const;
  int         SmoothP_gs()const;
protected: 
  int         MakeAMGASMBlocks();
  int         FormDAP( const PromMatrix * const AA ) const;
  int         ApplySmoothing( const PromMatrix*const, double omega )const;
  int         FormDAP_gs( const PromMatrix * const AA ) const;
  int         MakeBlocksFromAggragates();
  // methods
public:
  int QRShapes( const PromAggGrid *lastGrid )const;
  // debug
  int checkParents()const;
protected: 
  int SendQRShapes( const PromAggGrid *lastGrid )const;  
  // data
public:
  int amg_factor_;
  int amg_smooths_;
  mutable PromVector *A_RBMs_[PROM_MAXNDF];
};

#endif //  __PROM_GRID_H__
