//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: coar_resolve.C,v 1.3 2011-01-22 20:13:39 madams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           coar_resolve.C
//
#include "prom_grid.hh"
 
#undef __FUNCT__  
#define __FUNCT__ "ResolveGhostData"  
/* PromGrid::ResolveGhostData *************************************************
 *
 *  - communicate ghost ID data, and adjacentcies, with neighbors
 *    copes with first (simple) step of aglomeration
 *
 *  INPUT:
 *    - lastG: finer grid, coarsened
 *
 *  SIDE EFFECTS:
 *    - kills nodes w/o adjacs
 * 
 *  RETURN:
 *     - PETSc error code
 */  
#define ISLOCAL(p) (p>=master&&p<endp)
int PromGrid::ResolveGhostData( const PromGrid *lastG )
{
  int              ierr, todo, xx, newgh, vb = options_.verbose_;   
  PromTable<int>  procTable( 57 ); assert(lastG!=NULL && lastG->isActive());
  
  if(vb>1)PetscPrintf(PETSC_COMM_WORLD,"[%d]%s level %d\n",
		      comm_.mype(),__FUNCT__,getLevel());

  // get base ghost data
  ierr = Resolve_phase1( lastG, procTable, todo ); CHKERRQ(ierr);
  if( isActive() ) {
    // set full ghost data structs - can 'addGhost'
    ierr = Resolve_phase2( procTable, todo, &newgh );  CHKERRQ(ierr);
  }
  else newgh = 0;

  // reset new ghost local ids
  ierr = Resolve_phase3( lastG, procTable, newgh ); CHKERRQ(ierr);
  
  // clear last marks
  for( xx = 0 ; xx < lastG->nNodes_ ; xx++ ){
    PromNode *curr = lastG->nodes_[xx];
    curr->mark(FALSE);
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "Resolve_phase1"  
/* PromGrid::Resolve_phase1 ******************************
 *
 *  -  get base ghost data
 *
 *  INPUT:
 *    - lastG: finer grid, coarsened
 *    - procTable: 
 *    - todo: the number of tag2 sent (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::Resolve_phase1( const PromGrid *lastG, 
			      PromTable<int> &procTable,
			      int &todo )
{  
  MPI_Status status; 
  int        ierr, xx ,newOtherID, oldLocID, gid,proc,ndeps,tt,ndf,geq,nrealb;
  const int tag1 = lastG->getNewTag();
  const int mype = comm_.mype();
  PromNode       *ghos, *boun, *fine;
  const int master = masterProc(), endp = getEndProc();
  const int lastfactor = lastG->factor_, reqsz=sizeof(MPI_Request)/sizeof(int);
  MPI_Comm  lastcomm = lastG->gComm_; 
  PromList<MPI_Request*>  smessages; 
  PromTable<PromList<PromNode*>*> proc_list(57);
  PromList<PromNode*> *list;
  PromGlobalID *fdep,*cdep,*dep; 

  // 1a) make new dep lists in boundary nodes, send other's ghosts this data
  for( xx = todo = nrealb = ndeps = 0 ; xx < nLocalNd_ ; xx++ ){
    boun = nodes_[xx];
    if( boun->getState() == PROM_NOTUSED ) {
      PetscPrintf( PETSC_COMM_SELF, "[%d]%s: ERROR %d state=%s,type=%s\n",
		   mype,__FUNCT__,boun->index_(),
		   PromContext::stateNames_[boun->getState()],
		   PromContext::typeNames_[boun->getTopoType()] );
    }
    if( boun->symbData_->child == NULL ){ 
      assert( !boun->isBound() && factor_ > lastfactor); 
      continue;  // split processor
    }
    fine = boun->symbData_->child;               // could be a ghost
    if( !fine->isBound() ) continue;  
    
    // make new deps from fine
    Position pos2, pos = fine->symbData_->dependents->getHeadPosition();
    assert(pos!=NULL);
    while( pos ) {
      fdep = fine->symbData_->dependents->getNext( pos );  
      proc = fdep->proc();                              assert(proc!=mype);
      if(ISLOCAL(proc)){ assert(factor_!=lastfactor);continue;} // split proc
      cdep = new PromGlobalID( -1 ,proc );                CHKNEW( cdep, 1 );
      if( boun->symbData_->dependents == NULL ){
	boun->symbData_->dependents = new PromList<PromGlobalID*>(3); 
	CHKNEW( boun->symbData_->dependents, 1 );
      }
      boun->symbData_->dependents->addTail( cdep );
      assert(proc!=mype);
    }
    if( !boun->isBound() ) continue;  
    nrealb++;

    // clear dups to proc groups
    assert(endp - master == factor_);
    if( factor_ > lastfactor ) {
      pos = boun->symbData_->dependents->getHeadPosition(); assert(pos!=NULL); 
      while( (pos2 = pos) ) {
	cdep = boun->symbData_->dependents->getNext( pos );
	proc = cdep->proc();
	assert(!ISLOCAL(proc));
	Position pos3 = boun->symbData_->dependents->getHeadPosition();
	assert(pos3!=NULL);
	while( pos3 ) {
	  dep = boun->symbData_->dependents->getNext( pos3 ); 
	  tt = dep->proc();   
	  if( tt > proc && isSameGroup(tt,proc) ) { // bound to ghost - uniq.
	    boun->symbData_->dependents->removeAt(pos2);  
	    delete cdep; cdep = NULL; 
	    break;
	  }
	}
      }
    }
    assert( !boun->symbData_->dependents->isEmpty() );

    // collect send deps updates
    pos2 = fine->symbData_->dependents->getHeadPosition(); assert(pos2!=NULL);
    pos = boun->symbData_->dependents->getHeadPosition(); 
    while( pos ) {
      cdep = boun->symbData_->dependents->getNext(pos);
      proc = cdep->proc();                            assert(!ISLOCAL(proc));
      do{ fdep = fine->symbData_->dependents->getNext( pos2 );
      }while( (tt=fdep->proc()) != proc && pos2 );
      assert(proc == tt);
      // 
      if( (list=proc_list.find(proc+1)) == NULL ) {
	list = new PromList<PromNode*>(100); CHKNEW(list,1);
	proc_list.add( list, proc+1 );
      }
      list->addTail(boun); 
      list->addTail((PromNode*)fdep); // hack
      list->addTail(fine);
      //
      todo++;
      proc = masterProc(proc);
      if( proc != mype ) procTable.add( proc+1 );	
      ndeps++;
    }
  }  
  // send data
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    tt = list->getCount(); assert(tt%3==0); tt = (tt/3)*5;
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr);    smessages.addTail( sreq );
    int *buff = (int*)(sreq+1), *pb = buff;
    while( !list->isEmpty() ) {
      boun = list->removeHead(); 
      fdep = (PromGlobalID*)list->removeHead(); // hack
      fine = list->removeHead(); 
      *pb++ = fdep->local(); 
      *pb++ = boun->index_(); 
      *pb++ = boun->index_.global(); 
      *pb++ = boun->geq();
      *pb++ = boun->getNDF();
    }
    assert(tt == pb-buff);
    // send(tag1) ghost data *< oldFID, newCID, newCgid, geq, ndf >
    MPI_Isend( buff, tt, MPI_INT, proc/lastfactor, tag1, lastcomm, sreq ); 
    delete list; 
  }

  // 1b) receive ghost data, send back (killed)
  assert( ghosts_.isEmpty() ); 
  for( xx = nLocalNd_, tt = 0 ; xx < nNodes_ ; xx++ ){ 
    ghos = nodes_[xx];  
    if(!ghos->symbData_->child || ghos->symbData_->child->isMarked()) continue;
    tt++;
  }
  xx = tt;
  while( xx ) {
    MPI_Probe( MPI_ANY_SOURCE, tag1, lastcomm, &status );
    MPI_Get_count(&status, MPI_INT, &tt); proc = status.MPI_SOURCE*lastfactor;
    proc = masterProc(proc); procTable.add(proc+1); 
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // recive(tag1) ghost data < oldFID, newCID, gid, geq, ndf >
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag1, lastcomm, &status );
    while(pb-buff < tt){
      xx--; assert(xx>=0);
      PromContext::Read( oldLocID, pb ); PromContext::Read( newOtherID, pb ); 
      PromContext::Read( gid, pb ); PromContext::Read( geq, pb );
      PromContext::Read( ndf, pb );
      // get old ghost - (small) linear search
      fine = lastG->nodes_[oldLocID]; 
      assert( fine && fine->isGhost() && fine->index_() == oldLocID );
      if(fine->getState()!=PROM_SELECTED)SETERRQ(PETSC_COMM_SELF,1,"wrong state");
      ghos = dynamic_cast<PromNode*>(fine->symbData_->parent);
      assert(ghos!=NULL);
      // get and set ghost
      ghos->symbData_->ghostID.setLocal( newOtherID ); 
      ghos->index_.setGlobal( gid ); 
      ghos->setGEQ( geq ); ghos->SetNDF( ndf );
      // add to table for fast lookup (key is one based)
      bool add = ghosts_.add( ghos, gid + 1 );              assert( add );
    }
    assert(pb-buff == tt);    PetscFree(buff); 
  }
  // free
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  
  // trade ghost gids in local group
  if( factor_ > lastfactor ) {  
    ierr = LocalGhostDepSwapp( lastG, procTable, nrealb, ndeps, todo );
    CHKERRQ(ierr);
  }
  MPI_Iprobe(MPI_ANY_SOURCE,tag1,lastcomm,&tt,&status); assert(!tt);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "LocalGhostDepSwapp"  
/* PromGrid::LocalGhostDepSwapp ******************************
 *
 *  - set full ghost data structs
 * 
 *  INPUT:
 *    - lastG:
 *    - procTable (in/out)
 *    - nrealb
 *    - ndeps
 *    - todo (in/out): number of boundary dependencies, out:+ ghosts and gdeps
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::LocalGhostDepSwapp( const PromGrid *lastG, 
				  PromTable<int> &procTable, const int nrealb, 
				  const int ndeps, int &todo )
{
  MPI_Status      status; 
  const int       mype = comm_.mype(), tag = lastG->getNewTag();
  int             ierr,xx,gid,proc,flag,nadj,id,tt,nlp,bsz,lid,ii,tbsz,geq,ndf;
  PromNode       *ghos, *boun;
  const int       master = masterProc(), endp = getEndProc(), SENDSZ=6;
  const int       lastfactor = lastG->factor_; 
  MPI_Comm        lastcomm = lastG->gComm_; 
  int            *sbuff, *rbuff, *pb; 
  PromGlobalID   *cdep, *ndep;

  // buffer size
  tbsz = bsz = 2 + (SENDSZ*ghosts_.getCount()) + 1 + (2*nrealb) + (2*ndeps); 
  for( xx = master, nlp = 0 ; xx < endp ; xx += lastfactor, nlp++) {
    if( xx == mype ) continue; //
    MPI_Send(&tbsz, 1, MPI_INT, xx/lastfactor, tag, lastcomm );
    MPI_Recv(&flag,1,MPI_INT, xx/lastfactor, tag, lastcomm, &status);
    if( flag > bsz ) bsz = flag;
  }
  ierr = PetscMalloc( bsz * sizeof(int),&sbuff );             CHKERRQ(ierr);
  ierr = PetscMalloc( bsz * sizeof(int),&rbuff );             CHKERRQ(ierr);
  // collect local data, set realID.proc tp master
  // <todo,n,n*<lid,gid,geq,ndf,otherid,realproc>,nb, nb*<lid,nd<lid,proc>>>
  pb = sbuff; *pb++ = todo; *pb++ = ghosts_.getCount();
  for( xx = nLocalNd_, ii = 0; xx < nNodes_ ; xx++ ){ 
    ghos = nodes_[xx];
    if( ghos->symbData_->dependents == NULL ) {
      ghos->symbData_->dependents = new PromList<PromGlobalID*>(3); 
      CHKNEW( ghos->symbData_->dependents, 1 );
    }
    // set realid to master
    PromGlobalID *id = &ghos->symbData_->ghostID;
    id->setProc( masterProc(id->proc()) );
    if( ghos->symbData_->child == NULL || ghos->symbData_->child->isMarked() ){ 
      //if( ghos->symbData_->ghostID.local(ierr) < 0 ){ 
      assert(factor_>lastfactor); continue; // split processor 
    } 
    // <lid,gid,geq,otherid,realproc>
    *pb++ = xx; 
    *pb++ = ghos->index_.global(); 
    *pb++ = ghos->geq();
    *pb++ = ghos->getNDF(); 
    *pb++ = ghos->symbData_->ghostID.local(); 
    *pb++ = ghos->symbData_->ghostID.proc(); 
    ii++;
  }
  assert( ii == ghosts_.getCount()); 
  assert(pb-sbuff == 2 + SENDSZ * ghosts_.getCount());
  // write deps, set to master
  *pb++ = nrealb;
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ){
    boun = nodes_[xx]; assert(boun->getState() != PROM_NOTUSED);
    if( boun->isBound() ) {
      *pb++ = xx; *pb++ = boun->symbData_->dependents->getCount(); 
      Position pos = boun->symbData_->dependents->getHeadPosition(); 
      assert(pos!=NULL);
      while( pos ) { 
	cdep = boun->symbData_->dependents->getNext(pos);
	// set bound deps to master
	cdep->setProc( masterProc( cdep->proc() ) );
	*pb++ = cdep->local(ierr);  *pb++ = cdep->proc(); // no value yet ?
      }
    }
  }
  assert(pb-sbuff == tbsz);
  // send data along
  int sproc = (mype+lastfactor), rproc = (mype-lastfactor); 
  if( sproc >= endp ) sproc = master;
  if( rproc < master ) rproc = endp - lastfactor;
  MPI_Request sreq;
  for( xx = 0 ; xx < nlp-1; xx++ ){ 
    // send/recv<td,ng,ng*<lid,gid,geq,ndf,oid,rproc>,nb,nb*<lid,nd<lid,proc>>>
    MPI_Isend(sbuff, tbsz, MPI_INT, sproc/lastfactor, tag, lastcomm, &sreq);
    MPI_Recv(rbuff, bsz, MPI_INT, rproc/lastfactor, tag, lastcomm, &status);
    MPI_Get_count( &status, MPI_INT, &tbsz ); // tt has size of bounent buffer
    //
    pb = rbuff; 
    tt = *pb++; todo += tt; 
    nadj = *pb++;
    while( nadj-- ) {
      lid = *pb++;gid = *pb++;geq = *pb++;ndf = *pb++;id = *pb++;proc = *pb++;
      assert( lid >= nLocalNd_ && lid < nNodes_ ); assert(!ISLOCAL(proc));
      ghos = nodes_[lid]; assert(ghos != NULL);
      //
      if( ghosts_.add( ghos, gid + 1) ) {
	ghos->index_.setGlobal( gid );
	ghos->setGEQ( geq );
	ierr = ghos->SetNDF( ndf );  CHKERRQ(ierr);
	ghos->symbData_->ghostID.setLocal( id ); 
	ghos->symbData_->ghostID.setProc( proc );
	procTable.add( proc+1 );
      }
    }
    // read deps
    nadj = *pb++; 
    while( nadj-- ) {
      lid = *pb++; flag = *pb++;         assert( lid >= 0 && lid < nLocalNd_ );
      boun = nodes_[lid];
      if( boun->symbData_->dependents == NULL ) {
	boun->symbData_->dependents = new PromList<PromGlobalID*>(3); 
	CHKNEW( boun->symbData_->dependents, 1 );
      }
      while( flag-- ) { 
	lid = *pb++; proc = *pb++; assert( lid < 0 );
	ndep = new PromGlobalID( lid, proc );            CHKNEW( ndep, 1 );
        boun->symbData_->dependents->addTail( ndep );
      }
    }
    assert(pb-rbuff==tbsz);
    // ready for next
    MPI_Wait( &sreq, &status );
    // swap buffers
    pb = rbuff; rbuff = sbuff; sbuff = pb;
  }
  PetscFree(sbuff); PetscFree(rbuff);
  // set all procTable
  int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; 
  pb = sbuf; *pb++ = procTable.getCount();
  TablePos tpos = procTable.getHeadPosition(); 
  while( tpos ) { 
    proc = procTable.getNext( tpos )-1;
    *pb++ = proc;
  }
  bsz = pb-sbuf; assert(bsz<=PROM_BSZ);
  for( xx = master, nlp = 0 ; xx < endp ; xx += lastfactor, nlp++) {
    if( xx == mype ) continue; //
    MPI_Request rreq,sreq; 
    MPI_Irecv( rbuf, PROM_BSZ, MPI_INT, xx/lastfactor, tag, lastcomm, &rreq );
    MPI_Isend( sbuf, bsz, MPI_INT, xx/lastfactor, tag, lastcomm, &sreq );
    MPI_Wait( &rreq, &status ); 
    pb = rbuf; tt = *pb++;
    while( tt-- ) { proc = *pb++; procTable.add( proc+1 ); }
    MPI_Wait( &sreq, &status ); 
  }
  assert( procTable.getCount()*factor_ <= comm_.np() );
  assert( ghosts_.getCount() == nNodes_ - nLocalNd_ );  
  MPI_Iprobe(MPI_ANY_SOURCE,tag,lastcomm,&tt,&status); assert(!tt);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "Resolve_phase2"  
/* PromGrid::Resolve_phase2 ******************************
 *
 *  - set full ghost data structs
 *
 *  INPUT:
 *    - procTable: (in/out)
 *    - todo: number of tag1 to recv
 *    - newghosts: (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::Resolve_phase2( PromTable<int> &procTable, int todo,
			      int *newghosts )
{
  MPI_Status status; 
  const int mype = comm_.mype(),tag1=getNewTag(),gtag=getNewTag();
  int        ierr, xx, newOtherID, newLocID, gid, proc, flag;
  int        nadj, id, receive, todo2,tt;
  PromNode *ghos, *boun, *n1;
  PromTable<int>  procTable2(57);                         assert(isActive());
  const int  master = masterProc(), endp = getEndProc();
  const int  reqsz = sizeof(MPI_Request)/sizeof(int);
  PromList<PromNode*> *list;
  PromList<MPI_Request*> smessages;
  PromTable<PromList<PromNode*>*> proc_list(57);
  PromGlobalID *dep;

  // 2) send new indices and adjacs for ghost to real
  for( xx = nLocalNd_ ; xx < nNodes_ ; xx++ ){
    ghos = nodes_[xx];      
    if( ghos->getState() == PROM_NOTUSED ) continue; 
    if( ghos->symbData_->dependents == NULL ){
      ghos->symbData_->dependents = new PromList<PromGlobalID*>(3); 
      CHKNEW( ghos->symbData_->dependents, 1 );
    }
    assert( ghos->symbData_->dependents->isEmpty() );
    proc = ghos->symbData_->ghostID.proc();            assert(!ISLOCAL(proc));
    //
    if( (list=proc_list.find(proc+1)) == NULL ) {
      list = new PromList<PromNode*>(100); CHKNEW(list,1);
      proc_list.add( list, proc+1 );
    }
    list->addTail(ghos);
  }
  // isend data
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    tt = 3*list->getCount(); assert(tt>0);
    Position pos = list->getHeadPosition(); assert(pos!=NULL);
    while( pos ) {
      ghos = list->getNext(pos); 
      tt += ghos->symbData_->adjac.getCount(); // *gidJ
    }
    MPI_Request *sreq; ierr = PetscMalloc( (tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr);    smessages.addTail( sreq );
    int *buff = (int*)(sreq+1), *pb = buff;
    while( !list->isEmpty() ) {
      ghos = list->removeHead(); 
      *pb++ = ghos->symbData_->ghostID.local(); *pb++ = ghos->index_();
      // send ghost's adjacs
      *pb++ = ghos->symbData_->adjac.getCount(); 
      TablePos tpos = ghos->symbData_->adjac.getHeadPosition(); 
      while( tpos ) {
	PromNode *n1 = ghos->symbData_->adjac.getNext( tpos ); 
	gid = *pb++ = n1->index_.global();
      }
    }
    assert(tt == pb-buff);
    // send(tag1) ghost data *< oldFID, newCID, newCgid, state >
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag1, gComm_, sreq ); 
    delete list; 
  }
  // 2b) receive bound data from deps (ghosts on other procs), new ghosts
  int procstodo = procTable.getCount(), sent_done_mess = 0; todo2 = 0; 
  while( procstodo || !sent_done_mess ) {
    // send(tag1) tell all I am done
    if( !todo && !sent_done_mess ) {
      sent_done_mess = 1; 
      TablePos tpos = procTable.getHeadPosition(); 
      while( tpos ) { 
	proc = procTable.getNext( tpos ) - 1; 
	if( proc != mype ) {
	  MPI_Send( NULL, 0, MPI_INT, proc/factor_, tag1, gComm_ );
	}
      } 
      if(options_.verbose_>2)PetscPrintf(PETSC_COMM_WORLD,
					  "[%d]%s:Send(tag1) to %d procs\n",
					  mype,__FUNCT__,procstodo);
    }
    // recv(tag1) check for done procs
    MPI_Iprobe( MPI_ANY_SOURCE, tag1, gComm_, &receive, &status );
    if( receive ) {  // recive(tag1)
    } 
    // recv(tag1), send(gtag)
    do{ 
      MPI_Iprobe( MPI_ANY_SOURCE, tag1, gComm_, &receive, &status);
      if( receive ) {
	MPI_Get_count(&status, MPI_INT, &tt); 
	if( tt == 0 ) {
	  procstodo--;
	  MPI_Recv(NULL,0,MPI_INT, status.MPI_SOURCE, tag1, gComm_,&status);
	}
	else {
	  int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
	  CHKERRQ(ierr);
	  // recive(tag1) < newlocalid, newOtherID, nadj, nadj*<gid> >
	  MPI_Recv(buff,tt,MPI_INT, status.MPI_SOURCE, tag1, gComm_, &status);
	  while(pb-buff < tt) {
	    assert(todo); todo--;
	    PromContext::Read(newLocID,pb); PromContext::Read(newOtherID,pb);
	    assert( newLocID >= 0 && newLocID < nLocalNd_ );
	    boun = nodes_[newLocID];                 assert(boun->isBound());
	    assert(newOtherID >= 0 );
	    // add adjacs (fine gids)
	    PromContext::Read( nadj, pb ); 
	    while( nadj-- ) {
	      PromContext::Read(gid,pb); id = gid - proc_gnode_[master]; 
	      n1 = NULL;
	      if( id < 0 || id >= nLocalNd_ ) {
		ierr = GetProc( gid, proc );  CHKERRQ(ierr);
		proc = masterProc(proc);
		id = gid - proc_gnode_[proc];
		// can send(gtag) < otherid, locid, myproc >
		todo2 += addGhost( n1, proc, gid, id, gtag );
		if( !procTable.findi(proc+1) && !procTable2.findi(proc+1) ) {
		  procTable2.add(proc+1);
		}
	      }
	      else n1 = nodes_[id];
	      assert(n1!=NULL); assert( n1 != boun );
	      if(boun->symbData_->adjac.add(n1))n1->symbData_->adjac.add(boun);
	    }
	    proc = status.MPI_SOURCE*factor_;
	    // set new data
	    flag = 0;
	    Position pos = boun->symbData_->dependents->getHeadPosition();
	    assert(pos!=NULL);
	    while( pos ) { 
	      dep = boun->symbData_->dependents->getNext(pos); 
	      if( dep->proc() == proc ) {
		assert(dep->local(ierr)==-1); flag++;  
		dep->setLocal( newOtherID ); 
	      }
	    }
	    assert(flag == 1 || newOtherID == -1); 
	  }
	  assert(pb-buff == tt);   PetscFree(buff); 
	}
      }
    }while(receive);
    assert(todo>=0);
  }
  assert(todo == 0);
  while( !procTable2.isEmpty() ) procTable.add( procTable2.removeHead() );
  // free
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }

  // 2c) receive(gtag)
  *newghosts = todo2; // ret
  ierr = ReceiveAddGhost( todo2, gtag, procTable );            CHKERRQ(ierr);
  assert( ghosts_.getCount() == nNodes_ - nLocalNd_ );
  // debug
  MPI_Iprobe(MPI_ANY_SOURCE,tag1,gComm_,&tt,&status); assert(!tt);
  MPI_Iprobe(MPI_ANY_SOURCE,gtag,gComm_,&tt,&status); assert(!tt);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "Resolve_phase3"  
/* PromGrid::Resolve_phase3 ******************************
 *
 *  - union deps and ghosts on clumps - inactive procs 
 *
 *  INPUT:
 *    - procTable:
 *    - newghosts: 
 *
 *  SIDE EFFECTS:
 *    - set ghosts proc to master  
 *
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::Resolve_phase3( const PromGrid *lastG, 
			      const PromTable<int> &procTable, 
			      const int newghosts )
{
  int  ierr, tag3 = lastG->getNewTag();;
  
  assert( procTable.getCount() <= comm_.np()/factor_ ); // only active

  if( isActive() ) {
    ierr = Resolve_phase3_act(lastG, procTable, newghosts, tag3);
    CHKERRQ(ierr);
  }
  else { // inactive
    ierr = Resolve_phase3_ina( lastG, tag3 );            CHKERRQ(ierr);
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "Resolve_phase3_act"  
/* PromGrid::Resolve_phase3_act ******************************
 *
 *  - union deps and ghosts on clumps - active procs 
 *
 *  INPUT:
 *    - procTable:
 *    - newghosts: 
 *    - tag3: MPI tag to use  (any recv)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::Resolve_phase3_act( const PromGrid *lastG, 
				  const PromTable<int> &procTable, 
				  const int newghosts, const int tag3 )
{ 
  MPI_Status  status;  
  const int   mype = comm_.mype(),tag2=getNewTag(),tag1=getNewTag();
  int         xx,gid,proc,*pb,tt,ierr;
  int         ndep, id, ii, jj, lid, kk, nd, *buff; 
  PromNode   *ghos, *boun;
  const int master = masterProc(); assert(master==mype);
  const int reqsz = sizeof(MPI_Request)/sizeof(int);
  PromList<MPI_Request*> smessages; 
  PromTable<PromList<PromNode*>*> proc_list(57);
  PromList<PromNode*> *list; 
  PromGlobalID *dep; Position pos; TablePos tpos;

  // count sends
  for(nd = xx = 0, gid = proc_gnode_[master] ; xx < nLocalNd_ ; xx++, gid++ ){
    boun = nodes_[xx];    assert( boun->getState() != PROM_NOTUSED );   
    if( boun->isBound() ) {
      pos = boun->symbData_->dependents->getHeadPosition(); assert(pos!=NULL);
      while( pos ) {
	dep = boun->symbData_->dependents->getNext( pos ); 
	proc = dep->proc(); assert(proc != mype);
	if( (list=proc_list.find(proc+1)) == NULL ) {
	  list = new PromList<PromNode*>(64); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	} 
	list->addTail( boun );
	nd++;
      }
    }
  }
  const int ndeps = nd;
  // 3) send other's ghosts this adjacs, deps
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &tt ); proc = tt - 1;
    tt = 3 * list->getCount(); // 3
    pos = list->getHeadPosition(); assert(pos!=NULL);
    while( pos ) {
      boun = list->getNext( pos ); 
      tt += 2 * boun->symbData_->dependents->getCount();
      tpos = boun->symbData_->adjac.getHeadPosition(); 
      while( tpos ) {
	PromNode *n1 = boun->symbData_->adjac.getNext( tpos ); 
	if( n1->isExtended() ) tt++;
      }
    }
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr); smessages.addTail( sreq );
    buff = (int*)(sreq+1); pb = buff;
    while( !list->isEmpty() ){
      boun = list->removeHead(); 
      // send< gid, nadj, nadj*<gid>, ndeps, ndeps*<local,proc> >
      *pb++ = boun->index_.global(); 
      int *tpb = pb; *pb++ = -1; 
      tpos = boun->symbData_->adjac.getHeadPosition(); int nadj = 0;
      while( tpos ) {
	PromNode *n1 = boun->symbData_->adjac.getNext( tpos ); 
	if( n1->isExtended() ) { *pb++ = n1->index_.global(); nadj++; }
      }
      assert(*tpb==-1);  *tpb = nadj; //assert(nadj>0);
      *pb++ = boun->symbData_->dependents->getCount(); 
      pos = boun->symbData_->dependents->getHeadPosition(); assert(pos!=NULL);
      while( pos ) {
	dep = boun->symbData_->dependents->getNext( pos ); 
	*pb++ = dep->local(); *pb++ = dep->proc(); 
      }
    }
    assert(tt == pb-buff); delete list;
    // send(tag1) *< gid, nadj, nadj*<gid>, ndeps, ndeps*<local,proc> >
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag1, gComm_, sreq);  
  }

  // receive ghost data
  int todo = nNodes_ - nLocalNd_; kk = 0; 
  while( todo ){
    MPI_Probe( MPI_ANY_SOURCE, tag1, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &jj );  const int sz = jj;
    ierr = PetscMalloc(sz*sizeof(int),&buff); CHKERRQ(ierr);
    // recive(tag1) *< gid, nadj, nadj*<gid>, ndeps, ndeps*<local,proc>>
    MPI_Recv(buff,sz,MPI_INT,status.MPI_SOURCE,tag1,gComm_,&status);  
    pb = buff;
    while( pb-buff < sz ) {
      todo--; assert(todo>=0);
      gid = *pb++; ghos = ghosts_.find(gid+1);
      if( ghos==NULL || ghos->index_.global()!=gid || !ghos->isGhost() ){
	PromContext::printArr(stderr,buff,sz,"\tERROR(1): recv[%d]");// debug
	//SETERRQ4(PETSC_COMM_SELF,1,"gid=%d, ghos=%p, g.gid=%d, tag=%d",gid,ghos,(ghos==NULL) ? -1 : ghos->index_.global(),tag1);
      }
      else {
	// add to table (key is one based)
	int nadj = *pb++; jj = nadj; //assert(nadj > 0);
	while( jj-- ) {
	  gid = *pb++; id = gid - proc_gnode_[master]; // gid
	  PromNode *n1;
	  if( id < 0 || id >= nLocalNd_ ) n1 = ghosts_.find( gid + 1 );
	  else n1 = nodes_[id];  
	  if(n1&&ghos->symbData_->adjac.add(n1)) n1->symbData_->adjac.add(ghos); 
	}
	// read ghosts deps
	assert( ghos->symbData_->dependents != NULL );
	ndep = *pb++; assert( ndep > 0 );
	for( ii = tt = 0 ; ii < ndep ; ii++ ) {
	  // < local, proc > 
	  id = *pb++; proc = *pb++; 
	  if( proc == mype ) { // do ghosts point back to bound
	    proc = status.MPI_SOURCE * factor_; 
	    assert(proc%factor_==0);// point back
	    jj = proc_gnode_[proc];
	    id = ghos->index_.global() - jj; // boundary id
	  }
	  Position pos = ghos->symbData_->dependents->getHeadPosition(); 
	  PromGlobalID *dep = NULL;
	  while( pos && dep==NULL ) {
	    dep = ghos->symbData_->dependents->getNext( pos ); 
	    if( dep->proc() != proc ) dep = NULL;
	  }
	  if( dep == NULL ) {
	    // ref to other ghost
	    dep = new PromGlobalID( id, proc );        CHKNEW( dep, 1 );
	    ghos->symbData_->dependents->addTail( dep ); 
	    assert(proc!=mype);
	    tt++;
	  }
	  else{
	    dep->setLocal( id ); // for splitters 
	  }
	}
	assert( ndep == ghos->symbData_->dependents->getCount() );
	// splitting -- send update of lid to all deps of ghosts
	if( factor_ != lastG->factor_ ) { 
	  Position pos = ghos->symbData_->dependents->getHeadPosition(); 
	  while( pos ) {
	    dep = ghos->symbData_->dependents->getNext( pos );
	    proc = dep->proc();
	    // send(tag2) < gid, lid >
	    if( (list=proc_list.find(proc+1)) == NULL ) {
	      list = new PromList<PromNode*>(64); CHKNEW(list,1);
	      proc_list.add( list, proc+1 );
	    } 
	    list->addTail( ghos );
	    kk++; assert( proc != mype );
	  }
	}
      }
    }
    if( sz != pb-buff ) {
      PromContext::printArr(stderr,buff,pb-buff,sz,"\tERROR:recv[%d]sz=%d");
      SETERRQ3(PETSC_COMM_SELF,1,"sz != pb-buff. %d %d tag=%d",sz,pb-buff,tag1);
    }
    PetscFree( buff ); 
  }
  const int ngdeps = kk;
  // clean up
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  // send
  while( !proc_list.isEmpty() ) {
    assert(factor_ != lastG->factor_);
    list = proc_list.removeHead( &tt ); 
    const int proc = tt - 1;
    tt = 2*list->getCount(); 
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr); smessages.addTail( sreq );
    buff = (int*)(sreq+1); pb = buff;
    while( !list->isEmpty() ){
      ghos = list->removeHead(); 
      *pb++ = ghos->index_.global(); *pb++ = ghos->index_();
    }
    assert(tt == pb-buff);
    // send(tag2) ghost data *< gid, lid >
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag2, gComm_, sreq ); 
    delete list; 
  }
  // splitting -- recv update of lid
  if( factor_ != lastG->factor_ ) { 
    assert( master == mype );
    int torecv = ngdeps - ghosts_.getCount() + ndeps;
    while( torecv ) {
      MPI_Probe( MPI_ANY_SOURCE, tag2, gComm_, &status );
      proc = status.MPI_SOURCE * factor_; 
      MPI_Get_count( &status, MPI_INT, &tt ); assert(tt%2==0); 
      ierr = PetscMalloc(tt*sizeof(int),&buff); CHKERRQ(ierr);
      // recv(tag2) *< gid of node, newlid for dep >
      MPI_Recv(buff, tt, MPI_INT, status.MPI_SOURCE, tag2, gComm_, &status);
      pb = buff;
      while( pb-buff < tt ) {
	gid = *pb++; lid = *pb++;
	id = gid - proc_gnode_[mype];
	if( id < 0 || id >= nLocalNd_ ) {
	  ghos = ghosts_.find( gid + 1 );
	  assert(ghos!=NULL);
	}
	else ghos = nodes_[id];
	assert( ghos->isExtended() );
	Position pos = ghos->symbData_->dependents->getHeadPosition(); 
	PromGlobalID *dep = NULL;
	while( pos && dep == NULL ) {
	  dep = ghos->symbData_->dependents->getNext( pos ); 
	  if( dep->proc() != proc ) dep = NULL;
	}
	assert(dep!=NULL);
	dep->setLocal( lid );
	torecv--; assert(torecv >= 0);
      }
      assert( pb-buff == tt );   PetscFree( buff ); 
    }
    MPI_Iprobe(MPI_ANY_SOURCE,tag2,gComm_,&tt,&status); assert(!tt);
    // send deps back to slaves
    int bsz,*buff; 
    PromGlobalID *dep;
    const int lastfactor = lastG->factor_;
    const int sproc = mype + lastfactor;
    // send dependents to master
    for( tt = xx = 0; xx < nNodes_ ; xx++ ) {
      boun = nodes_[xx];
      if( boun->isExtended() ) {     
	tt += 3; // gid,ndep,nadj
	if( xx >= nNodes_ - newghosts ) {
	  if( tt%(sizeof(double)/sizeof(int)) != 0 ) tt++;
	  boun->GetWrite_size(jj); tt += jj;
	}
	jj = boun->symbData_->dependents->getCount(); 	tt += 2*jj;
	jj = boun->symbData_->adjac.getCount();	        tt += jj;
      }
      else assert( !boun->isGhost() );
    }
    bsz = tt;
    ierr = PetscMalloc((bsz+1)*sizeof(int),&buff); CHKERRQ(ierr);
    // send 
    for( xx = 0, pb = buff; xx < nNodes_ ; xx++ ) {
      boun = nodes_[xx];
      if( boun->isExtended() ) {     
	*pb++ = boun->index_.global(); 
	*pb++ = boun->symbData_->dependents->getCount();  
	*pb++ = boun->symbData_->adjac.getCount();
	if( xx >= nNodes_ - newghosts ) {
	  if( (pb-buff)%(sizeof(double)/sizeof(int)) != 0 ) *pb++ = -3;
	  boun->Write( &pb );
	}
	Position pos = boun->symbData_->dependents->getHeadPosition(); 
	assert(pos!=NULL); 
	while( pos ) {
	  dep = boun->symbData_->dependents->getNext( pos ); 
	  dep->setProc( masterProc(dep->proc()) );
	  *pb++ = dep->local(); *pb++ = dep->proc();
	}
	TablePos tpos = boun->symbData_->adjac.getHeadPosition();
	while( tpos ) {
	  PromNode *n1 = boun->symbData_->adjac.getNext( tpos ); 
	  *pb++ = n1->index_.global();
	}
      }
      else assert( !boun->isGhost() );
    }
    assert(pb-buff==bsz);
    //send(tag3) *<gid,ndep,nadj,[[-3],node],ndep*<local,proc>,nadj*<gid>>
    MPI_Send( buff, bsz, MPI_INT, sproc/lastfactor, tag3, lastG->gComm_ );
    PetscFree(buff);
  }
  // cleanup
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe(MPI_ANY_SOURCE,tag2,gComm_,&tt,&status); assert(!tt);
  MPI_Iprobe(MPI_ANY_SOURCE,tag1,gComm_,&tt,&status); assert(!tt);

  return 0;
}
  
#undef __FUNCT__  
#define __FUNCT__ "Resolve_phase3_ina"  
/* PromGrid::Resolve_phase3_ina ******************************
 *
 *  - union deps and ghosts on clumps - inactive procs 
 *
 *  INPUT:
 *    - tag3:
 *
 *  SIDE EFFECTS:
 *    - set ghosts/dep proc to master  
 *
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::Resolve_phase3_ina( const PromGrid *lastG, const int tag3 )
{  
  const int master=masterProc(),endp=getEndProc();
  const int lastfactor=lastG->factor_;
  const int mype = comm_.mype();
  MPI_Comm  lastcomm = lastG->gComm_;  assert(lastG->isActive());
  MPI_Status status; 
  int        ierr, xx, gid, proc, bsz, id, ii, *pb, nadj, ndep;
  int       *buff=NULL, my0 = proc_gnode_[master]; 
  MPI_Request sreq;
  PromNode   *boun; 
  assert(endp-master > lastfactor); assert(mype != master);
  const int rproc = mype - lastfactor;
  const int sproc = (mype+lastfactor != endp) ? mype+lastfactor : -1;
  
  // clear old deps
  for( xx = 0; xx < nNodes_ ; xx++ ) {
    boun = nodes_[xx];
    if( boun->isExtended() ) {     
      while( !boun->symbData_->dependents->isEmpty() ) {
	delete boun->symbData_->dependents->removeHead();
      }
    }
  }

  // recv new deps
  MPI_Probe( rproc/lastfactor, tag3, lastcomm, &status );
  MPI_Get_count( &status, MPI_INT, &bsz ); 
  if( bsz != 0 ){ierr = PetscMalloc(bsz*sizeof(int),&buff); CHKERRQ(ierr); }
  else buff = NULL;
  // recv(tag3) *<gid,ndep,nadj,[[-3],node],ndep*<local,proc>,nadj*<gid>>
  MPI_Recv( buff, bsz, MPI_INT, rproc/lastfactor, tag3, lastcomm, &status);
  if( sproc >= 0 ) { // send allong
    MPI_Isend( buff, bsz, MPI_INT, sproc/lastfactor, tag3, lastcomm, &sreq );
  }
  pb = buff; 
  while( pb-buff<bsz ) {
    gid = *pb++; ndep = *pb++; nadj = *pb++; assert(ndep>0); 
    assert( ndep > 0 );
    id = gid - proc_gnode_[master];  
    if( id >= 0 && id < nLocalNd_ ) boun = nodes_[id];
    else {
      boun = ghosts_.find(gid+1); 
      if( boun == NULL ){
	id = nNodes_;   // a new ghost
	ierr = AddNNodePointers( 1 ); CHKERRQ(ierr);
	nodes_[id] = boun = newNode( id ); 
	GetProc( gid, proc ); assert(master != proc);
	boun->symbData_->ghostID.setProc( proc );  assert(proc%factor_ == 0);
	boun->symbData_->ghostID.setLocal( gid - proc_gnode_[proc] ); 
	boun->index_.setGlobal( gid );
	ghosts_.add( boun, gid+1 );
	//boun->setNDF(ndf); boun->setGEQ(geq); // hack for necc. data!!!
	if((pb-buff)%(sizeof(double)/sizeof(int))!=0){ii=*pb++;assert(ii==-3);}
	boun->Read( parBank_, &pb );
      }
    }
    // deps
    if( boun->symbData_->dependents == NULL ){
      boun->symbData_->dependents = new PromList<PromGlobalID*>(3);         
      CHKNEW(boun->symbData_->dependents,1);
    }
    for( ii = 0 ; ii < ndep ; ii++ ) {
      id = *pb++; proc = *pb++;    assert(proc>=0 && proc<comm_.np());
      assert(id >= 0);
      PromGlobalID *dep = NULL;
      Position pos = boun->symbData_->dependents->getHeadPosition(); 
      while( pos && dep == NULL ) {
	dep = boun->symbData_->dependents->getNext( pos ); 
	if( dep->proc() != proc ) dep = NULL;
      }
      if( dep == NULL ){ 
	dep = new PromGlobalID( id, proc );                 CHKNEW( dep, 1 );
	boun->symbData_->dependents->addTail( dep ); 
      }
      else {
	assert( dep->local(ierr) < 0 || dep->local() == id );
	dep->setLocal(id);
      }
    }
    assert(ndep == boun->symbData_->dependents->getCount());
    // adjac
    for( ii = 0 ; ii < nadj ; ii++ ) {
      gid = *pb++;  assert( gid != boun->index_.global() );
      PromNode *ghos;
      if( (id=gid-my0) >= 0 && id < nLocalNd_ ) ghos = nodes_[id];
      else ghos = ghosts_.find( gid + 1 );
      if(ghos&&ghos->symbData_->adjac.add(boun))boun->symbData_->adjac.add(ghos);
    }
  }
  assert(pb-buff==bsz); 
  if( sproc >= 0 ) MPI_Wait( &sreq, &status );
  if( buff ) PetscFree(buff);
  MPI_Iprobe(MPI_ANY_SOURCE,tag3,lastcomm,&ii,&status); assert(!ii);
    
  return 0;
}
