//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: geo_coar_adjac.C,v 1.2 2011-01-22 20:13:39 madams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           geo_coar_adjac.C
//
#include "prom_grid.hh"
   
#undef __FUNCT__  
#define __FUNCT__ "MakeExactAdjac"
/* PromGeoGrid::MakeExactAdjac ******************************
 *
 *  - current top grid - make exact adjacs from prev (symbolic mat. tri. prod.)
 *  - ready to coarsen (for next grid), and/or make numeric
 *  - makes full adjacecetcies (w/o regaurd to fixity)
 *
 *  INPUT:
 *  SIDE EFFECTS:
 *    - adds extra ghost nodes
 *
 *  RETURN:
 *     - PETSc error code
 */
int PromGeoGrid::MakeExactAdjac( PromGrid *lastG )
{ 
  int         ierr, xx; assert(this!=NULL);
  PromNode   *curr;

  if( lastG->isActive() ) {
    // remove all adjacs - count num selected
    for( xx = 0 ; xx < nNodes_ ; xx++ ){
      curr = nodes_[xx];
      assert( curr->index_() == xx && curr->getState() != PROM_NOTUSED );
      curr->symbData_->adjac.removeAll();
    }

    // make numeric adjacentcies
    ierr = MakeExactAdjac_private( lastG );                CHKERRQ(ierr);
    
    // diagnostics
    int npar[2],sum[2];
    for( xx = npar[0] = 0 ; xx < nLocalNd_ ; xx++ ){
      npar[0] += nodes_[xx]->symbData_->adjac.getCount() + 1;
    }
    if( !isActive() ) npar[0] = 0;
    // count parents
    for( xx = npar[1] = 0 ; xx < lastG->nLocalNd_ ; xx++ ){
      npar[1] += lastG->nodes_[xx]->nParents();
    }

    MPI_Allreduce( npar, sum, 2, MPI_INT, MPI_SUM, lastG->gComm_ ); 
    assert((sum[0]-nglobal())%2 == 0);

    if( options_.verbose_>1 && comm_.mype()==0 )
      PetscPrintf(PETSC_COMM_SELF,
		  "[%d]%s %d non-zero cells, %d parents tot, %g pars ave\n",
		  comm_.mype(),__FUNCT__,sum[0],sum[1],
		  (float)sum[1]/(float)lastG->nglobal());
  }

  return 0;
}

#define ADJ_MESS_SZ 2
#undef __FUNCT__  
#define __FUNCT__ "MakeExactAdjac_private" 
/* PromGeoGrid::MakeExactAdjac_private ******************************
 *
 *  - set adjacentcies, send off messages, receive nessages
 *
 *  INPUT:
 *
 *  SIDE EFFECTS: 
 *  RETURN:
 *     - PETSc error code
 */
int PromGeoGrid::MakeExactAdjac_private( PromGrid *lastG )
{ 
  int         ierr,proc,ii,todo,locidi,gidj,tt,procj;
  PromNode   *curri, *currj;
  TablePos    tpos;
  PromTable< PromTable<PromTable<int>*>*> proci_table_gidi_tablej(57);
  PromTable<PromTable<int>*> *gidi_tablej;
  PromTable<int> procTable(57);
  const int   lastfactor=lastG->factor_,gtag=getNewTag(); assert(lastG->isActive());
  const int   reqsz = sizeof(MPI_Request)/sizeof(int),tag=lastG->getNewTag();
  PromList<MPI_Request*>  rmessages; const int tag1=lastG->getNewTag();
  MPI_Status  status;  
  int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; 
 
  // collect data to send
  ierr = MakeAdjacData( lastG, gtag, todo, proci_table_gidi_tablej, procTable);
  CHKERRQ( ierr );

  // send (tag) sizes -- global
  for( ii = 0 ; ii < comm_.np()/lastfactor ; ii++ ) sbuf[ii] = 0;
  tpos = proci_table_gidi_tablej.getHeadPosition(); 
  while( tpos ) { 
    gidi_tablej = proci_table_gidi_tablej.getNext(tpos, &proc);
    TablePos tpos2 = gidi_tablej->getHeadPosition();     proc--;    tt = 0; 
    while( tpos2 ) { 
      PromTable<int> *tablej = gidi_tablej->getNext( tpos2 );
      tt += tablej->getCount() * ADJ_MESS_SZ; 
    }
    assert(sbuf[proc/lastfactor] == 0); sbuf[proc/lastfactor] = tt;
  }

  assert(sbuf[comm_.mype()/lastfactor] == 0);
  // global!
  MPI_Alltoall( sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, lastG->gComm_ );
  // post recv(tag)
  for( ii = 0 ; ii < comm_.np()/lastfactor ; ii++ ) {
    if( rbuf[ii] != 0 ) {
      // recv(tag)
      int sz = rbuf[ii], part = ii, *pb; 
      ierr = PetscMalloc((sz+reqsz)*sizeof(int),&pb); CHKERRQ(ierr);
      MPI_Request *rreq = (MPI_Request*)pb; pb += reqsz;
      // irecv(tag) *< oldgid, loclid >
      MPI_Irecv( pb, sz, MPI_INT, part, tag, lastG->gComm_, rreq );
      rmessages.addTail( rreq );
    }
  }

  // send data buffers (tag)
  int **sbuffs = NULL; MPI_Request *sreqs = NULL;
  const int nSendProc = proci_table_gidi_tablej.getCount();
  if( nSendProc ) {
    ierr = PetscMalloc(sizeof(int*)*nSendProc,&sbuffs); CHKERRQ(ierr);
    ierr = PetscMalloc(sizeof(MPI_Request)*nSendProc,&sreqs);  CHKERRQ(ierr);
  }
  ierr = SendAdjacData( tag, proci_table_gidi_tablej, sreqs, sbuffs);
  CHKERRQ(ierr);
  // recv(tag) adjacData
  while( !rmessages.isEmpty() ) {
    MPI_Request *req = rmessages.removeHead(); 
    assert(*req != MPI_REQUEST_NULL);
    // recv(tag) * < oldgid, loclid >
    MPI_Test( req, &ii, &status );
    if( !ii ){ rmessages.addTail( req ); continue; }

    int *buff = (int*)req + reqsz, *pb = buff;    
    MPI_Get_count( &status, MPI_INT, &tt );     assert(tt > 0);
    proc = status.MPI_SOURCE*lastfactor; 
    // recive(tag) n * < locidi, gidj >
    assert(tt%ADJ_MESS_SZ == 0); 
    while( pb-buff < tt ){  
      // recive(0) < locidi, gidj >
      locidi = *pb++;              assert(locidi>=0 && locidi<nLocalNd_);
      curri = nodes_[locidi]; 
      gidj = *pb++;
      ierr = GetProc( gidj, procj ); CHKERRQ(ierr);  assert(procj%factor_==0); 
      int locidj = gidj - proc_gnode_[procj];
      // get other node
      if( procj == comm_.mype() ) { // local adjac from other proc
	assert( locidj >= 0 && locidj < nLocalNd_ );
	currj = nodes_[locidj]; 
      }
      else{ // procj != comm_.mype()
	// search for ghost - can send(1) < otherid, locid, proc >
	assert(procj<comm_.np());
	todo += addGhost( currj, procj, gidj, locidj, gtag );  // start game
	if( procTable.add(procj+1) && options_.verbose_>2 ){ // has recvs now
	  PetscPrintf(PETSC_COMM_SELF,"\t[%d]%s ERROR: 1) NEW PROC %d\n",
		      comm_.mype(),__FUNCT__,procj);
	}
	assert( currj != curri ); 
      } 
      curri->symbData_->adjac.add( currj ); currj->symbData_->adjac.add(curri);
    }
    assert(pb-buff == tt);      PetscFree(req);
  }
  MPI_Iprobe(MPI_ANY_SOURCE,tag,lastG->gComm_,&ii,&status);assert(!ii);

  // collect proctable
  if( isActive() ){
    for(proc=comm_.mype()+lastfactor ; proc < comm_.mype()+factor_; proc += lastfactor){
      MPI_Recv(rbuf,PROM_BSZ,MPI_INT, MPI_ANY_SOURCE,tag1,lastG->gComm_,
	       &status);
      int *pb = rbuf, tt = *pb++;
      while(tt--){
	ii = *pb++; 
	if(ii!=comm_.mype() && procTable.add(ii+1) && options_.verbose_>0){
	  PetscPrintf(PETSC_COMM_SELF,"\t[%d]%s ERROR: 2) NEW PROC %d\n",
		      comm_.mype(),__FUNCT__,ii);  
	}
      }
    }
  }
  else{ //send procs
    int *pb = sbuf;  *pb++ = procTable.getCount();
    TablePos tpos = procTable.getHeadPosition();
    while( tpos ) { 
      int proc = procTable.getNext(tpos) - 1; *pb++ = proc;
    }
    MPI_Request sreq;
    MPI_Isend( sbuf, pb-sbuf, MPI_INT,masterProc(comm_.mype())/lastfactor, tag1,
	       lastG->gComm_, &sreq);
    MPI_Wait( &sreq, &status ); 
  }
  // clean up sends
  if( nSendProc ) {
    MPI_Status *statuses;
    ierr = PetscMalloc(nSendProc*sizeof(MPI_Status),&statuses);  CHKERRQ(ierr);
    MPI_Waitall( nSendProc, sreqs, statuses ); 
    for( ii = 0 ; ii < nSendProc ; ii++ ) {     
      if(sbuffs[ii] != NULL){ PetscFree(sbuffs[ii]); sbuffs[ii]=NULL; }
    }
    PetscFree( statuses ); PetscFree( sreqs );  PetscFree( sbuffs ); 
  }
  // process addGhost() messages
  if( isActive() ) {
    if( options_.verbose_>2 && (comm_.mype()==0) ) {
      PetscPrintf(PETSC_COMM_SELF,"[%d]%s %d new ghos\n",comm_.mype(),__FUNCT__,todo);
    }
    ierr = ReceiveAddGhost( todo, gtag, procTable );        CHKERRQ(ierr);
    MPI_Iprobe(MPI_ANY_SOURCE,gtag,gComm_,&ii,&status);assert(!ii);
  }
  MPI_Iprobe(MPI_ANY_SOURCE,tag,lastG->gComm_,&ii,&status);assert(!ii);
  MPI_Iprobe(MPI_ANY_SOURCE,tag1,lastG->gComm_,&ii,&status);assert(!ii);

  return 0;
}
 
#undef __FUNCT__  
#define __FUNCT__ "MakeAdjacData" 
/* PromGeoGrid::MakeAdjacData ******************************
 *
 *  - make data
 *
 *  INPUT:
 *    - gtag:
 *    - todo: number of new ghost responses to wait for (out)
 *    - proci_table_gidi_tablej: processor->adjac(i,j) to send messages to(out)
 *    - procTable: processors to send messages to, receive ghost from (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */ 
int PromGeoGrid::MakeAdjacData( PromGrid *lastG, const int gtag, int &todo, 
				PromTable<PromTable<PromTable<int>*>*> 
				&proci_table_gidi_tablej, 
				PromTable<int> &procTable )
{
  int       xx,ii,jj,Msz=0;
  const int Isz = 2512;
  PromNode *curri, *currj;

  // add local adjacs to 'adjac' in nodes on this THIS grid with PREV grids nds
  for( xx = 0, todo = 0 ; xx < lastG->nLocalNd_ ; xx++ ){
    const PromNode * const finei = lastG->nodes_[xx];  
    const PromParent * const *parI = finei->parents_->data_;
    const int nparI = finei->parents_->getSize();

    // loop over all the adjacent nodes, others elements, add adjac. here.
    TablePos tpos = finei->symbData_->adjac.getHeadPosition();
    const PromNode * finej = finei;
    do{ 
      const PromParent * const *parJ = finej->parents_->data_; 
      const int nparJ = finej->parents_->getSize();
      // for all coarse nodes (I)
      for( ii = 0 ; ii < nparI ; ii++ ){
	const int proci = parI[ii]->proc(), gidi = parI[ii]->global(); 
	const int locidi = gidi - proc_gnode_[proci];	
	// for all course nodes (J)
	for( jj = 0 ; jj < nparJ ; jj++ ){
	  assert(parJ[jj]->getPsz()==1);
	  const int procj = parJ[jj]->proc(), gidj = parJ[jj]->global();
	  if( gidj == gidi ) continue;
	  const int locidj = gidj - proc_gnode_[procj];
	  // add node(j) to others adjac(i).
	  currj = NULL;
	  if ( proci == comm_.mype() && procj == comm_.mype() ) {
	    curri = nodes_[locidi]; currj = nodes_[locidj];
	    curri->symbData_->adjac.add(currj); currj->symbData_->adjac.add(curri);
	  }
	  else if( proci == comm_.mype() ){ //  procj != comm_.mype()
	    curri = nodes_[locidi]; 
	    // search for ghost - can send(1) < otherid, locid, comm_.mype() >
	    todo += addGhost( currj, procj, gidj, locidj, gtag );
	    curri->symbData_->adjac.add(currj) ; currj->symbData_->adjac.add(curri);
	    procTable.add(procj+1); // make 'procTable'
	  }
	  else { // ( proci != comm_.mype() )
	    PromTable<PromTable<int>*> *gidi_tablej;
	    gidi_tablej = proci_table_gidi_tablej.find(proci+1);
	    if( !gidi_tablej ) {
	      gidi_tablej = new PromTable<PromTable<int>*>(Isz);
	      CHKNEW(gidi_tablej,Isz);
	      proci_table_gidi_tablej.add( gidi_tablej, proci+1 );
	    }
	    PromTable<int> *tablej = gidi_tablej->find( gidi+1 );
	    if( !tablej ) {
	      tablej = new PromTable<int>(163);        CHKNEW(tablej,1);
	      gidi_tablej->add( tablej, gidi+1 );
	    }
	    tablej->add(gidj+1);
	    if( Msz < gidi_tablej->getCount() ) Msz = gidi_tablej->getCount();
	  }
	}	                             // course nodes jj 
      }                                      // course nodes ii
      if( tpos ) finej = finei->symbData_->adjac.getNext(tpos);
      else break;
    }while(1);                               // fine nodes (j)   
  }                                          // fine nodes (i)
  if(options_.verbose_>2 && comm_.mype()==0) PetscPrintf(PETSC_COMM_SELF,
					 "[%d]%s():--DONE (Msz=%d/%d)\n",
					 comm_.mype(),__FUNCT__,Msz,Isz);
  
  return 0; 
}
      
#undef __FUNCT__  
#define __FUNCT__ "SendAdjacData" 
/* PromGeoGrid::SendAdjacData ******************************
 *
 *  - send data
 *
 *  INPUT:
 *    - tag:
 *    - proci_table_gidi_tablej: table of tables of nodes (emptied)
 *    - sreqs: (out)
 *    - sbuffs: (out)
 *   
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */ 
int PromGeoGrid::SendAdjacData( const int tag, 
				PromTable<PromTable<PromTable<int>*>*> 
				&proci_table_gidi_tablej, 
				MPI_Request sreqs[], int *sbuffs[])
{
  int          nsent=0, tt, gidi, proc, ierr, *buff;
  PromTable<PromTable<int>*> *gidi_tablej;
  PromGrid *lastG = prevgrid_; assert(lastG!=NULL);

  if(options_.verbose_>2)PetscPrintf(PETSC_COMM_SELF,"[%d]%s BEGIN\n", comm_.mype(),__FUNCT__);
  // send(tag)
  while( !proci_table_gidi_tablej.isEmpty() ) { 
    gidi_tablej = proci_table_gidi_tablej.removeHead( &proc ); proc--;
    tt = 0; 
    TablePos tpos = gidi_tablej->getHeadPosition();     assert(tpos);
    while( tpos ) { 
      PromTable<int> *tablej = gidi_tablej->getNext( tpos );
      tt += (tablej->getCount() * ADJ_MESS_SZ); 
    }
    ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;  CHKERRQ(ierr);
    while( !gidi_tablej->isEmpty() ) {
      PromTable<int> *tablej = gidi_tablej->removeHead(&gidi); gidi--;
      while( !tablej->isEmpty() ) {					    
	int gidj = tablej->removeHead() - 1;  
	int locidi = gidi - proc_gnode_[masterProc(proc)];
	// send(tag) message proc <locidi, gidj > 
	*pb++ = locidi;      *pb++ = gidj;
      }
      delete tablej;  
    }
    delete gidi_tablej; assert(pb == buff+tt);

    // send(tag) < *<locidi,gidj> >
    MPI_Isend(buff,tt,MPI_INT,proc/lastG->factor_,tag,lastG->gComm_,&sreqs[nsent]);
    sbuffs[nsent++] = buff;
  }

  return 0;
}
