//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: prom_solv.C,v 1.8 2011-10-09 20:24:27 madams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           prom_solve.C
//    
#include "prometheus.hh"

const char * const PromPC_base::PromPCStrings[PROMNUMPC] = {"ILU","LU","Jacobi","ASM","None","MG","Gauss-Seidel","G-S(sym.)", "Nodal ASM"};
const char * const PromSolver::PromKSPStrings[PROMNUMKSP] = {"GMRES","Richardson","BCGS","CR","TFQMR","PC only","CG","Chebychev", "Brezina", "FGMRES", "SYMMLQ", "MINRES"};

/******************************************************************************
 * core MG methods
 *************************************************************************/

static int create_top_is_arr( PromGrid *grid, const bool use_iter_on_top);

#undef __FUNCT__   
#define __FUNCT__ "Create"
/* PromMG::Create***********************************************
 *
 *   set up PETSc multigrid data structures (BEFORE assembly)
 *     -- per configuration
 *
 *   INPUT:
 *     - grid: grid communicator (for dot products)
 *     - npre:
 *     - npost:
 *     - p_map: map for lagrange nultipliers (==NULL)
 *     - use_iter_on_top:
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code 
 */    
int PromMG::Create( PromGrid * const grid, int npre, int npost, 
		    const PromMap *const p_map /*=NULL*/,
		    const bool use_iter_on_top /*=FALSE*/ )
{ 
  assert(this!=NULL);
  A_ = grid->stiffness_; assert(A_ != NULL);
  const PromMap &map = A_->map_; const int np = map.comm_.np();
  int ierr;
  MPI_Comm wcomm = A_->MPIComm(); assert(x_==NULL);
  PetscFunctionBegin;
  assert(x_==NULL && b_==NULL && r_==NULL);
  if( p_map != NULL ) {
    x_ = new PromCRVector( map, p_map );
    b_ = new PromCRVector( map, p_map );
    r_ = new PromCRVector( map, p_map );
  }
  else if( A_->isComplex() ){
    x_ = new PromComplexVector( map );
    b_ = new PromComplexVector( map );
    r_ = new PromComplexVector( map );
  }
  else {
    x_ = new PromVector( map );
    b_ = new PromVector( map );
    r_ = new PromVector( map );
  } 
  interpolate_ = grid->prolongator_;
  restrict_ = grid->prolongator_;
  // helmholtz?
  char str[32], *ksp_str = NULL, *pc_str = NULL; bool helm;
  ierr = grid->options_.GetOptionString("-mg_levels_ksp_type", str, 32, &helm );
  CHKERRQ(ierr);
  if( helm && strcmp( str,"helmholtz") != 0 ) helm = FALSE;
  // set working vectors - needed on all except the coarsest 
  assert(presmooth_ == NULL && postsmooth_ == NULL);
  presmooth_ = new PromSolver( grid->options_, grid->perf_mon_, map, p_map );
  CHKNEW(presmooth_,1);
  if( grid->isTop() ){ // top multigrid
    // make pc
    PromPC *pc = NULL; 
    if( p_map == NULL ) {
      // make solver
      if( !use_iter_on_top ) {
	ierr = presmooth_->CreatePC( PROMPCASM, x_ );         CHKERRQ(ierr);
	ierr = presmooth_->SetKSPType( PROMKSPPREONLY );      CHKERRQ(ierr);
      }
      else { // ksp 
	int kmax = 3*(np/grid->factor_) + 10; if(kmax>120) kmax = 120; // helmholtz heur.!
	if( np/grid->factor_ == 1 ) {
	  kmax = 50;
	  ierr = presmooth_->CreatePC( PROMPCGAUSSSEIDEL, x_ ); CHKERRQ(ierr);
	  kmax = kmax/4 + 1; 
	}
	else {
	  ierr = presmooth_->CreatePC( PROMPCASM, x_ );      CHKERRQ(ierr);
	}
	double rtol = 1.0e-1, atol = 1.e-29, dtol = 1.e2;
	ierr = presmooth_->SetTolerances( kmax, rtol, atol, dtol );    CHKERRQ(ierr);
	if( A_->isIndefinite() || helm || A_->isComplex() ) { // -- symmetric???
	  ierr = presmooth_->SetNEig( kmax ); CHKERRQ(ierr); // called before 'SetKSPType'
	  ierr = presmooth_->SetKSPType( PROMKSPGMRES );       CHKERRQ(ierr);
	}
	else {
	  ierr = presmooth_->SetKSPType( PROMKSPCG );       CHKERRQ(ierr);
	}
	if( helm ) presmooth_->history_.helmholtz = 1;
      }
      // make 1 or 2 blocks 
      if(grid->nBlocks_!=0 || grid->blockISs_!=NULL){
	SETERRQ(PETSC_COMM_SELF,-1,"grid->nBlocks_!=0 || grid->blockISs_!=NULL");
      }
      ierr = create_top_is_arr( grid, use_iter_on_top ); CHKERRQ(ierr);
      if( use_iter_on_top && np/grid->factor_ == 1 ) { 
	PromPCGaussSeidel *gspc;
	gspc = dynamic_cast<PromPCGaussSeidel*>( presmooth_->pc_ ); 
	assert(gspc != NULL);    assert(gspc->pc_ == NULL);
	pc = gspc->pc_ = new PromPC(grid->options_,grid->perf_mon_,wcomm);
	ierr = gspc->SetNumIts(2); CHKERRQ(ierr);
	gspc->symmetric_ = TRUE;
      }
      else pc = dynamic_cast<PromPC*>(presmooth_->pc_);
      //
      ierr = pc->SetType( PROMPCASM, x_ );  CHKERRQ(ierr);
      ierr = SetOperator( A_ ); CHKERRQ(ierr); // restes
      if( use_iter_on_top && np == grid->factor_ ) {
	PromPCGaussSeidel *gspc;
	gspc=dynamic_cast<PromPCGaussSeidel*>(presmooth_->pc_);
	assert(gspc!=NULL);
	// setup g-s data structures
	ierr = gspc->Create( grid->stiffness_,  grid->blockISs_, 
			     grid->isActive() ? grid->nBlocks_ : 0 ); 
	CHKERRQ(ierr); 
      }
      // make the blocks -- does a setup
      ierr = pc->CreateASM( grid->nBlocks_, grid->blockISs_ );  CHKERRQ(ierr);
    }
    else { // kkt -- 'pc' is not used for 'solver' (no iterative for 1 proc)
      ierr = presmooth_->CreatePC( PROMPCNONE, x_ );         CHKERRQ(ierr);
      ierr = presmooth_->SetKSPType( PROMKSPPREONLY );      CHKERRQ(ierr);
      pc = dynamic_cast<PromPC*>(presmooth_->pc_);
      ierr = SetOperator( A_ ); CHKERRQ(ierr); // restes
      const PromCRMatrix*K=dynamic_cast<const PromCRMatrix*>(grid->stiffness_);
      PromCRVector *xx = dynamic_cast<PromCRVector*>(x_); assert(xx!=NULL);
      ierr = pc->CreateKKTSolver( K, K->C_, xx ); CHKERRQ(ierr);
      assert(grid->CRAggs_ == NULL); assert(grid->nCRAggs_ == 0); 
      assert( pc->setup_called_ == 2 ); // skip around
      ierr = pc->kkt_->SetUp(); CHKERRQ(ierr); //keep in same state as rest-noop?
    }
  }
  else{ // non-top grid
    bool kspset;
    PromVector *xx = dynamic_cast<PromVector*>(x_); assert(xx!=NULL);
    if( helm ){
      int level = grid->getLevel(), hfact = 4, ii;
      if( grid->isBase() ) ksp_str = "chebychev";
      else{ ksp_str = "gmres"; pc_str = "nodal"; } 
      // heuristic!
      for(ii=0;ii<level;ii++) npost *= hfact; // hueristic
      if( npost > 64 ) npost = 64;
    }
    else assert(presmooth_->history_.helmholtz == 0);

    // make pc
    ierr = presmooth_->MakePC( grid, grid->nBlocks_, grid->blockISs_, 
			       kspset, xx, "mg_levels_", pc_str );
    CHKERRQ(ierr);
    if( kspset ) ksp_str = "preonly"; // gs

    // make ksp -- allocs vectors for gmres so set max limit
    ierr = presmooth_->SetKSP((npre>npost)?npre:npost,"mg_levels_",ksp_str);
    CHKERRQ(ierr);
    // post smooth is copy 
    postsmooth_ =  new PromSolver( presmooth_ ); CHKNEW( postsmooth_, 1 ); 
    // G-S sets this -- lower where needed for gmres (hack!)
    if( npost > npre ) {
      ierr = presmooth_->SetTolerances( npre ); CHKERRQ(ierr); 
    }
    else if( npre > npost ) {
      ierr = postsmooth_->SetTolerances( npost ); CHKERRQ(ierr); // overwrite!
    }
    // helmholtz -- post smoothing only
    if( helm ) {
      postsmooth_->history_.helmholtz = 1;
      postsmooth_->history_.level = grid->getLevel();
      postsmooth_->history_.isbase = grid->isBase();
      postsmooth_->history_.istop = grid->isTop();
    }
    // kkt data 
    if( p_map != NULL ) {
      const PromCRMatrix*K=dynamic_cast<const PromCRMatrix*>(grid->stiffness_);
      PromPC_base *pc = presmooth_->pc_;
      const PromMatrix *AA = grid->stiffness_; 
      const PromCRVector *xx = dynamic_cast<const PromCRVector*>(x_); 
      assert(xx!=NULL);
      ierr = pc->CreateKKTSmoother( AA, K->C_, xx, grid->nCRAggs_, 
				    grid->CRAggs_, K->CDinvCt_); 
      CHKERRQ(ierr);
      grid->CRAggs_ = NULL;    grid->nCRAggs_ = 0; // takes these 
      assert(pc->setup_called_==2); // skip around
      ierr = pc->kkt_->SetUp(); CHKERRQ(ierr); // keep in same state as rest
      if( kspset ) { // G-S with kkt
	const char *str = pc->kkt_->getPCString();
	if( strcmp(str,"KKT-ILU") == 0 ){ // use cheb for ILU
	  ierr = presmooth_->SetKSPType( PROMKSPCHEB ); CHKERRQ(ierr);
	}
      }
    }

    // setup to be consistant w/ top grid
    assert( presmooth_->pc_->setup_called_ == 2 );
    //ierr = presmooth_->pc_->SetUp(); CHKERRQ(ierr); // noop?
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__ 
#define __FUNCT__ "Create"
/* PromPCMG::Create ****************************************
 *
 *   * 
 *
 *   INPUT:
 *     - grid: used for 'factor' and map, stiffness
 *     - p_map: all-at-once-solver
 *
 *   SIDE EFFECTS:
 *     - sets 'nwork_' and allocs 'work_'
 *
 *   RETURN:
 *     - PETSc error code
 */
int PromPCMG::Create( PromGrid *const grid0, const PromMap *p_map )
{  
  int ierr,cycles,smup,smdn,lev;
  const int np=grid0->comm_.np(),mype=grid0->comm_.mype();
  MPI_Comm wcomm = grid0->comm_();
  char str[128];
  bool dflag, uflag, topit;

  PetscFunctionBegin;
  ierr = options_.GetOptionInt("-pc_mg_cycles",&cycles,&uflag); CHKERRQ(ierr);
  if( !uflag ) cycles = 1; 
  // mg type
  ierr = options_.GetOptionString( "-pc_mg_type", str, 128, &uflag );
  CHKERRQ( ierr ); 
  if( uflag && strcmp(str,"full")==0 ) algo_ = PROMMGFULL;
  else if( uflag && strcmp(str,"additive")==0 ) algo_ = PROMMGADDITIVE;
  else if(uflag && strcmp(str,"multiplicative")==0) algo_=PROMMGMULTIPLICATIVE;
  else{algo_ = PROMMGMULTIPLICATIVE; strcpy(str,"multiplicative");} // default
  ierr=options_.HasOptionName("-prometheus_use_parallel_top_solver",&topit);
  CHKERRQ(ierr);

  // smoother steps
  ierr = options_.GetOptionInt("-pc_mg_smoothup",&smup,&uflag); CHKERRQ(ierr);
  ierr = options_.GetOptionInt("-pc_mg_smoothdown",&smdn,&dflag); CHKERRQ(ierr);
  if( !uflag && dflag ) smup = smdn;
  else if( uflag && !dflag ) smdn = smup;
  else if( !uflag && !dflag ) smdn = smup = 1; // default (1)

  if( options_.verbose_ > 1 )
    PetscPrintf(wcomm,
		"[%d] %s multigrid (%d cycles), %d,%d smoothing iterations, %s top solve\n",
		mype, str, cycles, smdn, smup, topit ? "iterative" : "direct");
  // create ... next grids: A2 = R1 * A1 * R1'
  assert(grid0->CRmap_ == p_map);
  PromMG *mgprev=NULL; assert(mgstack_ == NULL);  PromGrid *grid;
  for( grid = grid0, lev = 0 ; grid != NULL ; lev++, grid = grid->nextgrid_ ){
    // create MG
    PromMG *mg = new PromMG( perf_mon_ );
    if( mgstack_ == NULL ) mgstack_ = mg;
    else { mgprev->next_ = mg; mg->prev_ = mgprev; }
    mgprev = mg;
    mg->factor_ = grid->factor_;
    mg->cycles_ = cycles;

    if( options_.verbose_ > 0 ) { // output  
      PromContext::PrintMatInfo(grid->stiffness_, lev, grid->factor_, mype, np);
      if( options_.matrix_ || options_.verbose_ > 2 ){  
	PetscPrintf(wcomm,"[%d]%s Write R Mat\n",mype,__FUNCT__);
	char str[256],pre[256];
	options_.GetOptionString("-out_path",str,256,&uflag);
	if( uflag ) sprintf( pre, "%s/prom_A_", str );
	else sprintf( pre, "./prom_A_" );
	PromContext::WriteMat( grid->stiffness_, pre, lev );
	if( grid->prolongator_ != NULL ) {
	  if( uflag ) sprintf( pre, "%s/prom_P_", str );
	  else sprintf( pre, "./prom_P_" );
	  PromContext::WriteMat( grid->prolongator_, pre, lev-1 );
	  const PromCRMatrix *pp = dynamic_cast<const PromCRMatrix*>(grid->prolongator_);
	  if( pp != NULL && pp->D_ != NULL ) {
	    if( uflag ) sprintf( pre, "%s/prom_Pc_", str );
	    else sprintf( pre, "./prom_Pc_" );
	    PromContext::WriteMat( pp->D_, pre, lev-1 );
	  }
	}
	const PromCRMatrix *aa = dynamic_cast<const PromCRMatrix*>(grid->stiffness_); 
	if( aa != NULL && aa->C_ != NULL ) {
	  if( uflag ) sprintf( pre, "%s/prom_C_", str );
	  else sprintf( pre, "./prom_C_" );
	  PromContext::WriteMat( aa->C_, pre, lev );
	}
	if( aa != NULL && aa->Ct_ != NULL ) {
	  if( uflag ) sprintf( pre, "%s/prom_Ct_", str );
	  else sprintf( pre, "./prom_Ct_" );
	  PromContext::WriteMat( aa->Ct_, pre, lev );
	}
      }
    }
    // setup PETSc SLES (real work: block factors), hack to test for uzawa
    const PromMatrix *AA = grid->stiffness_; assert(AA != NULL);
    ierr = mg->Create( grid, smdn, smup, grid->CRmap_, topit );
    CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__ 
#define __FUNCT__ "create_top_is_arr"
/* create_top_is_arr ****************************************
 *
 */     
int create_top_is_arr( PromGrid *grid, const bool use_iter_on_top )
{
  int ierr,np=grid->comm_.np();
  if( grid->isActive() ) {
    const int ngsb = (np/grid->factor_ == 1 && use_iter_on_top) ? 2 : 1;
    const int nd = grid->nLocalNd_, my0 = grid->my0Nd();
    const int fact = grid->stiffness_->isComplex() ? 2 : 1;
    grid->nBlocks_ = ngsb;
    ierr = PetscMalloc(ngsb*sizeof(PromIS),&grid->blockISs_);CHKERRQ(ierr);
    int *ipeq,_buf[PROM_BSZ],jj,ii,eq,eq0=grid->my0Eq();
    for( eq = jj = 0 ; jj < nd ; jj++ ) eq += grid->nodes_[jj]->getNDF();
    if(eq*fact>PROM_BSZ){
      ierr = PetscMalloc(sizeof(int)*eq*fact,&ipeq); CHKERRQ(ierr);
    }
    else ipeq = _buf;
    if( ngsb == 1 ) { // multi proc top grid
      for( jj = 0, ii = 0, eq = eq0*fact ; jj < nd ; jj++ ) {
	int ndf = grid->nodes_[jj]->getNDF()*fact;
	for(int kk=0 ; kk < ndf ; kk++, eq++, ii++ ) ipeq[ii] = eq; 
      }
      ierr = PromISCreateGeneral( PETSC_COMM_SELF, ii, ipeq, PETSC_COPY_VALUES,
				  &grid->blockISs_[0]);  CHKERRQ(ierr);
    }
    else if( ngsb == -2 ) { // heavy smoother (simple 1/rest)
      for( jj = ii = 0, eq = grid->nodes_[0]->geq()*fact ; jj < nd-1 ; jj++ ) {
	int ndf = grid->nodes_[jj]->getNDF()*fact;
	for(int kk=0 ; kk < ndf ; kk++, eq++, ii++ ) ipeq[ii] = eq; 
      }
      ierr = PromISCreateGeneral(PETSC_COMM_SELF,ii,ipeq, PETSC_COPY_VALUES,
				 &grid->blockISs_[0]); CHKERRQ(ierr);
      for( jj = 1, ii = 0, eq = grid->nodes_[1]->geq()*fact ; jj < nd ; jj++ ) {
	int ndf = grid->nodes_[jj]->getNDF()*fact;
	for(int kk=0 ; kk < ndf ; kk++, eq++, ii++ ) ipeq[ii] = eq;
      }
      ierr=PromISCreateGeneral(PETSC_COMM_SELF,ii,ipeq, PETSC_COPY_VALUES,
			       &grid->blockISs_[1]); CHKERRQ(ierr);
    }
    else { // even split
      int seed,bi,deq,*ipnd,*lid_leq;
      int _buf2[PROM_BSZ], _buf3[PROM_BSZ];
      for( eq = jj = 0 ; jj < nd ; jj++ ) eq += grid->nodes_[jj]->getNDF();
      const int neq = eq;
      if(nd+1>PROM_BSZ){
	ierr=PetscMalloc(sizeof(int)*nd,&ipnd);CHKERRQ(ierr);
	ierr=PetscMalloc(sizeof(int)*(nd+1),&lid_leq);CHKERRQ(ierr);
      }
      else{ ipnd = _buf2; lid_leq = _buf3; }
      for( jj = 0, lid_leq[0] = 0 ; jj < nd ; jj++ ){
	ipnd[jj] = -1;
	lid_leq[jj+1] = lid_leq[jj] + grid->nodes_[jj]->getNDF();
      }
      PromList<int> nextseeds;
      for( deq = bi = 0, seed = nd/2 ; bi < ngsb ; bi++ ) {
	assert(deq<neq); assert(ipnd[seed] == -1);
	int ndf = grid->nodes_[seed]->getNDF(); 
	eq = ndf; ipnd[seed] = bi; deq += ndf; // number done
	const int limit = (bi == ngsb-1) ? neq : neq/ngsb ;
	while( eq < limit && deq < neq ) {
	  assert(ipnd[seed]!=-1);
	  int ncols,adjacs[PROM_MAXNIEGH];
	  ierr = grid->stiffness_->GetLocalNodeRow(seed,PROM_MAXNIEGH,ncols,
						   adjacs);
	  CHKERRQ(ierr);
	  for( jj = 0 ; jj < ncols && eq < limit && deq < neq ; jj++ ) {
	    int lid = adjacs[jj] - my0;
	    if( lid < 0 || lid >= nd ) continue;
	    if( ipnd[lid] == -1 ) {
	      ndf = grid->nodes_[lid]->getNDF();
	      eq += ndf; ipnd[lid] = bi; deq += ndf; // number done
	      nextseeds.addTail(lid+1);
	    }
	  }
	  assert(ipnd[seed] != -1);
	  if( !nextseeds.isEmpty() ) seed = nextseeds.removeHead()-1;
	  else {
	    for(jj=0;jj<nd;jj++){
	      if(ipnd[jj] == -1){ 
		seed = jj; 
		ndf = grid->nodes_[seed]->getNDF(); 
		eq += ndf; ipnd[seed] = bi; deq += ndf; // number done
		break; 
	      }
	    }
	    assert(ipnd[seed] != -1);
	  }
	  assert(ipnd[seed] != -1);
	}
	// block made, construct it
	for( ii = jj = 0; jj < nd ; jj++ ){
	  if( ipnd[jj] == bi ) {
	    int ndf = grid->nodes_[jj]->getNDF()*fact, leq = lid_leq[jj]*fact;
	    for(int kk=0 ; kk < ndf ; kk++, leq++, ii++ ) ipeq[ii] = leq;
	  }
	}
	assert(ii==eq*fact);
	ierr=PromISCreateGeneral(PETSC_COMM_SELF,ii,ipeq,PETSC_COPY_VALUES,&grid->blockISs_[bi]);
	CHKERRQ(ierr);

	// start over
	nextseeds.removeAll();
	for(jj=0;jj<nd;jj++) if(ipnd[jj] == -1){ seed = jj; break; }
      }
      if(deq != neq) SETERRQ2(PETSC_COMM_SELF,-1,"deq(%d) != neq(%d)",deq,neq);
      if( ipnd != _buf2 ) { PetscFree(ipnd); PetscFree(lid_leq); }
    }
    if( ipeq != _buf ) PetscFree(ipeq);
  }
  else { // inactive
    grid->nBlocks_ = 1;
    ierr = PetscMalloc(sizeof(PromIS),&grid->blockISs_);CHKERRQ(ierr); 
    ierr = PromISCreateGeneral(PETSC_COMM_SELF,0,NULL,PETSC_COPY_VALUES,&grid->blockISs_[0]);
    CHKERRQ(ierr);
  }    
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromMG"
/* PromMG::PromMG *****************************************
 *
 */
PromMG::PromMG( PromPerfMonitor &perf ) :
  presmooth_(NULL), postsmooth_(NULL), b_(NULL), x_(NULL), r_(NULL),
  interpolate_(NULL), restrict_(NULL), A_(NULL), prev_(NULL), next_(NULL),
  factor_(1), perf_mon_(perf), cycles_(1)
{}

#undef __FUNCT__  
#define __FUNCT__ "SetNewPVec"
/* PromMG::SetNewPVec *****************************************
 *
 */
int PromMG::SetNewPVec( const PromMap &map )
{
  int ierr; assert( presmooth_ != NULL);
  ierr = presmooth_->SetNewPVec( map ); CHKERRQ(ierr);
  if( postsmooth_ != NULL ){ // noop!!!
    ierr = postsmooth_->SetNewPVec( map );CHKERRQ(ierr);
  }
  PromCRVector *cr;
  cr = dynamic_cast<PromCRVector*>(b_); assert(cr!=NULL);
  assert(cr->p_ != NULL); delete cr->p_;cr->p_ = new PromVector(map);
  cr = dynamic_cast<PromCRVector*>(r_); assert(cr!=NULL);
  assert(cr->p_ != NULL); delete cr->p_;cr->p_ = new PromVector(map);
  cr = dynamic_cast<PromCRVector*>(x_); assert(cr!=NULL);
  assert(cr->p_ != NULL); delete cr->p_;cr->p_ = new PromVector(map);
 
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "~PromMG"
/* PromMG::~PromMG *****************************************
 *
 */
PromMG::~PromMG()
{
  PromCRVector *cr;

  cr = dynamic_cast<PromCRVector*>(x_); 
  if( cr != NULL ) delete cr; 
  else {
    PromVector *xx = dynamic_cast<PromVector*>(x_); assert(xx!=NULL);
    delete xx;
  }
  
  cr = dynamic_cast<PromCRVector*>(b_); 
  if( cr != NULL ) delete cr; 
  else {
    PromVector *bb = dynamic_cast<PromVector*>(b_); assert(bb!=NULL);
    delete bb;
  }
  cr = dynamic_cast<PromCRVector*>(r_); 
  if( cr != NULL ) delete cr; 
  else {
    PromVector *rr = dynamic_cast<PromVector*>(r_); assert(rr!=NULL);
    delete rr;
  }

  if( presmooth_ != NULL ) delete presmooth_;
  if( postsmooth_ != NULL ) delete postsmooth_;
}

#undef __FUNCT__
#define __FUNCT__ "~PromPCMG"
/* PromPCMG::~PromPCMG ****************************************
 *
 */
PromPCMG::~PromPCMG()
{
  PromMG *mg = mgstack_;
  while( mg != NULL ){
    PromMG *mgx = mg->next_; 
    delete mg; 
    mg = mgx; 
  }
}

/************************************************************************
 * PromSolver
 ************************************************************************/

#undef __FUNCT__  
#define __FUNCT__ "MakePC" 
/* PromSolver::MakePC *************************************
 *
 *  set parameters for level pc -- called once per object
 *  setup 'pc_'
 *
 *  INPUT: 
 *    - grid:
 *    - blockISs: (in/out)
 *    - kspset: did I set the KSP method (out)
 *    - xx: PETSc needs a vector for a PC
 *    - prestring:
 *    - pcname:
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */     
int PromSolver::MakePC( PromGrid * const grid, 
			int &nblocks, PromIS *&blockISs, bool &kspset, 
			const PromVector *const xx, 
			const char* const prestring /*=NULL*/,
			const char* const pcname /*=NULL*/ )
{
  assert(this!=NULL);
  if(!alloced_)SETERRQ(PETSC_COMM_SELF,-1,"called PromSolver::MakePC on aliased");
  const PromMatrix_base * const AA = grid->stiffness_;
  PromPC *levpc = NULL;  MPI_Comm comm = AA->MPIComm();
  int         ii,ierr,mype = AA->map_.comm_.mype();
  bool        flag; char  str[32], opstr[128]="-"; 

  PetscFunctionBegin;
  if( pcname != NULL ) strcpy( str, pcname );
  else {
    if( prestring != NULL ) strcat( opstr, prestring ); // eg mg_levels_
    strcat( opstr, "pc_type" );
    ierr = options_.GetOptionString( opstr, str, 32, &flag ); CHKERRQ(ierr);
    if( !flag ) {
      if( prestring != NULL ) strcpy(str,"nodal_asm");
      else strcpy(str,"jacobi");
      if( options_.verbose_ > 1 ) {
	PetscPrintf(comm,"\t[%d]%s WARNING no '%s' (use default '%s')\n",
		    mype, __FUNCT__, opstr, str );
      }
    }
  }
  bool gs=FALSE,gs_asm=FALSE,asmm=FALSE,none=FALSE,jacobi=FALSE,nd_asm=FALSE;
  bool ilu=FALSE,lu=FALSE;

  if( strcmp(str,"gs")==0 ) gs = TRUE;
  else if( strcmp(str,"gs_asm")==0 ) gs_asm = TRUE;
  else if( strcmp(str,"asm")==0 ) asmm = TRUE;
  else if( strcmp(str,"none")==0 ) none = TRUE;
  else if( strcmp(str,"jacobi")==0 ) jacobi = TRUE;
  else if( strcmp(str,"nodal_asm")==0 || strcmp(str,"nodal")==0) nd_asm = TRUE;
  else if( strcmp(str,"ilu")==0 ) ilu = TRUE;
  else if( strcmp(str,"lu")==0 ) lu = TRUE;
  else {
    PetscPrintf(comm,"\t[%d]%s ERROR Invalid argument to %s:'%s'\n",
		mype,__FUNCT__,opstr,str);
    SETERRQ1(PETSC_COMM_SELF,-1,"Invalid argument to -pc_type: '%s'",str);
  }
  //opstr[1] = '\0'; // back to "-"
  // set PC params -- Create pc_
  kspset = FALSE; // out
  if( none ) {
    ierr = CreatePC( PROMPCNONE, xx );                CHKERRQ(ierr);
  }
  else if( jacobi ) {
    ierr = CreatePC( PROMPCJACOBI, xx );                CHKERRQ(ierr);
  }
  else if( lu ) {
    ierr = CreatePC( PROMPCLU, xx );                CHKERRQ(ierr);
  }
  else if( nd_asm ) {
    ierr = CreatePC( PROMPCNODALASM, xx );                CHKERRQ(ierr);
  }
  else if( gs || gs_asm ) {
    if( prestring != NULL ) { // for smoother only (hack)
      // set GS -- hack ksp 
      ierr = SetKSPType( PROMKSPPREONLY ); CHKERRQ(ierr); 
      kspset = TRUE; // out
    }
    if( gs && blockISs ){ // delete blockISs
      assert(blockISs != NULL);
      while( nblocks-- ){ 
	assert(blockISs[nblocks]!=NULL);
	ierr = PromISDestroy( &blockISs[nblocks] );  assert(ierr==0);
      }
      ierr = PetscFree(blockISs); blockISs=NULL; nblocks=0; CHKERRQ(ierr);
    }
    if( gs && nblocks != 0 ) SETERRQ(PETSC_COMM_SELF,-1,"have blocks and Gauss-Seidel");
    if(gs_asm && nblocks==0) SETERRQ(PETSC_COMM_SELF,-1,"no blocks and blocked Gauss-Seidel");
    ierr = CreatePC( PROMPCGAUSSSEIDEL, xx );           CHKERRQ(ierr);
    PromPCGaussSeidel *gspc = dynamic_cast<PromPCGaussSeidel*>( pc_ ); 
    // blocks
    if( nblocks ) { 
      // manually make a PC for G-S
      assert( gspc->pc_ == NULL && gs_asm ); // set this now
      gspc->pc_ = levpc = new PromPC( options_, perf_mon_, AA->MPIComm() );
      CHKNEW(levpc,1); 
      // create block pc and save for setting
      ierr = levpc->SetType( PROMPCASM, xx );  CHKERRQ(ierr);
    }
    else assert( gspc->pc_ == NULL && gs );
    // Create
    int nGSb = (blockISs != NULL) ? nblocks : 0;
    if( !grid->isActive() ) nGSb = 0;
    //double dmp;
    //ierr = options_.GetOptionDouble("-pc_sor_omega", &dmp, &flag);
    //if( !flag || dmp <= 0.0 || dmp > 2.0 ) dmp = 1.0;
    ierr = gspc->Create( grid->stiffness_,  grid->blockISs_, nGSb ); 
    CHKERRQ(ierr); 
    ierr = options_.HasOptionName( "-pc_symmetric", &flag );
    gspc->symmetric_ = flag;
  }
  else if( asmm ) {
    // create block pc and save for setting
    ierr = CreatePC( PROMPCASM, xx );  CHKERRQ(ierr);
    levpc = dynamic_cast<PromPC*>(pc_); 
    assert(levpc!=NULL); // sp2 has problems with dynamic_cast  
  }
  else if( ilu ) {
    ierr = CreatePC( PROMPCASM, xx );                CHKERRQ(ierr);    
  }
  else SETERRQ(PETSC_COMM_SELF,-1,"No");
  
  ierr = SetOperator( AA ); CHKERRQ(ierr); // new?
  
  if( (gs_asm || asmm) && nblocks == 0 ){
    SETERRQ(PETSC_COMM_SELF,-1,"asm smoother but no blocks (ie, -prometheus_blocks NB)");
  }
  if( !gs_asm && !asmm && nblocks != 0 ){
    assert( blockISs != NULL );
    PetscPrintf(comm,"\t[%d]%s WARNING deleting blocks, %s PC, %d blocks\n",
		mype,__FUNCT__,str,nblocks);
    while( nblocks-- ){ 
      assert(blockISs[nblocks]!=NULL);
      ierr = PromISDestroy( &blockISs[nblocks] );  assert(ierr==0);
    }
    ierr = PetscFree(blockISs); blockISs = NULL; nblocks = 0; CHKERRQ(ierr);
  }
  else if( ilu ) {
    ierr = pc_->CreateASM( 1, NULL );  CHKERRQ(ierr);
    assert( !nblocks );
  }
  // sub pc types 
  if( nblocks ){
    assert(levpc!=NULL);
    ierr = levpc->CreateASM( nblocks, blockISs );    CHKERRQ(ierr);
    if( gs_asm ) { 
      // setup to be consistant w/ ASM
      ierr = pc_->SetUp(); CHKERRQ(ierr);
    }
  }
  else{
    assert( !asmm && !gs_asm );
    ierr = pc_->SetUp(); CHKERRQ(ierr); // post condition: setup is called
  }

  // numits
  ierr = options_.GetOptionInt("-prometheus_pc_steps",&ii,&flag);
  if( flag && ii > 1 ) {
    ierr = pc_->SetNumIts(ii); CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}
 
#undef __FUNCT__
#define __FUNCT__ "CreatePC"
/* PromSolver::CreatePC ****************************************
 *
 */
int PromSolver::CreatePC( PromPCType type, const PromVector_base * const vv )
{
  int ierr;

  PetscFunctionBegin;
  assert(this!=NULL); assert(pc_==NULL);
  if( !alloced_ ) SETERRQ(PETSC_COMM_SELF,-1,"called PromSolver::CreatePC on aliased");
  
  switch( type ) {
  case PROMPCMG : 
    pc_ = new PromPCMG( options_, perf_mon_ ); CHKNEW( pc_, 1 ); 
    break;
  case PROMPCGAUSSSEIDEL : 
    pc_ = new PromPCGaussSeidel( options_, perf_mon_ ); CHKNEW(pc_,1); 
    break; 
  case PROMPCNODALASM : 
    {
      PromPCNodalASM *pc =  new PromPCNodalASM( options_, perf_mon_ ); 
      CHKNEW(pc,1);
      pc_ = pc;
    }
    break;
  case PROMPCLU :
  case PROMPCILU : 
  case PROMPCJACOBI :
  case PROMPCASM : 
  case PROMPCNONE :
    pc_ = new PromPC( options_, perf_mon_, vv->map_.comm_() ); // low level PCs
    CHKNEW(pc_,1);
    ierr = pc_->SetType( type, vv ); CHKERRQ(ierr);
    if(type == PROMPCLU)
      {
	PromPC *pc = dynamic_cast<PromPC*>( pc_ ); assert(pc); 
	ierr = pc->SetFactor(); // this could be missing!!	
      }
    break;
  default : SETERRQ1(PETSC_COMM_SELF,-1,"ERROR: %d type",type);
  }
  PetscFunctionReturn(0);
}
 
#undef __FUNCT__
#define __FUNCT__ "SetKSP"
/* PromSolver::SetKSP ****************************************
 *
 *  Create Iterative method - called after MakePC
 *
 */
int PromSolver::SetKSP(const int maxit,
		       const char* const prestring /*=NULL*/,
		       const char* const ksp_string /*=NULL */)
{
  assert(this != NULL); 
  if( !alloced_ ) SETERRQ(PETSC_COMM_SELF,-1,"called SetKSP::MakePC on aliased");
  int ierr,mype=map_.comm_.mype();
  char str[32],opstr[128]="-"; bool flag;
  MPI_Comm comm = map_.comm_();
  PetscFunctionBegin;

  // get ksp type strint
  if( ksp_string != NULL ) strcpy( str, ksp_string);
  else {
    if( prestring != NULL ) strcat( opstr, prestring ); // ie, mg_levels_
    strcat( opstr,"ksp_type");
    ierr = options_.GetOptionString(opstr,str,32,&flag);CHKERRQ(ierr);
    if( !flag ) {
      if( prestring == NULL ) strcpy(str,"cg"); 
      else strcpy(str,"chebychev"); 
      if( options_.verbose_>1 )
	PetscPrintf(comm,"\t[%d]%s WARNING no '%s' (default to '%s')\n",
		    mype,__FUNCT__,opstr,str);
    }
    opstr[1] = '\0'; // back to "-"
  }

  // get tolerances
  double rtol, atol, dtol;
  if( prestring != NULL ) { // ie, mg_levels_
    dtol = 1.e6; atol = 1.e-27; rtol = 1.e-2; // smoother
  }
  else {
    strcat( opstr,"ksp_rtol");
    ierr = options_.GetOptionDouble( opstr, &rtol, &flag); CHKERRQ(ierr);
    if( !flag ) rtol = -1.;
    opstr[1] = '\0'; // back to "-"
    //
    strcat( opstr,"ksp_atol");
    ierr = options_.GetOptionDouble( opstr, &atol, &flag); CHKERRQ(ierr);
    if( !flag ) atol = -1.;
    opstr[1] = '\0'; // back to "-"
    //
    strcat( opstr,"ksp_dtol");
    ierr = options_.GetOptionDouble( opstr, &dtol, &flag); CHKERRQ(ierr);
    if( !flag ) dtol = -1.;
    opstr[1] = '\0'; // back to "-"
  }

  // set tols
  if( maxit <= 0 )SETERRQ1(PETSC_COMM_SELF,-1,"maxit<=0 (%d)",maxit);
  ierr = SetTolerances( maxit, rtol, atol, dtol ); CHKERRQ(ierr);
  
  if( strcmp(str,"richardson") == 0 ){
    ierr = SetKSPType( PROMKSPRICHARDSON );        CHKERRQ(ierr);
  }
  else if( strcmp(str,"cg") == 0 ){
    ierr = SetKSPType( PROMKSPCG );              CHKERRQ(ierr);
  }
  else if( strcmp(str,"bcgs") == 0 ){
    ierr = SetKSPType( PROMKSPBCGS );              CHKERRQ(ierr);
  }
  else if( strcmp(str,"tfqmr") == 0 ){
    ierr = SetKSPType( PROMKSPTFQMR );             CHKERRQ(ierr);
  }
  else if( strcmp(str,"symmlq") == 0 ){
    ierr = SetKSPType( PROMKSPSYMMLQ );            CHKERRQ(ierr);
  }
  else if( strcmp(str,"minres") == 0 ){
    ierr = SetKSPType( PROMKSPMINRES );            CHKERRQ(ierr);
  }
  else if( strcmp(str,"cr") == 0 ){
    ierr = SetKSPType( PROMKSPCR );                CHKERRQ(ierr);
  }
  else if( strcmp(str,"preonly") == 0 ){
    ierr = SetKSPType( PROMKSPPREONLY );           CHKERRQ(ierr);
  }
  else if( strcmp(str,"chebychev") == 0 ){
    assert( prestring != NULL); // ie, mg_levels_
    ierr = SetKSPType( PROMKSPCHEB );              CHKERRQ(ierr);
  } 
  else if( strcmp(str,"brezina") == 0 ){
    ierr = SetKSPType( PROMKSPBREZ );              CHKERRQ(ierr);
  }
  else if( strcmp(str,"gmres")!=0 && strcmp(str,"fgmres")!=0 ){ // default
    ierr = SetKSPType( PROMKSPCG );              CHKERRQ(ierr);
  }
  
  // cg and gmres/fgmres
  if( strcmp(str,"gmres") == 0 || strcmp(str,"fgmres") == 0 ){
    assert(nEig_ == 0);
    int kmax = -1;
    ierr = options_.GetOptionInt("-ksp_gmres_restart",&kmax,&flag);
    CHKERRQ(ierr);
    if( !flag || kmax < 1 ) kmax = 50;
    if( kmax > maxit ) kmax = maxit;
    ierr = SetNEig( kmax ); CHKERRQ(ierr); // called before 'SetKSPType'
    // must set gmres restart before SetKSPType (allocs vectors)
    if( strcmp(str,"gmres") == 0 ) ierr = SetKSPType( PROMKSPGMRES );
    else ierr = SetKSPType( PROMKSPFGMRES );

    CHKERRQ(ierr);
  }

  if( prestring == NULL && strcmp(str,"cg") == 0 ) { // hack to avoid CV/subsp in smoothers
    // ksp -- compute Singular Values
    ierr = options_.HasOptionName("-prometheus_nocond",&flag);
    CHKERRQ(ierr);
    if( !flag ) { ierr = SetComputeSVD(); CHKERRQ(ierr); }
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SetUp"
/* PromSolver::SetUp ****************************************
 *
 *   KSP setup (eg, vectors)
 */
int PromSolver::SetUp()
{ 
  assert(this!=NULL); int ierr;
  PetscFunctionBegin;
  if( !alloced_ ) {
    PetscFunctionReturn(0);
  }
  //
  if(A_==NULL) SETERRQ(PETSC_COMM_SELF,-1,"A_ == NULL");
  ierr = pc_->SetUp();    CHKERRQ(ierr);

  // complete vectors
  for( int ii = 0 ; ii < nwork_ ; ii++ ) {
    if( work_[ii] == NULL ) {
      // factory
      if(p_map_!=NULL) work_[ii] = new PromCRVector(map_,p_map_ );
      else if( A_->isComplex() ) {
	work_[ii] = new PromComplexVector( map_ );
      }
      else work_[ii] = new PromVector( map_ );
    }
  }
  
  // history data and GMRES Hess...
  if(origin_.e==NULL && (getKSPType()==PROMKSPGMRES||getKSPType()==PROMKSPFGMRES)){
    if( nEig_  <= 0 ) SETERRQ(PETSC_COMM_SELF,-1,"nEig_  < = 0");
    // alloc required work space for GMRES and CG 
    int max_k         = nEig_;
    int hh            = (max_k + 2) * (max_k + 1);
    int hes           = (max_k + 1) * (max_k + 1);
    int rs            = (max_k + 2);
    int cc            = (max_k + 1);
    int size          = (hh + hes + rs + 2*cc + 1);
    if( A_->isComplex() ) {
#if defined(PROM_HAVE_COMPLEX)
      origin_.c = new PromComplex[size];      
      hes_.c = origin_.c + hh;   // hes_
      d_.c  = hes_.c + hes;      // rs_
      ee_.c  = d_.c + rs;        // cc_
      dd_.c = ee_.c + cc;        // ss_
#if defined(PETSC_USE_BOPT_g)
      for(int ii=0;ii<size-1;ii++) origin_.c[ii] = PromComplex(PROM_BIG_NUM,PROM_BIG_NUM);
#endif
      dd_.c[cc-1] = dd_.c[cc] = -13.0; 
      assert(origin_.c[(nEig_+2)*(nEig_+1)+(nEig_+1)*(nEig_+1)+(nEig_+2)+2*(nEig_+1)].real()==-13.0);
#endif
    }
    else{
      ierr = PetscMalloc(size*sizeof(double),&origin_.e); CHKERRQ(ierr); // hh_
      ierr = PetscMemzero(origin_.e,size*sizeof(double)); CHKERRQ(ierr);
      hes_.e = origin_.e + hh;   // hes_
      d_.e  = hes_.e + hes;      // rs_
      ee_.e  = d_.e + rs;        // cc_
      dd_.e  = ee_.e + cc;       // ss_
#if defined(PETSC_USE_BOPT_g)
      for( int ii = 0 ; ii < size-1 ; ii++ ) origin_.e[ii] = PROM_BIG_NUM;
#endif
      dd_.e[cc-1] = dd_.e[cc] = -13.0; 
      assert(origin_.e[hh + hes + rs + 2*cc] == -13.0);
    }
  }

  if( cg_restart_ < 0 ) { // check rho
    cg_restart_ = -cg_restart_;
    PromKSPType ty = alloced_ ? getKSPType() :  alias_->getKSPType();
    if( ty==PROMKSPCHEB || ty==PROMKSPBREZ ){ 
      const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
      const bool noRho = Amat->getDRho() < 1.;
      const bool isScalar = Amat->map_.isConstNDF() && Amat->map_.constNDFSize()==1;
      const bool samePC = pc_->getType()==PROMPCNODALASM || (isScalar && pc_->getType()==PROMPCJACOBI); 
      if(  noRho || !samePC ) {
	// need to compute rho(DA) here -- ignoring dirty_data_!!
	ierr = Amat->ComputeLambda(options_,perf_mon_,10,pc_); CHKERRQ(ierr);
      }
      //
      if( Amat->isIndefinite() && history_.helmholtz ){
	int itfact = 4;
	ierr = SetTolerances(maxits_*itfact, rtol_,atol_,dtol_); CHKERRQ(ierr);
	if( alias_ != NULL ) {
	  ierr = alias_->SetTolerances(alias_->maxits_*itfact,alias_->rtol_, 
				       alias_->atol_, alias_->dtol_);
	  CHKERRQ(ierr);
	}
	// use CR or GMRES?
	if( 1 == 1 ) { 
	  if( alloced_ ) ierr = SetKSPType( PROMKSPCR );
	  else ierr = alias_->SetKSPType( PROMKSPCR );   CHKERRQ(ierr);
	  if(verbose_>0)PetscPrintf(PETSC_COMM_WORLD,"\tIndefinite operator - switch to CR\n");
	}
	else {
	  int mits = (maxits_>alias_->maxits_) ? maxits_ : alias_->maxits_; assert(mits>0);
	  if( alloced_ ) {
	    ierr = SetNEig( mits );               CHKERRQ(ierr);
	    ierr = SetKSPType( PROMKSPGMRES );    CHKERRQ(ierr);
	    ierr = PromSolver::SetUp(); CHKERRQ(ierr); // recursive call to me (GMRES)
	  }
	  else {
	    assert(alias_ != NULL);
	    ierr = alias_->SetNEig( mits );             CHKERRQ(ierr);
	    ierr = alias_->SetKSPType( PROMKSPGMRES );  CHKERRQ(ierr);
	    ierr = alias_->PromSolver::SetUp(); CHKERRQ(ierr);
	  }
	  if(verbose_>0)PetscPrintf(PETSC_COMM_WORLD,"\tIndefinite operator - switch to GMRES\n");	  
	}
      }
    }
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "PrintInfo"
/* PromSolver::PrintInfo *****************************************
 *
 *   - Print info of the solver on this level
 *
 *   INPUT:
 *     level: level
 *
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromSolver::PrintInfo( const int nactivep, const int lev /*=-1*/ ) const
{
  assert(this!=NULL);
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const pc = alloced_ ? pc_ : alias_->pc_;
  int ierr, tt, mype = Amat->map_.comm_.mype();
  const char*const kspname = PromKSPStrings[alloced_?getKSPType():alias_->getKSPType()];
  const char * const pcname = pc->getPCString();
  const char * const subpcname = pc->getSubPCString();

  MPI_Comm comm = Amat->map_.comm_.MPIComm();
  PetscPrintf(comm,"[%d] ",mype);
  if(lev == -1 && pc->getType() == PROMPCMG) PetscPrintf(comm,"Base:    ");
  else if( lev != -1 ) PetscPrintf(comm,"Level %d: ",lev);
  PetscPrintf(comm,"%4d %s, %s:pc(%s.%s), ",nactivep,
	      (nactivep == 1) ? "processor " : "processors",
	      (lev != -1) ? "Smoother" : "  Solver",
	      pcname, subpcname);
  if( lev != -1 || pc->getType() != PROMPCMG ) { // non-accel.
    const int nblocks = pc->getNumBlocks(), nstp = pc->getNumIts();
    PetscPrintf(comm," (%d %s), %d blocks, ",nstp, 
		(nstp==1) ? "step " : "steps", nblocks );
  }
  if( alias_ != NULL && (tt=alias_->maxits_) != maxits_ ) { // this is post!
    PetscPrintf(comm,"ksp(%s, %d pre, %d post, max.it.)\n",kspname,tt,maxits_);
  }
  else {
    PetscPrintf(comm,"ksp(%s,%d max.it.)\n", kspname, maxits_ );
  }
  ierr = pc->PrintLevelInfo();    CHKERRQ(ierr); // collective?

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "SetTolerances"
/* PromSolver::SetTolerances *****************************************
 *
 *   INPUT:
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromSolver::SetTolerances(const int maxits, double rtol /*=-1.0*/,
			      double atol /*=-1.*/, double dtol /*=-1.0*/ )
{
  assert(this!=NULL); // !alloc
  PetscFunctionBegin;
  if( rtol == -1. ) rtol_ = 1.e-6;
  else rtol_ = rtol;
  if( dtol == -1. ) dtol_ = 1.e6; 
  else dtol_ = dtol;
  if( atol == -1. ) atol_ = 1.e-28;
  else atol_ = atol;
  if(maxits<0){SETERRQ1(PETSC_COMM_SELF,-1,"Error 'maxits'=%d < 0",maxits);}
  else if( maxits > 32000 ) maxits_ = 32000;
  else maxits_ = maxits;
  PetscFunctionReturn(0);
}
 
#undef __FUNCT__  
#define __FUNCT__ "SetOperator"
/* PromSolver::SetOperator *****************************************
 *
 *   INPUT:
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromSolver::SetOperator( const PromMatrix_base * const A )
{
  assert(this!=NULL); int ierr;

  if( !alloced_ ) {
    ierr = alias_->SetOperator( A ); CHKERRQ(ierr);
  }
  else {
    if( pc_ == NULL ) SETERRQ(PETSC_COMM_SELF,-1,"pc_ == NULL: call MakePC 1st");
    ierr = pc_->SetOperator( A ); CHKERRQ(ierr); 
    ierr = SetOperator_private( A ); CHKERRQ(ierr);     
  }
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "SetNewOperator"
/* PromSolver::SetNewOperator *****************************************
 *
 *   INPUT:
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromSolver::SetNewOperator( const PromMatrix_base * const A )
{
  assert(this!=NULL); int ierr;

  if( !alloced_ ) {
    ierr = alias_->SetNewOperator( A ); CHKERRQ(ierr);
  }
  else {
    if( pc_ == NULL ) SETERRQ(PETSC_COMM_SELF,-1,"pc_ == NULL: call MakePC 1st");
    ierr = pc_->SetNewOperator( A ); CHKERRQ(ierr); 
    ierr = SetOperator_private( A ); CHKERRQ(ierr); 
  }
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "SetOperator_private"
/* PromSolver::SetOperator_private *****************************************
 *
 *   INPUT:
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromSolver::SetOperator_private( const PromMatrix_base * const A )
{
  // so that next solve call will call setup
  A_ = A;
  if(work_ != NULL && work_[0] != NULL && A->map_ != work_[0]->map_)SETERRQ(PETSC_COMM_SELF,-1,"A->map_ != map_");
  if( cg_restart_ > 0 ) cg_restart_ = -cg_restart_; // hack flag for dirty
  return 0;
} 

#undef __FUNCT__
#define __FUNCT__ "Solve"
/* PromSolver::Solve ****************************************
 * 
 */
int PromSolver::Solve( const PromVector_base * const BB,
		       PromVector_base * const XX,
		       int *its, const bool zerox /*=TRUE*/, 
		       const bool avoidnorms /*=FALSE*/ )
{
  assert(this!=NULL); // !alloce
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  int ierr; 
  
  PetscFunctionBegin;
  if( maxits_ == 0 ) {
    if(zerox) XX->Set(0.0);
    PetscFunctionReturn(0);
  }
  ierr = SetUp(); CHKERRQ(ierr); // does nothing if setup called (should do n.)

  // solve
  switch( alloced_ ? getKSPType() : alias_->getKSPType() ) {
  case PROMKSPGMRES : ierr = SolveGMRES(BB,XX,its,zerox,avoidnorms,PROM_RIGHT);
    CHKERRQ(ierr);   break;
  case PROMKSPFGMRES :
    ierr = SolveGMRES(BB,XX,its,zerox,avoidnorms,PROM_FLEX); CHKERRQ(ierr);
    break;
  case PROMKSPTFQMR :ierr = SolveTFQMR(BB,XX,its,zerox,avoidnorms,PROM_LEFT);
    CHKERRQ(ierr);   break;
  case PROMKSPSYMMLQ :ierr = SolveSYMMLQ(BB,XX,its,zerox,avoidnorms,PROM_LEFT);
    CHKERRQ(ierr);   break;
  case PROMKSPMINRES :
    if( Amat->isComplex() ) ierr = cSolveMINRES(BB,XX,its,zerox,avoidnorms);
    else ierr = SolveMINRES(BB,XX,its,zerox,avoidnorms);
    CHKERRQ(ierr);   break;
  case PROMKSPRICHARDSON : 
    ierr = SolveRichardson( BB, XX, its, zerox, avoidnorms ); CHKERRQ(ierr); 
    break;
  case PROMKSPBCGS : 
    if( Amat->isComplex() ) ierr = cSolveBCGS(BB,XX,its,zerox,avoidnorms);
    else ierr = SolveBCGS(BB,XX,its,zerox,avoidnorms); 
    CHKERRQ(ierr); break;
  case PROMKSPCR : 
    if( Amat->isComplex() ) ierr = cSolveCR(BB,XX,its,zerox,avoidnorms);
    else  ierr = SolveCR(BB,XX,its,zerox,avoidnorms);
    CHKERRQ(ierr);     break;
  case PROMKSPCG : 
    if( Amat->isComplex() ) ierr = cSolveCG(BB,XX,its,zerox,avoidnorms);
    else ierr = SolveCG( BB, XX, its, zerox, avoidnorms ); 
    if( ierr < 0 ) ierr = 0; // cg return number of idefinates
    CHKERRQ(ierr);  
    break;
  case PROMKSPCHEB : 
    if( Amat->isComplex() ) ierr = cSolveCheb(BB, XX, its, zerox, avoidnorms);
    else  ierr = SolveCheb(BB, XX, its, zerox, avoidnorms); 
    CHKERRQ(ierr);  break; 
  case PROMKSPBREZ : ierr = SolveBrez( BB, XX, its, zerox, avoidnorms ); 
    CHKERRQ(ierr);  break;
  case PROMKSPPREONLY : // used for G-S
    {
      PromPC_base * const pc = alloced_ ? pc_ : alias_->pc_;
      ierr = pc->SetNumIts( maxits_ );  CHKERRQ(ierr);       
      ierr = pc->Apply( BB, XX, zerox ); *its = 1; CHKERRQ(ierr);
      ierr = pc->SetNumIts( 1 );  CHKERRQ(ierr); 
    }
    break;
  default : SETERRQ1(PETSC_COMM_SELF,-1,"ERROR: %d type",getKSPType());
  }
  PetscFunctionReturn(ierr); // cg has number of indefs
}

#undef __FUNCT__  
#define __FUNCT__ "~PromSolver"
/* PromSolver::~PromSolver *****************************************
 *
 */
PromSolver::~PromSolver()
{
  assert(this!=NULL);

  if( alloced_ ) {
    if( nwork_ ) {
      for( int i = 0; i < nwork_ ; i++ ) {
	assert(work_[i]!=NULL);
	PromCRVector *cr = dynamic_cast<PromCRVector*>(work_[i]); 
	if( cr != NULL ) delete cr; 
	else {
	  PromVector *vec = dynamic_cast<PromVector*>(work_[i]); 
	  assert(vec!=NULL);
	  delete vec; 
	}
      }
      PetscFree(work_);
    }
    nwork_ = 0; 
    if( origin_.e != NULL ) {
      if(getKSPType()==PROMKSPGMRES||getKSPType()==PROMKSPFGMRES){
	assert(dd_.e[nEig_+1] == -13.0);
      }
      int ierr = PetscFree( origin_.e );  assert(ierr==0);
      origin_.e = d_.e = dd_.e = ee_.e = NULL; 
      hes_.e = NULL;
    }
#if defined(PROM_HAVE_COMPLEX)
    else if(origin_.c != NULL) { delete [] origin_.c; assert(origin_.e==NULL); }
#endif

    if( pc_ != NULL ) delete pc_;  pc_ = NULL; 
  }
  if( projector_ ) delete projector_;

}

#undef __FUNCT__  
#define __FUNCT__ "SetKSPType"
/* PromSolver::SetKSPType *****************************************
 *
 *   * Set the KSP type, allocs work vectors, needs SetOperator called 1st
 *
 *   INPUT:
 *   SIDE EFFECTS:
 *     - sets 'nwork_' and allocs 'work_'
 *
 *   RETURN:
 *     - PETSc error code
 */ 
int PromSolver::SetKSPType( PromKSPType type )
{
  assert( this != NULL );
  int ierr;

  PetscFunctionBegin;  
  if( !alloced_ ){ SETERRQ(PETSC_COMM_SELF,-1,"SetKSPType called on aliased smoother"); }
  if( origin_.e != NULL ) { // old cg or gmres
    assert(A_!=NULL);
    if((getKSPType()==PROMKSPGMRES||getKSPType()==PROMKSPFGMRES) && 
       dd_.e[nEig_+1] != -13.0){
      SETERRQ2(PETSC_COMM_SELF,-1,"dd_.e[%d+1]=%e != -13.0",nEig_,dd_.e[nEig_+1]);
    }
    ierr = PetscFree( origin_.e );  CHKERRQ(ierr); 
    origin_.e = d_.e = ee_.e = dd_.e = NULL;
  }
#if defined(PROM_HAVE_COMPLEX)
  else if( origin_.c != NULL ){
    delete [] origin_.c;
    origin_.c = d_.c = ee_.c = dd_.c = NULL;
  }
#endif

#define VEC_OFFSET 3
  // work vectors
  int ntodo = 0;
  switch( type ) {
  case PROMKSPGMRES :
    if(nEig_ <= 0) SETERRQ(PETSC_COMM_SELF,-1,"nEig needs to be set 1st");
    ntodo = (nEig_+1) + VEC_OFFSET; 
    break;
  case PROMKSPFGMRES :
    if(nEig_ <= 0) SETERRQ(PETSC_COMM_SELF,-1,"nEig needs to be set 1st");
    ntodo = 2 * (nEig_+1) + VEC_OFFSET; 
    break;
  case PROMKSPRICHARDSON : ntodo = 2; break;
  case PROMKSPTFQMR : ntodo = 10; break;
  case PROMKSPSYMMLQ : ntodo = 8; break; 
  case PROMKSPMINRES : ntodo = 9; break; 
  case PROMKSPBCGS : ntodo = 7; break;
  case PROMKSPCR : ntodo = 6; break;
  case PROMKSPPREONLY : ntodo = 0; break;
  case PROMKSPCG : ntodo = 3; break;
  case PROMKSPBREZ : // fallthrough 
  case PROMKSPCHEB : ntodo = 3; break; 
  default : SETERRQ1(PETSC_COMM_SELF,-1,"ERROR: %d type",type );
  }
  if( nwork_ < ntodo ) {
    PromVector_base **work_arr;
    ierr = PetscMalloc( ntodo*sizeof(PromVector_base*), &work_arr ); 
    CHKERRQ(ierr);
    int ii=0;
    if( nwork_ > 0 ) {
      do work_arr[ii] = work_[ii]; while(++ii < nwork_);
      PetscFree(work_);
    }
    work_ = work_arr; nwork_ = ntodo;
    if( A_ != NULL ) {
      do{
	// factory
	if( p_map_ != NULL) work_[ii] = new PromCRVector( map_, p_map_ );
	else if( A_->isComplex() ) {
	  work_[ii] = new PromComplexVector( map_ );
	}
	else work_[ii] = new PromVector( map_ );
      }while( ++ii < nwork_ );
    }
    else {
      do{ work_[ii] = NULL; }while( ++ii < nwork_ );
    }
  }
  ksp_type_ = type;

  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "SetNewPVec"
/* PromSolver::SetNewPVec *****************************************
 *
 *   INPUT:
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code
 */
int PromSolver::SetNewPVec( const PromMap &map )
{
  if( alloced_ ) {
    // ksp vectors
    for( int ii = 0 ; ii < nwork_ ; ii++) {
      PromVector_base *vec = work_[ii]; 
      PromCRVector *cr = dynamic_cast<PromCRVector*>(vec); 
      assert(cr != NULL);
      if( cr->p_ != NULL ) delete cr->p_;
      cr->p_ = new PromVector( map ); 
    }
    //int ierr = pc_->SetNewPVec( map );  CHKERRQ(ierr);
  }
  return 0;
}

/************************************************************************
 * PC
 ************************************************************************/

#undef __FUNCT__
#define __FUNCT__ "SetUp"
/* PromPC_base::SetUp ****************************************
 * 
 */     
int PromPC_base::SetUp() 
{
  assert( this != NULL );
  PetscFunctionBegin;
  if( setup_called_ < 2 ) {
    int ierr = SetUp_private();  CHKERRQ(ierr);
    if( kkt_ != NULL ){ 
      assert( KKT_ != NULL );
      SETERRQ(PETSC_COMM_SELF,-1,"should not be here ???");
      //ierr = kkt_->SetUp(); CHKERRQ(ierr);
    }
    setup_called_ = 2;
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "Apply"
/* PromPC_base::Apply ****************************************
 *
 *  - apply operator
 * 
 *  INPUT: 
 *    - zeroguess: used only for G-S smoothers only
 */     
int PromPC_base::Apply(const PromVector_base * const BB,
		       PromVector_base*const XX,const bool zeroguess /*=TRUE*/)
{
  assert( this != NULL ); int ierr; 
  ierr = SetUp(); CHKERRQ(ierr); 
  if( zeroguess ){ ierr = XX->Set(0.0); CHKERRQ(ierr); }
  PromCRVector *crxx = dynamic_cast<PromCRVector*>(XX);
  if( kkt_ != NULL && crxx != NULL ){ // compute lambda is inconsistant - ok
    assert(/*crxx != NULL &&*/ crxx->p_ != NULL);
    const PromCRVector *crbb = dynamic_cast<const PromCRVector*>(BB);
    assert(crbb!=NULL);  assert(KKT_ != NULL);
    perf_mon_.EventBegin(PromContext::prom_logs_[LOG_AUX1]);
    ierr = kkt_->Apply( this, KKT_, crbb, crxx, zeroguess); CHKERRQ(ierr);
    perf_mon_.EventEnd(PromContext::prom_logs_[LOG_AUX1]);
  } 
  else{
    ierr = Apply_private( BB, XX, zeroguess );  CHKERRQ(ierr);
  }

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "~PromPC_base"
/* PromPC_base::~PromPC_base ****************************************
 */     
PromPC_base::~PromPC_base( )
{
  if( kkt_ != NULL ) delete kkt_; 
}

#undef __FUNCT__
#define __FUNCT__ "SetOperator"
/* PromPC_base::SetOperator ****************************************
 *
 *  - set new operator
 * 
 *  INPUT: 
 *    - :
 */     
int PromPC_base::SetOperator( const PromMatrix_base * const A ) 
{
  assert( this != NULL ); int ierr;

  if( setup_called_ == 2 ) setup_called_ = 1; 
  ierr = SetOperator_private( A ); CHKERRQ(ierr);
  const PromCRMatrix *AA = dynamic_cast<const PromCRMatrix*>(A); 
  KKT_ = AA;

  if( kkt_ != NULL ) {
    assert(AA!=NULL);
    ierr = kkt_->SetOperator( AA );  CHKERRQ(ierr);
  }

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "SetNewOperator"
/* PromPC_base::SetNewOperator ****************************************
 *
 *  - set new operator
 * 
 *  INPUT: 
 *    - :
 */     
int PromPC_base::SetNewOperator( const PromMatrix_base * const A ) 
{
  assert( this != NULL ); int ierr;
  
  setup_called_ = 0; 
  ierr = SetNewOperator_private(A); CHKERRQ(ierr);
  const PromCRMatrix *AA = dynamic_cast<const PromCRMatrix*>(A); 
  if( AA != NULL ) KKT_ = AA;

  if( kkt_ != NULL ) {
    assert(AA!=NULL);
    ierr = kkt_->SetNewOperator( AA );             CHKERRQ(ierr);
  }

  return 0;
}

#undef __FUNCT__   
#define __FUNCT__ "FCycle"
/* PromPCMG::FCycle ****************************************************
 *
 */
int PromPCMG::FCycle( PromMG * const mg0, const bool zerox /*=TRUE*/ )
{ 
  int    ierr, ii, lev, logbase = 1; 
  PromMG *mgc, *mg;

  PetscFunctionBegin; 
  /* restrict the RHS through all levels to coarsest. */
  for( mg = mg0, ii = logbase ; mg->next_ != NULL ; mg = mg->next_, ii++ ) {
#if !defined(PROM_NO_STAGES)
    ierr = perf_mon_.StagePush(PromContext::prom_stages_[ii]); CHKERRQ(ierr);
#endif
    mgc = mg->next_; 
    ierr = mgc->restrict_->MultTranspose( mg->b_, mgc->b_ ); CHKERRQ(ierr);
#if !defined(PROM_NO_STAGES)
    ierr = perf_mon_.StagePop(); CHKERRQ(ierr);               // end log  
#endif
  }
  /* work our way up through the levels */
  mgc = mg; assert(mgc->next_ == NULL && mgc->prev_ != NULL); // top grid
  for( mgc=mg, lev=0 ; mgc->prev_ != NULL ; mgc=mgc->prev_, ii--, lev++ ){
    ierr = MCycle( mgc, ii, (lev == 0) ? TRUE : FALSE );       CHKERRQ(ierr);
#if !defined(PROM_NO_STAGES)
    ierr = perf_mon_.StagePush(PromContext::prom_stages_[ii]); CHKERRQ(ierr);
#endif
    ierr = mgc->interpolate_->Mult( mgc->x_, mgc->prev_->x_);  CHKERRQ(ierr); 
#if !defined(PROM_NO_STAGES)
    ierr = perf_mon_.StagePop(); CHKERRQ(ierr);               // end log  
#endif
  }
  assert( ii == logbase );

  ierr = MCycle( mg0, logbase, FALSE );                   CHKERRQ(ierr);
  PetscFunctionReturn(0);
} 

#undef __FUNCT__
#define __FUNCT__ "PrintLevelInfo"
/* PromPCMG::PrintLevelInfo ****************************************
 *
 */
int PromPCMG::PrintLevelInfo() const
{
  PromMG *mg = mgstack_;  int lev,ierr,na;
  for( lev=0; mg != NULL ; lev++, mg = mg->next_){
    na = mg->A_->map_.comm_.np() / mg->factor_;
    if( mg->postsmooth_ != NULL ) {
      ierr = mg->postsmooth_->PrintInfo( na, lev ); CHKERRQ(ierr);
    }
    else {
      ierr = mg->presmooth_->PrintInfo( na, lev ); CHKERRQ(ierr);
    }
  }
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "MCycle"
/* PromPCMG::MCycle **************************************************
 * 
 */  
int PromPCMG::MCycle( PromMG * const mg, int level, const bool zerox /*=TRUE*/)
{
  int ierr, its, cycles = mg->cycles_;

  PetscFunctionBegin;
  if( mg->next_ == NULL ) {  // top solve
#if !defined(PROM_NO_STAGES)
    ierr=perf_mon_.StagePush(PromContext::prom_stages_[level]);CHKERRQ(ierr);
#endif
    perf_mon_.EventBegin(PromContext::prom_logs_[LOG_ASSEMBLE]);
    assert(zerox);
    ierr = mg->presmooth_->Solve( mg->b_, mg->x_, &its, TRUE, TRUE ); 
    if( ierr < 0 ){ CHKERRQ(ierr); }
    else if( ierr > 0 ) {
      if( options_.verbose_ > 0 ) {
	PetscPrintf(PETSC_COMM_WORLD,
		    "[0]%s WARNING: switch coarse grid solve to GMRES\n",
		    __FUNCT__);
      }
      assert(mg->presmooth_->getKSPType() == PROMKSPCG);
      ierr = mg->presmooth_->A_->SetIndefinite();     CHKERRQ(ierr);
      ierr = mg->presmooth_->SetNEig( mg->presmooth_->maxits_ ); CHKERRQ(ierr); // called before 'SetKSPType'
      ierr = mg->presmooth_->SetKSPType( PROMKSPGMRES );      CHKERRQ(ierr);
      ierr = mg->presmooth_->Solve(mg->b_, mg->x_, &its, TRUE, TRUE );
    }
    perf_mon_.EventEnd(PromContext::prom_logs_[LOG_ASSEMBLE]);
#if !defined(PROM_NO_STAGES)
    ierr = perf_mon_.StagePop();  CHKERRQ(ierr);                // end clog 
#endif
  }
  else{
    PromMG *mgc = mg->next_; 
#if !defined(PROM_NO_STAGES)
    ierr=perf_mon_.StagePush(PromContext::prom_stages_[level]);CHKERRQ(ierr);
#endif
    if( mg->presmooth_->maxits_ > 0 || !zerox ) {
      if( mg->presmooth_->maxits_ > 0 ) {
	ierr = mg->presmooth_->Solve(mg->b_,mg->x_,&its,zerox,TRUE); 
	if(ierr<0) CHKERRQ(ierr);
      }
      else {
	ierr = mg->x_->Set( 0.0 );    CHKERRQ(ierr);  
      }
      ierr = mg->A_->Mult( mg->x_, mg->r_ ); /* b - Ax */        CHKERRQ(ierr);
      ierr = mg->r_->AYPX( -1.0, mg->b_ );                       CHKERRQ(ierr);
    }
    else {
      ierr = mg->r_->Copy( mg->b_ );                       CHKERRQ(ierr);
      ierr = mg->x_->Set( 0.0 );    CHKERRQ(ierr);  
    }
    // r' <- R r
    ierr = mgc->restrict_->MultTranspose( mg->r_, mgc->b_ );    CHKERRQ(ierr);
#if !defined(PROM_NO_STAGES)
    ierr = perf_mon_.StagePop();                 CHKERRQ(ierr);// end log
#endif
    for( int i = 0 ; i < cycles ; i++ ) {
      ierr = MCycle( mgc, level+1, (i==0) );   CHKERRQ(ierr);
      if( mgc->next_==NULL ) break;//coarse grid just once!
    }
#if !defined(PROM_NO_STAGES)
    ierr=perf_mon_.StagePush(PromContext::prom_stages_[level]);CHKERRQ(ierr);
#endif
    // x <- x + P x'
    ierr = mgc->interpolate_->MultAdd(mgc->x_,mg->x_,mg->x_);  CHKERRQ(ierr);
    ierr = mg->postsmooth_->Solve(mg->b_, mg->x_, &its, FALSE, TRUE);
    if(ierr<0) CHKERRQ(ierr);
#if !defined(PROM_NO_STAGES)
    ierr = perf_mon_.StagePop();                 CHKERRQ(ierr);// end log
#endif
  }

  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "ACycle"
/* PromPCMG::ACycle *****************************************************
 *
 */
int PromPCMG::ACycle( PromMG * const mg0, const bool zerox /*=TRUE*/ )
{
  int    ii, its, ierr;
  PromMG *mgc, *mg;

  /* compute RHS on each level */
  for( mg = mg0 ; mg->next_ != NULL ; mg = mgc ) {
    mgc = mg->next_;
    ierr = mgc->restrict_->MultTranspose( mg->b_, mgc->b_ ); CHKERRQ(ierr);
  }
  /* solve seperately on each level */
  for( mg = mg0, ii = 1 ; mg != NULL ; mg = mg->next_, ii++ ) {
#if !defined(PROM_NO_STAGES)
    ierr = perf_mon_.StagePush(PromContext::prom_stages_[ii]); CHKERRQ(ierr);
#endif
    ierr = mg->presmooth_->Solve( mg->b_, mg->x_, &its, zerox, TRUE );
    if(ierr<0) CHKERRQ(ierr);
#if !defined(PROM_NO_STAGES)
    ierr = perf_mon_.StagePop();  CHKERRQ(ierr);// log 
#endif
    mgc = mg; // stash
  }
  assert(mgc!=NULL && mgc->prev_!=NULL && mgc->next_==NULL); // mgc is top
  do{
    mg = mgc->prev_;
    ierr = mgc->interpolate_->MultAdd( mgc->x_, mg->x_, mg->x_);CHKERRQ(ierr);
    mgc = mg;
  }while( mgc->prev_ != NULL );
  return 0;
}


#undef __FUNCT__
#define __FUNCT__ "Apply_private"
/* PromPCMG::Apply_private ****************************************
 *
 */     
int PromPCMG::Apply_private( const PromVector_base * const bb, 
			     PromVector_base * const xx,
			     const bool zeroguess /*=TRUE*/ )
{  
  int ierr; assert(zeroguess);

  PetscFunctionBegin;
  ierr = mgstack_->b_->Copy(bb); CHKERRQ(ierr); 

  if( algo_ == PROMMGMULTIPLICATIVE ) {
#if defined(PROM_USE_VAMPIRE) 
    static int vt_count = 0;  
    if(vt_count++ == 2 && mgstack_->A_->map_.comm_.np() <= 16) VT_traceon();
#endif
    ierr = MCycle( mgstack_, 1 ); CHKERRQ(ierr);
#if defined(PROM_USE_VAMPIRE) 
    VT_traceoff();
#endif
  } 
  else if( algo_ == PROMMGADDITIVE ) {
    ierr = ACycle( mgstack_, TRUE ); CHKERRQ(ierr);
  }
  else {
    assert(algo_ == PROMMGFULL);
    ierr = FCycle( mgstack_, TRUE );          CHKERRQ(ierr);
  }
  ierr = xx->Copy(mgstack_->x_); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "Create"
/* PromPCNodalASM::Create ****************************************
 *
 *  - alloc 'lid_ndof_' and 'invBlocks_', set 'lid_ndof_' and 'nLocNds_'
 *
 */
int PromPCNodalASM::Create()
{
  assert(A_ != NULL);
  int ierr,ndf,nnz,nloc=A_->map_.nlocal(),fact=A_->isComplex() ? 2 : 1; 

  if( A_->map_.isConstNDF() ){ 
    ndf = fact*A_->map_.constNDFSize(); nnz = nloc * ndf * ndf;
  }
  else {
    int xx; const PromMap &map = A_->map_;
    for(xx=nnz=0;xx<nloc;xx++){ ndf = fact*map.locID_ndf(xx); nnz += ndf*ndf; }
  }
  if( invBlocks_ != NULL ) SETERRQ(PETSC_COMM_SELF,-1,"invBlocks_ != NULL");
  ierr = PetscMalloc((nnz+1)*sizeof(double), &invBlocks_); CHKERRQ(ierr);
  invBlocks_[0] = 0.0; invBlocks_[nnz] = -22.0; 
  if( nnz > 0 ) invBlocks_[nnz-1] = -22.0; 

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "SetUp_private"
/* PromPCNodalASM::SetUp_private ****************************************
 *
 *  - create inverted blocks
 *
 */
int PromPCNodalASM::SetUp_private()
{
  if( A_ == NULL ) SETERRQ(PETSC_COMM_SELF,-1,"A_ == NULL"); 
  const PromMap &map = A_->map_;
  const int nloc=map.nlocal(),fact=A_->isComplex() ? 2:1;
  int ierr,xx,jj,done,ncols,adjacs[PROM_MAXNIEGH],igid;
  double *pv; double *Ai[PROM_MAXNIEGH];

#if defined(PROM_TRACE_PC_SETUP)
  PetscPrintf(map.comm_(),"\t[%d]%s this=%p\n",map.comm_.mype(),__FUNCT__,this);
#endif
  
  if( invBlocks_ == NULL ) { ierr = Create(); CHKERRQ(ierr); }
  // set up for timings 
  const int nndf = map.isConstNDF() ? fact*map.constNDFSize() : -1;
  for( xx = 0, pv = invBlocks_, igid = map.my0Nd() ; 
       xx < nloc ; xx++, igid++ ) {
    ierr = A_->GetLocalNodeRow(xx, PROM_MAXNIEGH, ncols, adjacs, Ai); 
    CHKERRQ(ierr);
    const int ndof = (nndf < 0) ? map.locID_ndf(xx)*fact : nndf;
    for( jj = 0, done = 0 ; jj < ncols ; jj++ ) {
      int jgid = adjacs[jj];
      if( igid == jgid ){
	const double * const aijT = Ai[jj];
#if defined(PETSC_USE_BOPT_g)
	if(ndof/fact>1 && aijT[0]!=0.0 && fabs(aijT[fact]-aijT[fact*ndof])/aijT[0] > 1.e-6 && fabs(aijT[0]-1.0)>1.e-8){
	  PetscPrintf(PETSC_COMM_SELF,"[%d]%s symmetric?? row %d, ndf %d, a(2,1)=%e a(1,2)=%e error=%e a(1,1)=%e ndof=%d\n",map.comm_.mype(),__FUNCT__,
		      igid,ndof,aijT[fact],aijT[fact*ndof],
		      fabs(aijT[fact]-aijT[fact*ndof])/aijT[0],aijT[0],ndof);
	}
#endif
       	if( PromContext::TInvert( aijT, pv, ndof ) ) {
	  PromContext::printArr(stderr,aijT,ndof*ndof,map.my0Nd()+xx,"pivot A^-1[%d] gid=%d",ndof);
	}
	done++;
	break;
      }
    }
    if(done!=1)SETERRQ1(PETSC_COMM_SELF,-1,"matrix error - diag not found %d",igid);
    pv += ndof*ndof;
  }
  perf_mon_.Flops( fact*2*(pv-invBlocks_) );
  assert(invBlocks_[pv-invBlocks_] == -22);
  if( pv-invBlocks_ > 0 ) assert(invBlocks_[pv-invBlocks_-1] != -22);

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Apply_private"
/* PromPCNodalASM::Apply_private ****************************************
 *
 */     
int PromPCNodalASM::Apply_private( const PromVector_base * const BB,
				   PromVector_base * const XX,
				   const bool zeroguess /*=TRUE*/ )
{
  int ierr,ii=0; assert(numits_ > 0);
  PromVector*xx = dynamic_cast<PromVector*>(XX); assert(xx!=NULL);
  const PromVector *bb = dynamic_cast<const PromVector*>(BB);assert(bb!=NULL);
  if( zeroguess ){ 
    ierr = Apply_one( bb, xx );    CHKERRQ(ierr);
    ii++;
  }
  while( ii++ < numits_ ) {
    if( work_ == NULL ) {
      assert(r_==NULL); 
      if( !A_->isComplex() ) {
	work_ = new PromVector( xx ); 
	r_ = new PromVector( bb );
      }
      else {
	PromComplexVector *x = dynamic_cast<PromComplexVector*>(XX);
	assert(x!=NULL);
	work_ = new PromComplexVector( x ); 
	r_ = new PromComplexVector( x );
      }
    }
    ierr = A_->PromMatrix::Mult( xx, r_ ); CHKERRQ(ierr); /*r <- b - A x'*/ 
    ierr = r_->PromVector::AYPX( -1.0, bb );  CHKERRQ(ierr);
    ierr = Apply_one( r_, work_ ); CHKERRQ(ierr);          /*t <- B * r */
    ierr = xx->PromVector::AXPY(1.0,work_); CHKERRQ(ierr);/*x' <- x' + t */ 
  }
  
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Apply_one"
/* PromPCNodalASM::Apply_one************************************
 *
 *  - apply: XX := M * BB
 *
 */     
int PromPCNodalASM::Apply_one( const PromVector * const BB,
			       PromVector * const XX )
{
  int ierr, eq, xx; const PromMap &map = A_->map_;
  const int nloc = map.nlocal(), fact=BB->isComplex() ? 2 : 1;
  const int nndf = map.isConstNDF() ? map.constNDFSize() : -1;
  double tmp,*barr,*xarr,*pv, buff[2*PROM_MAXNDF],p0,p1,p2,p3,p4,p5;

  ierr = BB->GetArray( &barr );               CHKERRQ(ierr); // !kkt
  ierr = XX->GetArray( &xarr );               CHKERRQ(ierr); // !kkt
  for( xx = 0, pv = invBlocks_, eq = 0 ; xx < nloc ; xx++ ){
    const int ndof = fact*((nndf < 0) ? map.locID_ndf(xx) : nndf);
    switch(ndof) { 
    case 1: buff[0] = barr[0] * pv[0]; break;
    case 3:
      tmp = barr[0]; p0 = pv[0]*tmp; p1 = pv[3]*tmp; p2 = pv[6]*tmp; 
      tmp = barr[1];p0 += pv[1]*tmp;p1 += pv[4]*tmp;p2 += pv[7]*tmp; 
      tmp = barr[2];p0 += pv[2]*tmp;p1 += pv[5]*tmp;p2 += pv[8]*tmp; 
      buff[0] = p0; buff[1] = p1; buff[2] = p2; break;
    case 6:
      tmp = barr[0];
      p0 = pv[0]*tmp;   p1 = pv[6]*tmp;   p2 = pv[12]*tmp;
      p3 = pv[18]*tmp;  p4 = pv[24]*tmp;  p5 = pv[30]*tmp;
      tmp = barr[1];
      p0 += pv[1]*tmp;  p1 += pv[7]*tmp;  p2 += pv[13]*tmp;
      p3 += pv[19]*tmp; p4 += pv[25]*tmp; p5 += pv[31]*tmp;
      tmp = barr[2]; 
      p0 += pv[2]*tmp;  p1 += pv[8]*tmp;  p2 += pv[14]*tmp; 
      p3 += pv[20]*tmp; p4 += pv[26]*tmp; p5 += pv[32]*tmp;
      tmp = barr[3]; 
      p0 += pv[3]*tmp;  p1 += pv[9]*tmp;  p2 += pv[15]*tmp; 
      p3 += pv[21]*tmp; p4 += pv[27]*tmp; p5 += pv[33]*tmp;
      tmp = barr[4]; 
      p0 += pv[4]*tmp;  p1 += pv[10]*tmp; p2 += pv[16]*tmp; 
      p3 += pv[22]*tmp; p4 += pv[28]*tmp; p5 += pv[34]*tmp;
      tmp = barr[5]; 
      p0 += pv[5]*tmp;  p1 += pv[11]*tmp; p2 += pv[17]*tmp; 
      p3 += pv[23]*tmp; p4 += pv[29]*tmp; p5 += pv[35]*tmp;
      buff[0] = p0; buff[1] = p1; buff[2] = p2; 
      buff[3] = p3; buff[4] = p4; buff[5] = p5;	break;
    default:
      for( int ii = 0, kk = 0 ; ii < ndof ; ii++ ) {
	tmp = 0.0; 
	for( int jj = 0 ; jj < ndof ; jj++, kk++ ) tmp += pv[kk] * barr[jj];
	buff[ii] = tmp; 
      }
      break;
    }
    for( int ii = 0 ; ii < ndof ; ii++ ) xarr[eq++] = buff[ii]; // set!
    barr += ndof;  pv += ndof*ndof;
  }
  ierr = BB->RestoreArray( &barr );     CHKERRQ(ierr); 
  ierr = XX->RestoreArray( &xarr );     CHKERRQ(ierr); 

  return 0;
}


#undef __FUNCT__
#define __FUNCT__ "SetUp_private"
/* PromPCMG::SetUp_private ****************************************
 *
 */
int PromPCMG::SetUp_private()
{
  int ierr;
  
  PetscFunctionBegin;
#if defined(PROM_TRACE_PC_SETUP)
  PetscPrintf(mgstack_->A_->map_.comm_(),"\t[%d]%s this=%p\n",
	      mgstack_->A_->map_.comm_.mype(),__FUNCT__,this);
#endif
  // set up for timings 
  for( PromMG *mg = mgstack_ ; mg != NULL ; mg = mg->next_ ){
    // setup PETSc SLES with smoother...
    if(mg->presmooth_!=NULL){ 
      ierr = mg->presmooth_->SetUp(); CHKERRQ(ierr);
    }
    if( mg->postsmooth_ != NULL ){
      ierr = mg->postsmooth_->SetUp(); CHKERRQ(ierr);
    }
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "SetNewPVec"
/* PromPC_base::SetNewPVec *****************************************
 *
 *   INPUT:
 *   SIDE EFFECTS:
 *   RETURN:
 *     - PETSc error code

int PromPC_base::SetNewPVec( const PromMap &map )
{
  if( work_ != NULL ) {
    // work_
    PromVector_base *vec = work_; 
    PromCRVector *cr = dynamic_cast<PromCRVector*>(vec); assert(cr!=NULL);
    if( cr->p_ != NULL ) delete cr->p_;
    cr->p_ = new PromVector( map ); 
    }
    if( r_ != NULL ) {
    // r_
    PromVector_base *vec = r_; 
    PromCRVector *cr = dynamic_cast<PromCRVector*>(vec); assert(cr!=NULL);
    if( cr->p_ != NULL ) delete cr->p_;
    cr->p_ = new PromVector( map ); 
    }
  if( kkt_ != NULL ) return kkt_->SetNewPVec( map );
  
  return 0;
}*/

//////////////////////////////////////////////////////////////////////////
// KSP methods
//////////////////////////////////////////////////////////////////////////

#undef __FUNCT__  
#define __FUNCT__ "TestSymmMA"
/* PromSolver::TestSymmMA ***************************************************
 *
 */
int PromSolver::TestSymmMA( const PromMatrix_base * const Amat,
			    PromPC_base * const pc, PromVector_base *const Z, 
			    PromVector_base *const P,
			    PromVector_base *const R ) const
{ 
  // set Z and P with ran -- test symm.
  PromComplexVector *cZ = dynamic_cast<PromComplexVector*>(Z);
  PromComplexVector *cP = dynamic_cast<PromComplexVector*>(P);
  int ierr;
  double ra, emin = (double)RAND_MAX/2.0, aplpha = 1. / emin;
  for(int ii=0,tt=Amat->map_.my0Eq(),lim=R->getn();ii<lim;ii++,tt++){
    ra = ( (double)rand() - emin ) * aplpha;
    ierr = Z->SetValues( 1, &tt, &ra, INSERT_VALUES ); CHKERRQ(ierr);
    ra = ( (double)rand() - emin ) * aplpha;
    ierr = P->SetValues( 1, &tt, &ra, INSERT_VALUES ); CHKERRQ(ierr);
    if( cZ != NULL ) { 
      ra = ( (double)rand() - emin ) * aplpha;     assert(cP != NULL);
      ierr = cZ->SetValues_i( 1, &tt, &ra, INSERT_VALUES ); CHKERRQ(ierr);
      ra = ( (double)rand() - emin ) * aplpha;
      ierr = cP->SetValues_i( 1, &tt, &ra, INSERT_VALUES ); CHKERRQ(ierr);
    }
  }
  ierr = Z->Assembly(); CHKERRQ(ierr);
  ierr = P->Assembly(); CHKERRQ(ierr);
  // test
  if( cZ == NULL ) { 
    ierr = pc->Apply(Z,R); CHKERRQ(ierr);  // r <- MA * Z
    ierr = R->Dot(P,&ra); CHKERRQ(ierr);   // ra <- P' * MA * Z
    ierr = pc->Apply(P,R); CHKERRQ(ierr);  // r <- MA * P
    ierr = R->Dot(Z,&emin); CHKERRQ(ierr); // emin <- Z' * MA * P
  }
  else {
#if defined(PROM_HAVE_COMPLEX)
    PromComplexVector *cR = dynamic_cast<PromComplexVector*>(R);  
    PromComplex cra,ce; assert(cR != NULL);
    ierr = pc->Apply(Z,R); CHKERRQ(ierr);  // r <- MA * Z
    ierr = cR->cDot( cP, cra); CHKERRQ(ierr);   // ra <- P' * MA * Z
    ra = cra.real();
    ierr = pc->Apply(P,R); CHKERRQ(ierr);  // r <- MA * P
    ierr = cR->cDot( cZ, ce ); CHKERRQ(ierr); // emin <- Z' * MA * P
    emin = ce.real();
#endif
  }
  if( fabs((emin-ra)/emin) > 1.e-8 && verbose_ > 0 ) {
    PetscPrintf(Amat->MPIComm(),
		"\t[%d]%s WARNING: NON-SYMMETRIC: (|P'*MA*Z|-|Z'*MA*P|)/|Z'*MA*P|=%e\n", 
		0,__FUNCT__,fabs((emin-ra)/emin));
  }

  return 0;
}

#if defined(TRC)

#include <sys/trcctl.h>
#include <sys/trcmacros.h>
#include <sys/trchkid.h>

#define START_TEST        ( 1 | HKWD_USER2)
#define BEFORE_MAT_MULT   ( 2 | HKWD_USER2)
#define AFTER_MAT_MULT    ( 3 | HKWD_USER2)
#define BEFORE_DOT        ( 4 | HKWD_USER2)
#define AFTER_DOT         ( 5 | HKWD_USER2)
#define BEFORE_PC         ( 6 | HKWD_USER2)
#define AFTER_PC          ( 7 | HKWD_USER2)

#define ENTERING_SOLVE     ( 1 | HKWD_USER3)
#define EXITING_SOLVE      ( 2 | HKWD_USER3)

#endif

#undef __FUNCT__  
#define __FUNCT__ "SolveCG"
/* PromSolver::SolveCG ***************************************************
 *  
 *  KSPSolve_CG_solver 
 *
 */
int PromSolver::SolveCG( const PromVector_base * const bb, 
			 PromVector_base * const xx, int *its,
			 const bool zerox /*=TRUE*/, 
			 const bool avoidnorms /*=FALSE*/, 
			 const PromPCSide /*= PROM_RIGHT*/ )
{ 
  int          ierr,maxit,conv,it,err; 
  double       dpi=1.,a=1.,beta=1.,betaold=1.,b,ma,*e=0,*d=0,oldRho=1.1;
  PromVector_base *const X = xx; double dp,r0=0.0;
  const PromVector_base * const B = bb;
  assert((alloced_ ? work_ : alias_->work_) != NULL);
  PromVector_base *const Z = alloced_ ? work_[1] : alias_->work_[1];
  PromVector_base *const R = alloced_ ? work_[0] : alias_->work_[0];
  PromVector_base *const P = alloced_ ? work_[2] : alias_->work_[2];
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const pc = alloced_ ? pc_ : alias_->pc_;

  PetscFunctionBegin;
#if defined(TRC)
  int mype = Amat->map_.comm_.mype(); static int CG_counter = 0;
  TRCHKGT(ENTERING_SOLVE,++CG_counter,mype,0,0,0);
#endif 
  const bool eigs = (nEig_ >= maxits_);
  maxit   = maxits_;
#if defined(PETSC_USE_BOPT_g) 
  ierr = TestSymmMA( Amat, pc, Z, P, R );   CHKERRQ(ierr);
#endif
  if (eigs) {
    e = origin_.e; d = d_.e; assert(e && d); e[0] = 0.0; b = 0.0; 
  }
  
  if( !zerox ){
#if defined(PETSC_USE_BOPT_g)
    double beta; X->Norm2( &beta );
    if(beta==0.0 && !zerox)
      PetscPrintf(Amat->MPIComm(),"\t\t[%d]%s WARNING: |x|=0\n",0,__FUNCT__);
#endif
    ierr = KSPMult( Amat,  X, R );  CHKERRQ(ierr);
    ierr = R->AYPX( -1.0, B );  CHKERRQ(ierr); /* Aux <- b - Ax */
  }
  else { 
    ierr = R->Copy(B);     CHKERRQ(ierr);  /* r <- b (x is 0) */
    ierr = X->Set( 0.0 );  CHKERRQ(ierr);    
  }
  ierr = R->Norm2( &dp ); CHKERRQ(ierr);   /* dp <- sqrt(r'*r) */
  r0 = dp;  
  if( !avoidnorms ){
    ierr = Monitor( 0, dp );    CHKERRQ(ierr);
    conv = converged( r0, dp );
    if (conv != 0){ *its = 0; PetscFunctionReturn(0); }
  }
  else if( r0 < 1.e-300 ) { *its = 0; PetscFunctionReturn(0); }
  
  for( it=0, err=0 ; it<maxit; it++ ) {
#if defined(TRC)
    TRCHKGT(BEFORE_PC,it+1,mype,0,0,0);
#endif
    ierr = pc->Apply( R, Z ); CHKERRQ(ierr);     /* z <- Br  */
#if defined(TRC)
    TRCHKGT(AFTER_PC,it+1,mype,0,0,0);
#endif
#if defined(TRC)
    TRCHKGT(BEFORE_DOT,it+1,mype,0,0,0);
#endif
    Z->Dot(R,&beta);                             /*     beta <- r'z     */
#if defined(TRC)
    TRCHKGT(AFTER_DOT,it+1,mype,0,0,0);
#endif
    if( beta <= 0.0 ) {
      err++;
      if( verbose_ > 0 )
	PetscPrintf(Amat->MPIComm(),"[%d]%s %d) indefinite!!! (%g)\n",0,__FUNCT__,it+1,beta);
      ierr = Amat->SetIndefinite(); CHKERRQ(ierr);  
    }
    //if ( beta == 0.0 ) break; // happy breakdown?, messes up eig.
    if (it == 0) {
      ierr = P->Copy(Z); CHKERRQ(ierr);         /*     p <- z          */
    }  
    else {
      b = beta/betaold;
      if (eigs) {
	e[it] = sqrt(fabs(b))/a;  
      }
      ierr = P->AYPX(b,Z); CHKERRQ(ierr);      /*     p <- z + b* p   */
    }
    betaold = beta;
#ifdef HPM_MM
    hpmStart( 2, "MatMult" );
#endif /* HPM_MM */
#if defined(TRC)
    TRCHKGT(BEFORE_MAT_MULT,it+1,mype,0,0,0);
#endif
    ierr = KSPMult( Amat, P,Z); CHKERRQ(ierr);     /*     z <- Kp         */
#if defined(TRC)
    TRCHKGT(AFTER_MAT_MULT,it+1,mype,0,0,0);
#endif
#ifdef HPM_MM  
    hpmStop( 2 );
#endif /* HPM_MM */
    P->Dot(Z,&dpi);                            /*     dpi <- z'p      */
    a = beta/dpi;                              /*     a = beta/p'z    */
    if (eigs) {
      d[it] = sqrt(fabs(b))*e[it] + 1.0/a;
    }
    ierr = X->AXPY(a,P); CHKERRQ(ierr); /* x <- x + ap */
    ma = -a; R->AXPY(ma,Z);                    /*     r <- r - az     */

    if ( !avoidnorms ) {
      ierr = R->Norm2( &dp );   CHKERRQ(ierr);  /* dp <- sqrt( r'*r ) */
      conv = converged( r0, dp );  // cheap
      ierr = Monitor(it+1,dp,conv!=0);  CHKERRQ(ierr);
      if( conv != 0 ){ it++;  break; }
    }
    else if( verbose_ < 0 && it%4 == 0 && it > 0 ) {
      //if( err != 0 ) { it++;  break; } 
      double emax, emin;
      ComputeExtremeSVD( it+1, emax, emin );
      if( it == 4 ) oldRho = emax;
      else{
	if( (emax-oldRho)/emax < 0.2 || emax>10.0 || emin<0.){ 
	  it++; break; 
	} 
	oldRho = emax;
      }
    }
  }
  //
  if ( !avoidnorms && (conv==-1 || conv==0) ) *its = -it;
  else *its = it;
#if defined(TRC)
  TRCHKGT(EXITING_SOLVE,CG_counter,mype,0,0,0);
#endif

  PetscFunctionReturn(-err);
}


#undef __FUNCT__  
#define __FUNCT__ "cSolveCG"
/* PromSolver::cSolveCG ***************************************************
 *  
 *  Complex CG olver 
 *
 */
int PromSolver::cSolveCG( const PromVector_base * const bb, 
			 PromVector_base * const xx, int *its,
			 const bool zerox /*=TRUE*/, 
			 const bool avoidnorms /*=FALSE*/, 
			 const PromPCSide /*= PROM_RIGHT*/ )
{ 
  PetscFunctionBegin;
  int err = 0;
#if defined(PROM_HAVE_COMPLEX)
  int          ierr,maxit,conv,it; 
  PromComplex  dpi=1.,b,beta=1.,betaold=1.,ma,a=1.;
  double dp,r0=0.0,*e=0,*d=0,beta_r,oldRho=1.1;
  const PromVector_base * const B = bb; PromVector_base *const X_b = xx;
  assert((alloced_ ? work_ : alias_->work_) != NULL);
  PromVector_base *const Z_b = alloced_ ? work_[1] : alias_->work_[1];
  PromVector_base *const R_b = alloced_ ? work_[0] : alias_->work_[0];
  PromVector_base *const P_b = alloced_ ? work_[2] : alias_->work_[2];
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const pc = alloced_ ? pc_ : alias_->pc_;
  PromComplexVector *Z = dynamic_cast<PromComplexVector*>(Z_b);
  assert(Z != NULL);
  PromComplexVector *R = dynamic_cast<PromComplexVector*>(R_b);
  assert(R != NULL);
  PromComplexVector *P = dynamic_cast<PromComplexVector*>(P_b);
  assert(P != NULL);
  PromComplexVector *X = dynamic_cast<PromComplexVector*>(X_b);
  assert(X != NULL);

  const bool eigs = (nEig_ >= maxits_);
  maxit   = maxits_;
  //#if defined(PETSC_USE_BOPT_g) 
  if( verbose_ > 1 ) {
    ierr = TestSymmMA( Amat, pc, Z, P, R );   CHKERRQ(ierr);
  }
  //#endif
  if (eigs) {
    e = origin_.e; d = d_.e; assert(e && d); e[0] = 0.0; b = 0.0; 
  }

  if( !zerox ){
    ierr = KSPMult( Amat,  X, R );  CHKERRQ(ierr);
    ierr = R->AYPX( -1.0, B );  CHKERRQ(ierr); /* Aux <- b - Ax */
  }
  else { 
    ierr = R->Copy(B);     CHKERRQ(ierr);  /* r <- b (x is 0) */
    ierr = X->Set( 0.0 );  CHKERRQ(ierr);    
  }
  ierr = R->Norm2( &dp ); CHKERRQ(ierr);   /* dp <- sqrt(r'*r) */
  r0 = dp;  
  if( !avoidnorms ){
    ierr = Monitor( 0, dp );    CHKERRQ(ierr);
    conv = converged( r0, dp );
    if (conv != 0){ *its = 0; PetscFunctionReturn(0); }
  }
  else if( r0 < 1.e-300 ) { *its = 0; PetscFunctionReturn(0); }

  for( it=0, err=0 ; it<maxit; it++ ) {
    ierr = pc->Apply( R, Z ); CHKERRQ(ierr);     /* z <- Br  */
    Z->cDot( R, beta );                         /*     beta <- r'z     */
    beta_r = beta.real();
    if( beta_r <= 0.0 ) {
      err++;
      if( verbose_ > 0 )
	PetscPrintf(Amat->MPIComm(),"[%d]%s %d) indefinite!!! (%g)\n",
		    0,__FUNCT__,it+1,beta_r);
      ierr = Amat->SetIndefinite(); CHKERRQ(ierr);  
    }
    if (it == 0) {
      ierr = P->Copy( Z ); CHKERRQ(ierr);       /*     p <- z          */
    }  
    else{
      b = beta/betaold;
      if( eigs ) { ma = std::sqrt(b/a); e[it] = ma.real(); }
      ierr = P->cAYPX( b, Z ); CHKERRQ(ierr);   /*     p <- z + b* p   */
    }
    betaold = beta;
    ierr = KSPMult( Amat, P,Z); CHKERRQ(ierr);  /*     z <- Kp         */
    P->cDot( Z, dpi );                         /*     dpi <- z'p      */
    a = beta/dpi;                               /*     a = beta/p'z    */
    if( eigs ){ ma = std::sqrt(b) * ma + 1.0/a; d[it] = ma.real(); }
    ierr = X->cAXPY( a, P) ; CHKERRQ(ierr);      /* x <- x + ap */
    ma = -a; R->cAXPY(ma,Z);                     /*     r <- r - az     */

    if ( !avoidnorms ) {
      ierr = R->Norm2( &dp );   CHKERRQ(ierr);  /* dp <- sqrt( r'*r ) */
      conv = converged( r0, dp );  // cheap
      ierr = Monitor(it+1,dp,conv!=0);  CHKERRQ(ierr);
      if( conv != 0 ){ it++;  break; }
    }
    else if( verbose_ < 0 && it%4 == 0 && it > 0 ) {
      //if( err != 0 ) { it++;  break; } 
      double emax, emin;
      ComputeExtremeSVD( it+1, emax, emin );
      if( it == 4 ) oldRho = emax;
      else{
	if( (emax-oldRho)/emax < 0.2 || emax>10.0 || emin<0.){ 
	  it++; break; 
	} 
	oldRho = emax;
      }
    }
  }
  //
  if ( !avoidnorms && (conv==-1 || conv==0) ) *its = -it;
  else *its = it;
#endif
  PetscFunctionReturn(err);
}

#undef __FUNCT__  
#define __FUNCT__ "SolveCheb"
/* PromSolver::SolveCheb ***************************************************
 *  
 *  Chebyshev smoother
 */
int PromSolver::SolveCheb( const PromVector_base * const bb, 
			   PromVector_base * const xx, int *its,
			   const bool zerox /*=TRUE*/, 
			   const bool avoidnorms /*=FALSE*/, 
			   const PromPCSide /*= PROM_RIGHT*/ )
{ 
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  int   ierr;
  PromVector_base *const X = xx;
  const PromVector_base * const B = bb;
  assert((alloced_ ? work_ : alias_->work_) != NULL);
  PromVector_base * const Res = alloced_ ? work_[0] : alias_->work_[0];
  PromVector_base * const Dk = alloced_ ? work_[1] : alias_->work_[1];
  PromVector_base * const Aux = alloced_ ? work_[2] : alias_->work_[2];
  PromPC_base * const pc = alloced_ ? pc_ : alias_->pc_; assert(pc!=NULL);
  double over,under,rhok,rhokp1,dtemp1,dtemp2;

  PetscFunctionBegin;
  pc->SetNumIts(1);

  over = 1.1; under = 1./50.; rhok = Amat->getDRho();
  if(rhok <= 0.0){ SETERRQ1(PETSC_COMM_SELF,-1,"|DA| = %e",rhok); }
  const int deg = maxits_; 
  double const beta =  rhok * over; 
  double const alpha = rhok * under;
  double const delta = (beta - alpha)/2.;
  double const theta = (beta + alpha)/2.;
  double const s1 = theta / delta;
  rhok = 1./s1;

  // put residual into 'Res'
  if( !zerox ) {
    ierr = KSPMult( Amat,  X, Aux );  CHKERRQ(ierr);
    ierr = Aux->AYPX( -1.0, B );  CHKERRQ(ierr); /* Aux <- b - Ax */
  }
  else {
    ierr = Aux->Copy( B );        CHKERRQ(ierr); /* Aux <- b */
  }
  ierr = pc->Apply( Aux, Dk );    CHKERRQ(ierr); /* Dk <- D^-1 Aux*/
  ierr = Dk->Scale( 1. / theta ); CHKERRQ(ierr);
  if( !zerox ) ierr = X->AXPY( 1.0, Dk ); /* x <- x + Dk */
  else X->Copy( Dk );             CHKERRQ(ierr);
  for( int k = 0 ; k < deg - 1 ; k++ ) {
    //
    rhokp1 = 1./(2.*s1 - rhok);
    dtemp1 = rhokp1*rhok;
    dtemp2 = 2.*rhokp1/delta;
    rhok = rhokp1;
    //
    ierr = KSPMult( Amat,  X, Aux );   CHKERRQ(ierr);
    ierr = Aux->AYPX( -1.0, B );       CHKERRQ(ierr);  /* Aux <- b - Ax */
    ierr = pc->Apply(Aux, Res );       CHKERRQ(ierr);  /* Res <- D^-1(b-Ax) */

    // Dk[] = rhokp1 * rhok * Dk[] + 2. * rhokp1 * res[] / ( delta * diag[] )
    ierr = Dk->AXPBY( dtemp2, dtemp1, Res ); CHKERRQ(ierr); 
    ierr = X->AXPY( 1.0, Dk );               CHKERRQ(ierr);   /* x <- x + Dk */
  }
  *its = deg;

  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "cSolveCheb"
/* PromSolver::SolveCheb ***************************************************
 *  
 *  complex Chebyshev smoother
 */
int PromSolver::cSolveCheb( const PromVector_base * const bb, 
			    PromVector_base * const xx, int *its,
			    const bool zerox /*=TRUE*/, 
			    const bool avoidnorms /*=FALSE*/, 
			    const PromPCSide /*= PROM_RIGHT*/ )
{ 
  PetscFunctionBegin;
#if defined(PROM_HAVE_COMPLEX)
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  int   ierr;
  PromVector_base *const X = xx;
  const PromVector_base * const B = bb;
  assert((alloced_ ? work_ : alias_->work_) != NULL);
  PromVector_base * const Res_b = alloced_ ? work_[0] : alias_->work_[0];
  PromVector_base * const Dk_b = alloced_ ? work_[1] : alias_->work_[1];
  PromVector_base * const Aux = alloced_ ? work_[2] : alias_->work_[2];
  PromPC_base * const pc = alloced_ ? pc_ : alias_->pc_; assert(pc!=NULL);
  double over,under,rhok,rhokp1,dtemp1,dtemp2;

  pc->SetNumIts(1);

  over = 1.1; under = 1./50.; rhok = Amat->getDRho();
  if(rhok <= 0.0){ SETERRQ1(PETSC_COMM_SELF,-1,"|DA| = %e",rhok); }
  const int deg = maxits_; 
  double const beta =  rhok * over; 
  double const alpha = rhok * under;
  double const delta = (beta - alpha)/2.;
  double const theta = (beta + alpha)/2.;
  double const s1 = theta / delta;
  rhok = 1./s1;

  PromComplex ct1,ct2;
  PromComplexVector *Dk = dynamic_cast<PromComplexVector*>(Dk_b);
  assert(Dk != NULL);
  PromComplexVector *Res = dynamic_cast<PromComplexVector*>(Res_b);
  assert(Res != NULL);

  // put residual into 'Res'
  if( !zerox ) {
    ierr = KSPMult( Amat,  X, Aux );  CHKERRQ(ierr);
    ierr = Aux->AYPX( -1.0, B );  CHKERRQ(ierr); /* Aux <- b - Ax */
  }
  else {
    ierr = Aux->Copy( B );        CHKERRQ(ierr); /* Aux <- b */
  }
  ierr = pc->Apply( Aux, Dk );    CHKERRQ(ierr); /* Dk <- D^-1 Aux*/
  ct1 = PromComplex( 1. / theta, 0.0 );
  ierr = Dk->cScale( ct1 ); CHKERRQ(ierr);
  if( !zerox ) ierr = X->AXPY( 1.0, Dk ); /* x <- x + Dk */
  else X->Copy( Dk );             CHKERRQ(ierr);
  for( int k = 0 ; k < deg - 1 ; k++ ) {
    //
    rhokp1 = 1./(2.*s1 - rhok);
    dtemp1 = rhokp1*rhok;
    dtemp2 = 2.*rhokp1/delta;
    rhok = rhokp1;
    //
    ierr = KSPMult( Amat, X, Aux );   CHKERRQ(ierr);
    ierr = Aux->AYPX( -1.0, B );      CHKERRQ(ierr);  /* Aux <- b - Ax */
    ierr = pc->Apply(Aux, Res_b );    CHKERRQ(ierr);  /* Res <- D^-1(b-Ax) */

    // Dk[] = rhokp1 * rhok * Dk[] + 2. * rhokp1 * res[] / ( delta * diag[] )
    ct2 = PromComplex( dtemp2, 0.0 ); ct1 = PromComplex( dtemp1, 0.0 );
    ierr = Dk->cAXPBY( ct2, ct1, Res ); CHKERRQ(ierr); 
    ierr = X->AXPY( 1.0, Dk );        CHKERRQ(ierr);   /* x <- x + Dk */
  }
  *its = deg;
#endif
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "BrezParameters"
/* PromPCBrez::BrezParameters*************************************
 *
 *  - creates Brexina's prameters
 * 
 */
int PromSolver::BrezParameters( const PromMatrix_base * const AA,
				double &mlsOver, double &mlsCf,
				double &mlsOm, double &mlsOm2) 
{
  double rho; 
  PromPC_base *const pc = alloced_ ? pc_ : alias_->pc_; assert(pc!=NULL);
  
  PetscFunctionBegin; 
  mlsOver = 1.2; // out 
  rho = AA->getDRho(); if(rho <= 0.0)SETERRQ1(PETSC_COMM_SELF,-1,"|DA| = %e",rho);

  // setup (mlsOm2,mlsOm,Arho)
  const double pi=4.e0 * atan(1.e0); /* 3.141592653589793115998e0; */
  double       mlsBoost,rho2,ddeg,aux0,aux1; 
  double       aux_om,om_loc;
  int          deg = 1;

  /* Boost the largest eigenvalue by a fudge factor  */
  /* (mlsOver is now set to 1.1 in the code[1/7/02]).*/
  /* We know that our eigenvalue estimate is a lower */
  /* bound. Undershooting the largest eigenvalue for */
  /* a polynomial method is a bit risky.             */

  rho *= mlsOver; // widget->mlsOver;

  /* compute a set of polynomial coefficients. These */
  /* almost look like Chebyshev coefficients. They   */
  /* correspond to                                   */
  /*        2/(rho * ( 1 - cos(2 pi (j+1)/(2n + 1))) */
  /* By the way, this polynomial is done via products*/
  /* I think? that it corresponds to                 */
  /*    (I - om_loc[0] A) * (I - om_loc[1] A) * ...  */

  ddeg = (double)deg; 
  aux1 = 1.e0/(2.e0 * ddeg + 1.e0); 
  
  for(int j=0; j<deg; j++) { 
    aux0 = 2.e0 * (double)(j+1) * pi;  
    aux_om = rho/2.e0 * (1.e0 - cos(aux0 * aux1));
    om_loc/*[j]*/ = 1.e0/aux_om;
  }
  /* Compute the coefficients of the polynomial      */
  /*    (I - om_loc[0] A) * (I - om_loc[1] A) * ...  */
  
  mlsCf = om_loc; // out

  /* Here I believe that we are trying to estimate the largest */
  /* to estimate the largest eigenvalue of p(A)*p(A)*A where   */
  /*   p(A) = (I - om_loc[0] A) * (I - om_loc[1] A) * ...      */
  /* Normally, deg = 1 so we just go the hardwired stuff in the*/
  /* else clause.                                              */
  
  rho2 = 4.0e0/(27.e0 * om_loc);

  /* Boost the eigenvalue estimate for p(A)*p(A)*A */
  /* and store coefficients for later application. */
  
  mlsBoost = 1.019e0;

  rho2 *= mlsBoost;
  mlsOm2 = 2.0e0/rho2; // out
  mlsOm = om_loc; 

  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "SolveBrez"
/* PromSolver::SolveBrez ***************************************************
 *  
 *  Marian Brezina (Chebyshev) smoother
 *
 */
int PromSolver::SolveBrez( const PromVector_base * const bb, 
			   PromVector_base * const xx, int *its,
			   const bool zerox /*=TRUE*/, 
			   const bool avoidnorms /*=FALSE*/, 
			   const PromPCSide /*= PROM_RIGHT*/ )
{ 
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  int   ierr;
  PromVector_base *const X = xx;
  const PromVector_base * const B = bb;
  assert((alloced_ ? work_ : alias_->work_) != NULL);
  PromVector_base * const Res = alloced_ ? work_[0] : alias_->work_[0];
  PromVector_base * const work = alloced_ ? work_[1] : alias_->work_[1];
  PromVector_base * const Aux = alloced_ ? work_[2] : alias_->work_[2];
  PromPC_base * const pc = alloced_ ? pc_ : alias_->pc_; assert(pc!=NULL);
  double mlsOver,mlsCf,mlsOm,mlsOm2;

  PetscFunctionBegin; 
  ierr = BrezParameters( Amat, mlsOver, mlsCf,mlsOm, mlsOm2 );
  CHKERRQ(ierr);
  pc->SetNumIts(1); // scaling 

  *its = 4; // hard wired
  // 1 -- get aux
  if( !zerox ) {
    ierr = KSPMult( Amat,  X, Aux );        CHKERRQ(ierr);
    ierr = Aux->AYPX( -1.0, B );        CHKERRQ(ierr); /* Aux <- b - Ax */
  }
  else {
    ierr = Aux->Copy( B );        CHKERRQ(ierr); /* Aux <- b */
  }
  ierr = pc->Apply( Aux, Res );   CHKERRQ(ierr); /* Res <-  D^-1 * Aux */

  // 2 -- scale and start
  double cf = mlsOver * mlsCf;
  if( !zerox ) X->AXPY( cf, Res );  /* x <- x - cf * Res */
  else{ X->Copy( Res ); CHKERRQ(ierr); X->Scale( cf ); } CHKERRQ(ierr);

  // 3 -- apply two 
  {
    ierr = KSPMult( Amat,  X, Aux );    CHKERRQ(ierr); /* D^-1 * (Ax - B)*/
    ierr = Aux->AXPY( -1.0, B );         CHKERRQ(ierr); 
    ierr = pc->Apply( Aux, Res );   CHKERRQ(ierr);
    ierr = SandwPre(&mlsOm,1, pc, Amat, Res, work, Aux); CHKERRQ(ierr);
    ierr = SandwPost(&mlsOm,1, pc, Amat, work, Aux, Res); CHKERRQ(ierr);

    double cf = mlsOver * mlsOm2;
    ierr = X->AXPY( -cf, Aux  );    CHKERRQ(ierr); /* x <- x - cf * Aux */
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "SandwPre"
/* PromSolver::SandwPre ***************************************************
 *  
 */
int PromSolver::SandwPre( const double omV[], const int deg, 
			  PromPC_base * const pc,
			  const PromMatrix_base * const Amat, 
			  PromVector_base *const xx,PromVector_base*const yy,
			  PromVector_base *const work ) const
{
  int dg,ierr;
  double om;
  
  PetscFunctionBegin; 
  ierr = yy->Copy( xx ); CHKERRQ(ierr); 
  for (dg=deg-1; dg>=0; dg--) { 
    ierr = KSPMult( Amat,  yy, xx );           CHKERRQ(ierr);
    ierr = pc->Apply( xx, work );   CHKERRQ(ierr);
    om = omV[dg]; 
    //for (i=0; i<n; i++) y[i] -= om * x[i]; 
    ierr = yy->AXPY( -om, work ); CHKERRQ(ierr); /* y <- y - om * x */
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "SandwPost"
/* PromSolver::SandwPost ***************************************************
 *  
 */
int PromSolver::SandwPost( const double omV[], const int deg, 
			   PromPC_base * const pc,
			   const PromMatrix_base * const Amat, 
			   PromVector_base *const xx,
			   PromVector_base *const yy, 
			   PromVector_base *const work ) const
{
  int dg,ierr;
  double om;
  
  PetscFunctionBegin; 
  ierr = yy->Copy( xx ); CHKERRQ(ierr); 
  for (dg=0; dg<deg; dg++) { 
    ierr = KSPMult( Amat,  yy, xx );           CHKERRQ(ierr);
    ierr = pc->Apply( xx, work );   CHKERRQ(ierr);
    om = omV[dg]; 
    //for (i=0; i<n; i++) y[i] -= om * x[i]; 
    ierr = yy->AXPY( -om, work ); CHKERRQ(ierr); /* y <- y - om * x */
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "SolveRichardson"
/* PromSolver::SolveRichardson **********************************************
 *  
 */
int PromSolver::SolveRichardson( const PromVector_base * const bb,
				 PromVector_base * const xx, int *its,
				 const bool zerox /*=TRUE*/, 
				 const bool avoidnorms /*=FALSE*/, 
				 const PromPCSide /*= PROM_RIGHT*/ ) 
{
  int it,ierr,conv = 0;
  double r0, norm;
  assert((alloced_ ? work_ : alias_->work_) != NULL);
  PromVector_base *rr = alloced_ ? work_[0] : alias_->work_[0];
  PromVector_base *temp = alloced_ ? work_[1] : alias_->work_[1];  
  PromPC_base * const pc = alloced_ ? pc_ : alias_->pc_;
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;

  if( zerox ){ ierr = rr->Copy( bb ); CHKERRQ(ierr); }
  else {
    ierr = KSPMult( Amat,  xx, rr );   /* b - ax */        CHKERRQ(ierr);
    ierr = rr->AYPX( -1.0, bb );                       CHKERRQ(ierr);
  }
  // doit 
  if( avoidnorms == FALSE ) {
    ierr = rr->Norm2( &r0 );   CHKERRQ(ierr); 
    ierr = Monitor(0,r0);      CHKERRQ(ierr);
    conv = converged(r0,r0);
    if (conv != 0){ *its = 0; return 0; }
  }
  for(it=0;it<maxits_;it++){
    ierr = pc->Apply( rr, temp );      CHKERRQ(ierr); /* t <- B * r  */ 
    if( zerox && it == 0 ) ierr = xx->Copy( temp ); 
    else ierr = xx->AXPY( 1.0, temp );    /* x' <- x' + w * t  */ 
    CHKERRQ(ierr); 
    if( it == maxits_ - 1 ) { it++; break; } // avoid last residual    

    ierr = KSPMult( Amat,  xx, rr );    CHKERRQ(ierr); /* r <- b - A x'     */ 
    ierr = rr->AYPX( -1.0, bb );        CHKERRQ(ierr);
    if( !avoidnorms ) {
      ierr = rr->Norm2( &norm );        CHKERRQ(ierr);
      conv = converged(r0,norm);
      ierr = Monitor(it+1,norm,conv!=0);        CHKERRQ(ierr);
      if( conv != 0 ){ it++; break; }
    }
  }
  if( !avoidnorms && (conv == -1 || conv == 0) ) *its = -it;
  else *its = it;

  return 0;
}
 
#undef __FUNCT__  
#define __FUNCT__ "SolveCR"
/* PromSolver::SolveCR ***************************************************
 *  
 */
int PromSolver::SolveCR( const PromVector_base * const bb,
			 PromVector_base * const xx, int *its,
			 const bool zerox /*=TRUE*/, 
			 const bool avoidnorms /*=FALSE*/, 
			 const PromPCSide /*= PROM_RIGHT*/ ) 
{
  double apq,btop,bbot,bi,ai; int ierr,conv,it,maxit=maxits_;
  PromVector_base *X,*R,*RT,*P,*AP,*ART,*Q; double r0,dp;
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const      pc = alloced_ ? pc_ : alias_->pc_;
  assert((alloced_ ? work_ : alias_->work_) != NULL);

  PetscFunctionBegin;
  X     = xx;
  const PromVector_base * const B = bb;
  R     = alloced_ ? work_[0] : alias_->work_[0];    
  RT    = alloced_ ? work_[1] : alias_->work_[1];
  P     = alloced_ ? work_[2] : alias_->work_[2];
  AP    = alloced_ ? work_[3] : alias_->work_[3];
  ART   = alloced_ ? work_[4] : alias_->work_[4];
  Q     = alloced_ ? work_[5] : alias_->work_[5];

#if defined(PETSC_USE_BOPT_g)
  ierr = TestSymmMA( Amat, pc, Q, P, R ); CHKERRQ(ierr);
#endif
  
  if (!zerox) {
    ierr = KSPMult( Amat,  X, R ); CHKERRQ(ierr);    /*   r <- b - Ax       */
    ierr = R->AYPX( -1.0, B ); CHKERRQ(ierr);
  } else { 
    ierr = R->Copy(B);CHKERRQ(ierr);            /*    r <- b (x is 0)  */
    ierr = X->Set( 0.0 );    CHKERRQ(ierr);
  }
  // monitor
  if ( !avoidnorms ) {
    ierr = R->Norm2(&dp);CHKERRQ(ierr);
    r0 = dp;
    conv = converged( r0, dp );
    ierr = Monitor(0,dp,conv!=0); CHKERRQ(ierr);
    if (conv!=0){ *its = 0; PetscFunctionReturn(0); }
  }
  // pc
  ierr = pc->Apply( R, P  );   CHKERRQ(ierr);/*     p <- Br         */

  ierr = KSPMult( Amat,  P, AP );  CHKERRQ(ierr);   /*   AP <- A p          */
  ierr = RT->Copy( P );        CHKERRQ(ierr);   /*   rt <- p            */
  ierr = ART->Copy( AP );      CHKERRQ(ierr);   /*   ART <- AP          */
  ierr = ART->Dot( RT, &btop); CHKERRQ(ierr);   /*   (RT,ART)           */
  for (it=0; it<maxit; it++) {
    ierr = pc->Apply( AP, Q );         CHKERRQ(ierr); /*   q <- B AP */
 
    ierr   = AP->Dot(Q,&apq);          CHKERRQ(ierr);  
    ai = btop/apq;                              /* ai = (RT,ART)/(AP,Q) */
    ierr   = X->AXPY(ai,P); CHKERRQ(ierr);      /*   x <- x + ai p      */
    ierr   = RT->AXPY(-ai,Q); CHKERRQ(ierr);    /*   rt <- rt - ai q    */
    ierr   = KSPMult( Amat, RT,ART); CHKERRQ(ierr); /*  RT <-   ART   */
    bbot = btop;
    ierr   = RT->Dot(ART,&btop); CHKERRQ(ierr); /*   btop <- RT *  ART  */
    if ( !avoidnorms ) {
      ierr = R->AXPY(-ai,AP);CHKERRQ(ierr);     /*   R   <- R - ai*AP    */
      ierr = R->Norm2( &dp );CHKERRQ(ierr);     /*   dp <- R'*R          */
      conv = converged(r0,dp);
      Monitor(it+1,dp,conv!=0);
      if (conv!=0) { it++;  break; }
    }
    bi = btop/bbot;
    ierr = P->AYPX(bi,RT); CHKERRQ(ierr);     /*   P <- rt + Bi P     */
    ierr = AP->AYPX(bi,ART); CHKERRQ(ierr);   /*   AP <- Art + Bi AP  */
  }
  if( !avoidnorms && (conv == -1 || conv == 0) ) *its = -it;
  else *its = it;

  PetscFunctionReturn(0);
}
 
#undef __FUNCT__  
#define __FUNCT__ "cSolveCR"
/* PromSolver::cSolveCR ***************************************************
 *  
 */
int PromSolver::cSolveCR( const PromVector_base * const bb,
			   PromVector_base * const xx, int *its,
			   const bool zerox /*=TRUE*/, 
			   const bool avoidnorms /*=FALSE*/, 
			   const PromPCSide /*= PROM_RIGHT*/ ) 
{ 
  PetscFunctionBegin;
#if defined(PROM_HAVE_COMPLEX)
  PromComplex btop,apq,bbot,bi,ai; int ierr,conv,it,maxit=maxits_;
  PromVector_base *X_b,*R_b,*RT_b,*P_b,*AP_b,*ART_b,*Q_b; double r0,dp;
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const      pc = alloced_ ? pc_ : alias_->pc_;
  assert((alloced_ ? work_ : alias_->work_) != NULL);

  X_b   = xx;
  const PromVector_base * const B = bb;
  R_b   = alloced_ ? work_[0] : alias_->work_[0];    
  RT_b  = alloced_ ? work_[1] : alias_->work_[1];
  P_b   = alloced_ ? work_[2] : alias_->work_[2];
  AP_b  = alloced_ ? work_[3] : alias_->work_[3];
  ART_b = alloced_ ? work_[4] : alias_->work_[4];
  Q_b   = alloced_ ? work_[5] : alias_->work_[5];
  PromComplexVector *ART = dynamic_cast<PromComplexVector*>(ART_b);
  assert(ART != NULL);
  PromComplexVector *AP = dynamic_cast<PromComplexVector*>(AP_b);
  assert(AP != NULL);
  PromComplexVector *Q = dynamic_cast<PromComplexVector*>(Q_b);
  assert(Q != NULL);
  PromComplexVector *RT = dynamic_cast<PromComplexVector*>(RT_b);
  assert(RT != NULL);
  PromComplexVector *R = dynamic_cast<PromComplexVector*>(R_b);
  assert(R != NULL);
  PromComplexVector *X = dynamic_cast<PromComplexVector*>(X_b);
  assert(X != NULL);
  PromComplexVector *P = dynamic_cast<PromComplexVector*>(P_b);
  assert(P != NULL);

#if defined(PETSC_USE_BOPT_g)
  ierr = TestSymmMA( Amat, pc, Q, P, R ); CHKERRQ(ierr);
#endif

  if (!zerox) {
    ierr = KSPMult( Amat,  X, R ); CHKERRQ(ierr);    /*   r <- b - Ax     */
    ierr = R->AYPX( -1.0, B ); CHKERRQ(ierr);
  } else { 
    ierr = R->Copy(B);CHKERRQ(ierr);                /*    r <- b (x is 0) */
    ierr = X->Set( 0.0 );    CHKERRQ(ierr);
  }
  // monitor
  if ( !avoidnorms ) {
    ierr = R->Norm2(&dp);CHKERRQ(ierr);
    r0 = dp;
    conv = converged( r0, dp );
    ierr = Monitor(0,dp,conv!=0); CHKERRQ(ierr);
    if (conv!=0){ *its = 0; PetscFunctionReturn(0); }
  }
  // pc
  ierr = pc->Apply( R, P  );   CHKERRQ(ierr);       /*     p <- Br    */
  ierr = KSPMult( Amat,  P, AP );  CHKERRQ(ierr);   /*   AP <- A p    */
  ierr = RT->Copy( P );        CHKERRQ(ierr);       /*   rt <- p      */
  ierr = ART->Copy( AP );      CHKERRQ(ierr);       /*   ART <- AP    */
  ierr = ART->cDot( RT, btop); CHKERRQ(ierr);      /*   (RT,ART)     */
  for (it=0; it<maxit; it++) {
    ierr = pc->Apply( AP, Q );         CHKERRQ(ierr); /*   q <- B AP  */
    ierr   = AP->cDot( Q, apq );      CHKERRQ(ierr);  
    ai = btop/apq;                              /* ai = (RT,ART)/(AP,Q) */
    ierr   = X->cAXPY(ai,P); CHKERRQ(ierr);    /*   x <- x + ai p      */
    ierr   = RT->cAXPY(-ai,Q); CHKERRQ(ierr);  /*   rt <- rt - ai q    */
    ierr   = KSPMult( Amat, RT,ART); CHKERRQ(ierr);/*   RT <-   ART   */
    bbot = btop;
    ierr   = RT->cDot( ART, btop ); CHKERRQ(ierr); /*  btop <- RT *  ART */
    if ( !avoidnorms ) {
      ierr = R->cAXPY(-ai,AP);CHKERRQ(ierr);     /*   R   <- R - ai*AP */
      ierr = R->Norm2( &dp );CHKERRQ(ierr);     /*   dp <- R'*R       */
      conv = converged(r0,dp);
      Monitor(it+1,dp,conv!=0);
      if (conv!=0) { it++;  break; }
    }
    bi = btop/bbot;
    ierr = P->cAYPX(bi,RT); CHKERRQ(ierr);     /*   P <- rt + Bi P     */
    ierr = AP->cAYPX(bi,ART); CHKERRQ(ierr);   /*   AP <- Art + Bi AP  */
  }
  if( !avoidnorms && (conv == -1 || conv == 0) ) *its = -it;
  else *its = it;
#endif
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "PromPCApplyBAorAB"
/* PromSolver::PromPCApplyBAorAB ******************************************
 *  
 */
int PromSolver::PromPCApplyBAorAB( const PromMatrix_base * const Amat,
		       PromPC_base * const pc, const PromPCSide pc_side,
		       const PromVector_base * const xx,
		       PromVector_base * const dest,
		       PromVector_base * const zz )const
{
  int ierr;
  PetscFunctionBegin;
  if( pc_side == PROM_RIGHT || pc_side == PROM_FLEX ) {
    ierr = pc->Apply( xx, zz );CHKERRQ(ierr);  /* AB */
    ierr = KSPMult( Amat, zz, dest ); CHKERRQ(ierr);
  } 
  else if( pc_side == PROM_LEFT ) {
    ierr = KSPMult( Amat,  xx, zz );CHKERRQ(ierr); /* BA */
    ierr = pc->Apply( zz, dest );CHKERRQ(ierr);
  }
  else if( pc_side == PROM_SYMM ) {
    SETERRQ(PETSC_COMM_SELF,-1,"pc_side == PROM_SYMM not supported???");
  }
  else assert(0);
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "SolveTFQMR"
/* PromSolver::SolveTFQMR ***************************************************
 *  
 */
int PromSolver::SolveTFQMR( const PromVector_base * const bb,
			    PromVector_base * const xx, int *its,
			    const bool zerox /*=TRUE*/, 
			    const bool avoidnorms /*=FALSE*/, 
			    const PromPCSide pc_side /*= PROM_LEFT*/ ) 
{
  double rho,rhoold,a,s,b,m,psi,eta,etaold,psiold,cf,cm;
  double dpest; double r0,dp,tau,dpold,w;
  int ierr,conv,it,maxit=maxits_;
  PromVector_base *X,*V,*P,*R,*RP,*T,*T1,*Q,*U,*D,*BINVF,*AUQ;
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const      pc = alloced_ ? pc_ : alias_->pc_;
  assert((alloced_ ? work_ : alias_->work_) != NULL);

  PetscFunctionBegin;
  X     = xx;
  const PromVector_base * const B = bb;
  R     = alloced_ ? work_[0] : alias_->work_[0];    
  RP    = alloced_ ? work_[1] : alias_->work_[1];
  V     = alloced_ ? work_[2] : alias_->work_[2];
  T     = alloced_ ? work_[3] : alias_->work_[3];
  Q     = alloced_ ? work_[4] : alias_->work_[4];
  P     = alloced_ ? work_[5] : alias_->work_[5];
  U     = alloced_ ? work_[6] : alias_->work_[6];
  D     = alloced_ ? work_[7] : alias_->work_[7];
  BINVF = alloced_ ? work_[8] : alias_->work_[8];
  T1    = alloced_ ? work_[9] : alias_->work_[9];
  AUQ   = V;

  /* Compute initial preconditioned residual */
  //assert(pc_side == PROM_RIGHT);
  ierr = PromResidual( pc_side, zerox, X, V, T, R, BINVF, B );
  CHKERRQ(ierr);

  ierr = R->Norm2( &dp );CHKERRQ(ierr);
  r0 = dp;
  if ( !avoidnorms ) {
    ierr = Monitor(0,dp); CHKERRQ(ierr);
    conv = converged( r0, dp );
    if (conv){ *its = 0; PetscFunctionReturn(0); }
  }
  /* Make the initial Rp == R */
  ierr = RP->Copy(R);CHKERRQ(ierr);
  
  /* Set the initial conditions */
  etaold = 0.0;
  psiold = 0.0;
  tau    = dp;
  dpold  = dp;
  
  ierr = R->Dot(RP,&rhoold);CHKERRQ(ierr);       /* rhoold = (r,rp)     */
  ierr = U->Copy(R);CHKERRQ(ierr);
  ierr = P->Copy(R);CHKERRQ(ierr);
  ierr = PromPCApplyBAorAB( Amat, pc, pc_side,P,V,T);CHKERRQ(ierr);
  ierr = D->Set( 0.0 ); CHKERRQ(ierr);

  for (it=0; it<maxit; it++) {
    ierr = V->Dot(RP,&s);CHKERRQ(ierr);          /* s <- (v,rp)          */
    a = rhoold / s;                              /* a <- rho / s         */
    Q->WAXPY(-a,V,U);CHKERRQ(ierr);              /* q <- u - a v         */
    ierr = T->WAXPY(1.0,U,Q);  CHKERRQ(ierr);    /* t <- u + q           */
    ierr = PromPCApplyBAorAB( Amat, pc, pc_side,T,AUQ,T1);CHKERRQ(ierr);
    ierr = R->AXPY(-a,AUQ);CHKERRQ(ierr);      /* r <- r - a K (u + q) */
    ierr = R->Norm2( &dp );CHKERRQ(ierr);
    for (m=0; m<2; m++) {
      if (!m) {
        w = sqrt(dp*dpold);
      } else {
        w = dp;
      }
      psi = w / tau;
      cm  = 1.0 / sqrt(1.0 + psi * psi);
      tau = tau * psi * cm;
      eta = cm * cm * a;
      cf  = psiold * psiold * etaold / a;
      if (!m) {
        ierr = D->AYPX(cf,U);CHKERRQ(ierr);
      } else {
	ierr = D->AYPX(cf,Q);CHKERRQ(ierr);
      }
      ierr = X->AXPY(eta,D);CHKERRQ(ierr);
      
      dpest = sqrt(m + 1.0) * tau;
      if ( !avoidnorms && m == 0 ) {
	conv = converged(r0,dpest);
	Monitor(it+1,dpest,conv!=0);
	if (conv!=0) { it++;  break; }
      }
      etaold = eta;
      psiold = psi;
    }
    if ( !avoidnorms && conv  ) break;
    
    ierr = R->Dot(RP,&rho);CHKERRQ(ierr);        /* rho <- (r,rp)       */
    b = rho / rhoold;                               /* b <- rho / rhoold   */
    ierr = U->WAXPY(b,Q,R);CHKERRQ(ierr);       /* u <- r + b q        */
    ierr = Q->AXPY(b,P);CHKERRQ(ierr);
    ierr = P->WAXPY(b,Q,U);CHKERRQ(ierr);       /* p <- u + b(q + b p) */
    ierr = PromPCApplyBAorAB(Amat,pc,pc_side,P,V,Q);CHKERRQ(ierr);/* v <- K p */
    
    rhoold = rho;
    dpold  = dp;
  }
  if( !avoidnorms && (conv == -1 || conv == 0) ) *its = -it;
  else *its = it;
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "SolveSYMMLQ"
/* PromSolver::SolveSYMMLQ ***************************************************
 *
 *  This code implements the SYMMLQ method. (from PETSc)
 *  Reference: Paige & Saunders, 1975.
 *
 */
int PromSolver::SolveSYMMLQ( const PromVector_base * const bb,
			    PromVector_base * const xx, int *its,
			    const bool zerox /*=TRUE*/, 
			    const bool avoidnorms /*=FALSE*/, 
			    const PromPCSide ) 
{
  double alpha,beta,betaold,beta1,ceta=0.0,ceta_oold = 0.0;
  double ceta_old = 0.0,ceta_bar,dp=0.0;
  double c=1.0,cold=1.0,s=0.0,sold=0.0,coold,soold,rho0,rho1,rho2,rho3;
  double np,r0,s_prod;
  int ierr,conv,it,maxit=maxits_;
  PromVector_base *X,*R,*Z,*U,*V,*W,*UOLD,*VOLD,*Wbar;
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const      pc = alloced_ ? pc_ : alias_->pc_;
  assert((alloced_ ? work_ : alias_->work_) != NULL);

  PetscFunctionBegin;
  X     = xx;
  const PromVector_base * const B = bb;
  if(nwork_!=8)SETERRQ1(PETSC_COMM_SELF,-1,"nwork(%d) != 8",nwork_);
  R     = alloced_ ? work_[0] : alias_->work_[0];   
  Z    = alloced_ ? work_[1] : alias_->work_[1];
  U     = alloced_ ? work_[2] : alias_->work_[2];
  V     = alloced_ ? work_[3] : alias_->work_[3];
  W     = alloced_ ? work_[4] : alias_->work_[4];
  UOLD     = alloced_ ? work_[5] : alias_->work_[5];
  VOLD     = alloced_ ? work_[6] : alias_->work_[6];
  Wbar     = alloced_ ? work_[7] : alias_->work_[7];

#if defined(PETSC_USE_BOPT_g)
  ierr = TestSymmMA( Amat, pc, Z, U, R ); CHKERRQ(ierr);
#endif

  ierr = UOLD->Set(0.0);CHKERRQ(ierr);            /* u_old <- zeros;  */
  ierr = VOLD->Copy(UOLD);CHKERRQ(ierr);          /* v_old <- u_old;  */
  ierr = W->Copy(UOLD);CHKERRQ(ierr);             /* w     <- u_old;  */ 
  ierr = Wbar->Copy(UOLD);CHKERRQ(ierr);          /* w_bar <- u_old;  */

  if (!zerox) {
    ierr = KSPMult( Amat,  X, R ); CHKERRQ(ierr);    /*   r <- b - Ax       */
    ierr = R->AYPX( -1.0, B ); CHKERRQ(ierr);
  } else { 
    ierr = R->Copy(B);CHKERRQ(ierr);            /*    r <- b (x is 0)  */
    ierr = X->Set( 0.0 );    CHKERRQ(ierr);
  }
  
  ierr = pc->Apply( R, Z );      CHKERRQ( ierr ); /*      z <- B*r;    */
  ierr = R->Dot( Z, &dp );       CHKERRQ(ierr);   /* dp = r'*z;      */
  if ( fabs(dp) < 1.e-99 ) {
    PetscPrintf(Amat->MPIComm(),"Detected happy breakdown %g tol=%g\n",
		fabs(dp),1.e-99);
    *its = 0;
    PetscFunctionReturn(0);
  }
  dp = sqrt(dp); 
  beta = dp;                         /*  beta <- sqrt(r'*z)  */
  beta1 = beta;
  s_prod = fabs(beta1); 

  ierr = V->Copy(R);CHKERRQ(ierr);  /* v <- r; */
  ierr = U->Copy(Z);CHKERRQ(ierr);  /* u <- z; */
  ierr = V->Scale(1.0 / beta);  CHKERRQ(ierr);     /* v <- ibeta*v; */
  ierr = U->Scale(1.0 / beta);  CHKERRQ(ierr);     /* u <- ibeta*u; */
  ierr = Wbar->Copy(U); CHKERRQ(ierr);        /* w_bar <- u;   */
  
  if ( !avoidnorms ) {
    ierr = Z->Norm2( &np );   CHKERRQ(ierr);      /*   np <- ||z||        */
    ierr = Monitor( 0, np );  CHKERRQ(ierr); r0 = np;
    conv = converged( r0, np );
    if (conv){ *its = 0; PetscFunctionReturn(0); }
  }

  for (it=0; it<maxit; it++) {
    /*    Update    */
    if( it > 0 ){
      ierr = VOLD->Copy(V);CHKERRQ(ierr);  /* v_old <- v; */     
      ierr = UOLD->Copy(U);CHKERRQ(ierr);  /* u_old <- u; */
      
      ierr = V->Copy(R);CHKERRQ(ierr);
      ierr = V->Scale(1.0 / beta);CHKERRQ(ierr); /* v <- ibeta*r; */
      ierr = U->Copy(Z);CHKERRQ(ierr);
      ierr = U->Scale(1.0 / beta);CHKERRQ(ierr); /* u <- ibeta*z; */
      
      ierr = W->Copy(Wbar);CHKERRQ(ierr);
      ierr = W->Scale(c);CHKERRQ(ierr);
      ierr = W->AXPY(s,U);CHKERRQ(ierr);   /* w  <- c*w_bar + s*u;    (w_k) */
      ierr = Wbar->Scale(-s);CHKERRQ(ierr);
      ierr = Wbar->AXPY(c,U);CHKERRQ(ierr); /* w_bar <- -s*w_bar + c*u; (w_bar_(k+1)) */
      ierr = X->AXPY(ceta,W);CHKERRQ(ierr); /* x <- x + ceta * w;   (xL_k)  */
      
      ceta_oold = ceta_old;
      ceta_old  = ceta;
    }
    
    /*   Lanczos  */
    ierr = KSPMult( Amat, U,R);CHKERRQ(ierr);       /*  r     <- Amat*u; */  
    ierr = U->Dot(R,&alpha);CHKERRQ(ierr);      /*  alpha <- u'*r;   */
    ierr = pc->Apply( R, Z  ); CHKERRQ( ierr ); /*      z <- B*r;    */
    ierr = R->AXPY(-alpha,V);CHKERRQ(ierr);     /*  r <- r - alpha* v;  */
    ierr = Z->AXPY(-alpha,U);CHKERRQ(ierr);     /*  z <- z - alpha* u;  */
    ierr = R->AXPY(-beta,VOLD);CHKERRQ(ierr);   /*  r <- r - beta * v_old; */
    ierr = Z->AXPY(-beta,UOLD);CHKERRQ(ierr);   /*  z <- z - beta * u_old; */
    betaold = beta;                                /* beta_k             */
    ierr = R->Dot(Z,&dp);CHKERRQ(ierr);          /* dp <- r'*z;             */
    if ( fabs(dp) < 1.e-99 ) {
      PetscPrintf(Amat->MPIComm(),"Detected happy breakdown %g tol=%g\n",
		 fabs(dp),1.e-99);
      dp = 0.0;
    }
    if (dp < 0.0) SETERRQ1(PETSC_COMM_SELF,1,"Indefinite preconditioner(%e)",dp);
    beta = sqrt(dp);                    /*  beta = sqrt(dp); */

    /*    QR factorization    */
    coold = cold; cold = c; soold = sold; sold = s;
    rho0 = cold * alpha - coold * sold * betaold;    /* gamma_bar */ 
    rho1 = sqrt(rho0*rho0 + beta*beta);   /* gamma     */
    rho2 = sold * alpha + coold * cold * betaold;    /* delta     */
    rho3 = soold * betaold;                          /* epsilon   */
    
    /* Givens rotation: [c -s; s c] (different from the Reference!) */
    c = rho0 / rho1; s = beta / rho1;
    
    if (it==0){
      ceta = beta1/rho1;
    } else {
      ceta = -(rho2*ceta_old + rho3*ceta_oold)/rho1;
    }

    s_prod = s_prod*fabs(s);
    if (c == 0.0){
      np = s_prod*1.e16;
    } else {
      np = s_prod/fabs(c);       /* residual norm for xc_k (CGNORM) */
    }

    if ( !avoidnorms ) {
      conv = converged(r0,np);
      Monitor(it+1,np,conv!=0);
      if (conv) { it++;  break; }
    } 
  }

  /* move to the CG point: xc_(k+1) */
  if (c == 0.0){
    ceta_bar = ceta*1.e15;
  } else {
    ceta_bar = ceta/c;
  }
  ierr = X->AXPY(ceta_bar,Wbar);CHKERRQ(ierr); /* x <- x + ceta_bar*w_bar */
  
  if( !avoidnorms && (conv == -1 || conv == 0) ) *its = -it;
  else *its = it;
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "SolveMINRES"
/* PromSolver::SolveMINRES ***************************************************
 *
 *  This code implements the MINRES method. (from PETSc)
 *  Reference: Paige & Saunders, 1975.
 *
 */
int PromSolver::SolveMINRES( const PromVector_base * const bb,
			    PromVector_base * const xx, int *its,
			    const bool zerox /*=TRUE*/, 
			    const bool avoidnorms /*=FALSE*/, 
			    const PromPCSide ) 
{
  int     ierr,it,conv,maxit=maxits_; double r0,np;
  double  alpha,beta,betaold,eta,c=1.0,ceta,cold=1.0,coold,s=0.0;
  double  rho0,rho1,rho2,rho3,dp = 0.0,sold=0.0,soold;
  PromVector_base *X,*R,*Z,*U,*V,*W,*UOLD,*VOLD,*WOLD,*WOOLD;
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const      pc = alloced_ ? pc_ : alias_->pc_;
  assert((alloced_ ? work_ : alias_->work_) != NULL);

  PetscFunctionBegin;
  X     = xx;
  const PromVector_base * const B = bb;
  R     = alloced_ ? work_[0] : alias_->work_[0];   
  Z    = alloced_ ? work_[1] : alias_->work_[1];
  U     = alloced_ ? work_[2] : alias_->work_[2];
  V     = alloced_ ? work_[3] : alias_->work_[3];
  W     = alloced_ ? work_[4] : alias_->work_[4];
  UOLD     = alloced_ ? work_[5] : alias_->work_[5];
  VOLD     = alloced_ ? work_[6] : alias_->work_[6];
  WOLD     = alloced_ ? work_[7] : alias_->work_[7];
  WOOLD     = alloced_ ? work_[8] : alias_->work_[8];

#if defined(PETSC_USE_BOPT_g)
  ierr = TestSymmMA( Amat, pc, Z, U, R ); CHKERRQ(ierr);
#endif

  ierr = UOLD->Set(0.0);CHKERRQ(ierr);         /*     u_old  <-   0   */
  ierr = VOLD->Copy(UOLD);CHKERRQ(ierr);       /*     v_old  <-   0   */
  ierr = W->Copy(UOLD);CHKERRQ(ierr);          /*     w      <-   0   */
  ierr = WOLD->Copy(UOLD);CHKERRQ(ierr);       /*     w_old  <-   0   */
  
  if (!zerox) {
    ierr = KSPMult( Amat, X, R ); CHKERRQ(ierr);/*   r <- b - Ax       */
    ierr = R->AYPX( -1.0, B ); CHKERRQ(ierr);
  } else { 
    ierr = R->Copy(B);CHKERRQ(ierr);            /*    r <- b (x is 0)  */
    ierr = X->Set( 0.0 );    CHKERRQ(ierr);
  }

  // pc apply
  ierr = pc->Apply( R, Z );      CHKERRQ( ierr ); /*      z <- B*r;    */

  ierr = R->Dot(Z,&dp);CHKERRQ(ierr);             /* dp = r'*z;      */
  if ( fabs(dp) < 1.e-99 ) {
    PetscPrintf(Amat->MPIComm(),"Detected happy breakdown %g tol=%g\n",
		fabs(dp),1.e-99);
    *its = 0;
    PetscFunctionReturn(0);
  }

  if (dp<0.0) SETERRQ1(PETSC_COMM_SELF,-1,"Indefinite preconditioner %e (at start)",dp);
  dp   = PetscSqrtScalar(dp); 
  beta = dp;                                        /*  beta <- sqrt(r'*z  */
  eta  = beta;

  // minitor
  if ( !avoidnorms ) {
    ierr = Z->Norm2( &np );CHKERRQ(ierr);      /*   np <- ||z||        */
    Monitor(0,np); r0 = np;
    conv = converged( r0, np );
    if (conv){ *its = 0; PetscFunctionReturn(0); }
  }

  ierr = V->Copy(R);CHKERRQ(ierr);  /* v <- r; */
  ierr = U->Copy(Z);CHKERRQ(ierr);  /* u <- z; */
  ierr = V->Scale(1.0 / beta);  CHKERRQ(ierr);     /*    v <- r / beta     */
  ierr = U->Scale(1.0 / beta);  CHKERRQ(ierr);     /*    u <- z / beta     */

  for (it=1; it<maxit; it++) {

/*   Lanczos  */

     ierr = KSPMult( Amat, U, R ); CHKERRQ(ierr);/*      r <- A*u   */
     ierr = U->Dot(R,&alpha);CHKERRQ(ierr);      /*  alpha <- r'*u  */
     ierr = pc->Apply( R, Z  ); CHKERRQ(ierr);   /*      z <- B*r   */
     ierr = R->AXPY(-alpha,V);CHKERRQ(ierr);     /*  r <- r - alpha v     */
     ierr = Z->AXPY(-alpha,U);CHKERRQ(ierr);     /*  z <- z - alpha u     */
     ierr = R->AXPY(-beta,VOLD);CHKERRQ(ierr);   /*  r <- r - beta v_old  */
     ierr = Z->AXPY(-beta,UOLD);CHKERRQ(ierr);   /*  z <- z - beta u_old  */
     
     betaold = beta;
     
     ierr = R->Dot(Z,&dp);CHKERRQ(ierr); 
     if ( fabs(dp) < 1.e-99 ) {
       PetscPrintf(Amat->MPIComm(),"Detected happy breakdown %g tol=%g\n",
		   fabs(dp),1.e-99);
       dp = fabs(dp);
     }
     if (dp < 0.0) SETERRQ1(PETSC_COMM_SELF,-1,"Indefinite preconditioner(%e)",dp);
     beta = sqrt(dp);                    /*  beta = sqrt(dp); */
     
     /*    QR factorisation    */

     coold = cold; cold = c; soold = sold; sold = s;
     rho0 = cold * alpha - coold * sold * betaold;
     rho1 = PetscSqrtScalar(rho0*rho0 + beta*beta);
     rho2 = sold * alpha + coold * cold * betaold;
     rho3 = soold * betaold;

/*     Givens rotation    */

     c = rho0 / rho1;
     s = beta / rho1;

/*    Update    */

     ierr = WOOLD->Copy(WOLD);CHKERRQ(ierr);     /*  w_oold <- w_old      */
     ierr = WOLD->Copy(W);CHKERRQ(ierr);         /*  w_old  <- w          */
     
     ierr = W->Copy(U);CHKERRQ(ierr);            /*  w      <- u          */
    
     ierr = W->AXPY(-rho2, WOLD); CHKERRQ(ierr);  /*  w <- w - rho2 w_old  */
     ierr = W->AXPY(-rho3, WOOLD); CHKERRQ(ierr); /*  w <- w - rho3 w_oold */
     ierr = W->Scale(1.0/rho1);CHKERRQ(ierr);    /*  w <- w / rho1        */

     ceta = c * eta;
     ierr = X->AXPY(ceta,W);CHKERRQ(ierr);      /*  x <- x + c eta w     */ 
     eta = - s * eta;
     
     ierr = VOLD->Copy(V);CHKERRQ(ierr);
     ierr = UOLD->Copy(U);CHKERRQ(ierr);
     ierr = V->Copy(R);CHKERRQ(ierr);
     ierr = U->Copy(Z);CHKERRQ(ierr);
     ierr = V->Scale(1.0/beta);CHKERRQ(ierr);      /*  v <- r / beta       */
     ierr = U->Scale(1.0/beta);CHKERRQ(ierr);      /*  u <- z / beta       */
     
     if ( !avoidnorms ) {
       np = np * fabs(s);
       conv = converged(r0,np);
       Monitor(it,np,conv!=0);
       if (conv) { it++;  break; }
     } 
  }
  if( !avoidnorms && (conv == -1 || conv == 0) ) *its = -it;
  else *its = it;

  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "cSolveMINRES"
/* PromSolver::cSolveMINRES ***************************************************
 *
 *  This code implements the MINRES method. (from PETSc)
 *  Reference: Paige & Saunders, 1975.
 *
 */
int PromSolver::cSolveMINRES( const PromVector_base * const bb,
			      PromVector_base * const xx, int *its,
			      const bool zerox /*=TRUE*/, 
			      const bool avoidnorms /*=FALSE*/, 
			      const PromPCSide ) 
{
  PetscFunctionBegin;
#if defined(PROM_HAVE_COMPLEX)
  int     ierr,it,conv,maxit=maxits_; double r0,np;
  PromComplex  alpha,beta,betaold,eta,c=1.0,ceta,cold=1.0,coold,s=0.0;
  PromComplex  rho0,rho1,rho2,rho3,sold=0.0,soold,dp=0.0,tmp;
  PromVector_base *X_b,*R_b,*Z_b,*U_b,*V_b,*W_b,*UOLD,*VOLD,*WOLD,*WOOLD;
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const      pc = alloced_ ? pc_ : alias_->pc_;
  assert((alloced_ ? work_ : alias_->work_) != NULL);

  X_b     = xx;
  const PromVector_base * const B = bb;
  R_b     = alloced_ ? work_[0] : alias_->work_[0];   
  Z_b    = alloced_ ? work_[1] : alias_->work_[1];
  U_b     = alloced_ ? work_[2] : alias_->work_[2];
  V_b     = alloced_ ? work_[3] : alias_->work_[3];
  W_b     = alloced_ ? work_[4] : alias_->work_[4];
  UOLD     = alloced_ ? work_[5] : alias_->work_[5];
  VOLD     = alloced_ ? work_[6] : alias_->work_[6];
  WOLD     = alloced_ ? work_[7] : alias_->work_[7];
  WOOLD     = alloced_ ? work_[8] : alias_->work_[8];

  PromComplexVector *V = dynamic_cast<PromComplexVector*>(V_b);
  assert(V != NULL);
  PromComplexVector *W = dynamic_cast<PromComplexVector*>(W_b);
  assert(W != NULL);
  PromComplexVector *Z = dynamic_cast<PromComplexVector*>(Z_b);
  assert(Z != NULL);
  PromComplexVector *R = dynamic_cast<PromComplexVector*>(R_b);
  assert(R != NULL);
  PromComplexVector *X = dynamic_cast<PromComplexVector*>(X_b);
  assert(X != NULL);
  PromComplexVector *U = dynamic_cast<PromComplexVector*>(U_b);
  assert(U != NULL);

#if defined(PETSC_USE_BOPT_g)
  ierr = TestSymmMA( Amat, pc, Z, U, R ); CHKERRQ(ierr);
#endif

  ierr = UOLD->Set(0.0);   CHKERRQ(ierr);    /*     u_old  <-   0   */
  ierr = VOLD->Copy(UOLD); CHKERRQ(ierr);    /*     v_old  <-   0   */
  ierr = W->Copy(UOLD);    CHKERRQ(ierr);    /*     w      <-   0   */
  ierr = WOLD->Copy(UOLD); CHKERRQ(ierr);    /*     w_old  <-   0   */
  
  if (!zerox) {
    ierr = KSPMult( Amat, X, R ); CHKERRQ(ierr); /*   r <- b - Ax       */
    ierr = R->cAYPX( -1.0, B ); CHKERRQ(ierr);
  } else { 
    ierr = R->Copy(B);CHKERRQ(ierr);            /*    r <- b (x is 0)  */
    ierr = X->Set( 0.0 );    CHKERRQ(ierr);
  }

  // pc apply
  ierr = pc->Apply( R, Z );      CHKERRQ( ierr ); /*      z <- B*r;    */

  ierr = R->cDot( Z, dp ); CHKERRQ(ierr);        /* dp = r'*z;      */
  if ( std::abs(dp) < 1.e-99 ) {
    PetscPrintf(Amat->MPIComm(),"Detected happy breakdown %g tol=%g\n",
		std::abs(dp), 1.e-99 );
    *its = 0;
    PetscFunctionReturn(0);
  }

  dp   = std::sqrt(dp); 
  beta = dp;                                        /*  beta <- sqrt(r'*z  */
  eta  = beta;

  ierr = V->Copy(R);CHKERRQ(ierr);  /* v <- r; */
  ierr = U->Copy(Z);CHKERRQ(ierr);  /* u <- z; */
  tmp = 1.0 / beta;
  ierr = V->cScale( tmp );  CHKERRQ(ierr);     /*    v <- r / beta     */
  ierr = U->cScale( tmp );  CHKERRQ(ierr);     /*    u <- z / beta     */

  // minitor
  if ( !avoidnorms ) {
    ierr = Z->Norm2( &np );CHKERRQ(ierr);      /*   np <- ||z||        */
    Monitor( 0, np ); r0 = np;
    conv = converged( r0, np );
    if (conv){ *its = 0; PetscFunctionReturn(0); }
  }

  for (it=1; it<maxit; it++) {

/*   Lanczos  */

     ierr = KSPMult( Amat, U, R ); CHKERRQ(ierr);/*      r <- A*u   */
     ierr = U->cDot( R, alpha );CHKERRQ(ierr);      /*  alpha <- r'*u  */
     ierr = pc->Apply( R, Z  ); CHKERRQ(ierr);   /*      z <- B*r   */
     ierr = R->cAXPY(-alpha,V);CHKERRQ(ierr);     /*  r <- r - alpha v     */
     ierr = Z->cAXPY(-alpha,U);CHKERRQ(ierr);     /*  z <- z - alpha u     */
     ierr = R->cAXPY(-beta,VOLD);CHKERRQ(ierr);   /*  r <- r - beta v_old  */
     ierr = Z->cAXPY(-beta,UOLD);CHKERRQ(ierr);   /*  z <- z - beta u_old  */
     
     betaold = beta;
     
     ierr = R->cDot( Z, dp ); CHKERRQ(ierr); 
     if ( std::abs(dp) < 1.e-99 ) {
       PetscPrintf(Amat->MPIComm(),"Detected happy breakdown %g tol=%g\n",
		   std::abs(dp),1.e-99);
       dp = std::abs(dp);
     }
     beta = std::sqrt(dp);                 /*  beta = sqrt(dp); */
     
     /*    QR factorisation    */

     coold = cold; cold = c; soold = sold; sold = s;
     rho0 = cold * alpha - coold * sold * betaold;
     rho1 = std::sqrt( rho0*rho0 + beta*beta );
     rho2 = sold * alpha + coold * cold * betaold;
     rho3 = soold * betaold;

/*     Givens rotation    */

     c = rho0 / rho1;
     s = beta / rho1;

/*    Update    */

     ierr = WOOLD->Copy(WOLD);CHKERRQ(ierr);     /*  w_oold <- w_old      */
     ierr = WOLD->Copy(W);CHKERRQ(ierr);         /*  w_old  <- w          */
     
     ierr = W->Copy(U);CHKERRQ(ierr);            /*  w      <- u          */
    
     ierr = W->cAXPY(-rho2, WOLD); CHKERRQ(ierr);  /*  w <- w - rho2 w_old  */
     ierr = W->cAXPY(-rho3, WOOLD); CHKERRQ(ierr); /*  w <- w - rho3 w_oold */
     tmp = 1.0/rho1;
     ierr = W->cScale( tmp ); CHKERRQ(ierr);       /*  w <- w / rho1        */

     ceta = c * eta;
     ierr = X->cAXPY(ceta,W);CHKERRQ(ierr);        /*  x <- x + c eta w     */ 
     eta = - s * eta;
     
     ierr = VOLD->Copy(V);CHKERRQ(ierr);
     ierr = UOLD->Copy(U);CHKERRQ(ierr);
     ierr = V->Copy(R);CHKERRQ(ierr);
     ierr = U->Copy(Z);CHKERRQ(ierr);
     tmp = 1.0/beta;
     ierr = V->cScale( tmp ); CHKERRQ(ierr);      /*  v <- r / beta       */
     ierr = U->cScale( tmp ); CHKERRQ(ierr);      /*  u <- z / beta       */
     
     if ( !avoidnorms ) {
       np = np * std::abs( s );
       conv = converged( r0, np );
       Monitor( it, np, conv!=0 );
       if (conv) { it++;  break; }
     } 
  }
  if( !avoidnorms && (conv == -1 || conv == 0) ) *its = -it;
  else *its = it;
#endif
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "SolveBCGS"
/* PromSolver::SolveBCGS ***************************************************
 *  
 */
int PromSolver::SolveBCGS( const PromVector_base * const bb,
			   PromVector_base * const xx, int *its,
			   const bool zerox /*=TRUE*/, 
			   const bool avoidnorms /*=FALSE*/, 
			   const PromPCSide pc_side /*= PROM_RIGHT*/) 
{
  int         it,maxit,ierr,conv;
  double      rho,rhoold,alpha,beta,omega,omegaold,d1,d2,tmp;
  PromVector_base *X,*V,*P,*R,*RP,*T,*S,*BINVF;
  double      dp = 0.0,r0;
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const       pc = alloced_ ? pc_ : alias_->pc_;
  assert((alloced_ ? work_ : alias_->work_) != NULL);

  PetscFunctionBegin;
  maxit   = maxits_;
  X       = xx;
  const PromVector_base * const B = bb;
  R       = alloced_ ? work_[0] : alias_->work_[0];
  RP      = alloced_ ? work_[1] : alias_->work_[1];
  V       = alloced_ ? work_[2] : alias_->work_[2];
  T       = alloced_ ? work_[3] : alias_->work_[3];
  S       = alloced_ ? work_[4] : alias_->work_[4];
  P       = alloced_ ? work_[5] : alias_->work_[5];
  BINVF   = alloced_ ? work_[6] : alias_->work_[6];

  /* Compute initial preconditioned residual */
  assert(pc_side == PROM_RIGHT);
  ierr = PromResidual( pc_side, zerox, X, V, T, R, BINVF, B );
  CHKERRQ(ierr);

  /* Test for nothing to do */
  if (!avoidnorms) {
    ierr = R->Norm2( &dp ); CHKERRQ(ierr);
    r0 = dp;
    ierr = Monitor(0,dp); CHKERRQ(ierr);
    conv = converged( r0, dp );
    if (conv != 0){ *its = 0; PetscFunctionReturn(0); }
  }

  /* Make the initial Rp == R */
  ierr = RP->Copy(R);CHKERRQ(ierr);
  rhoold   = 1.0;
  alpha    = 1.0;
  omegaold = 1.0;
  ierr = P->Set( 0.0 );CHKERRQ(ierr);
  ierr = V->Set( 0.0 );   CHKERRQ(ierr);
  for (it=0; it<maxit; it++) {
    ierr = R->Dot(RP,&rho);CHKERRQ(ierr);       /*   rho <- (r,rp)      */
    if (rho == 0.0)SETERRQ(PETSC_COMM_SELF,-1,"Breakdown, rho = r . rp = 0");
    beta = (rho/rhoold) * (alpha/omegaold);
    tmp = -omegaold; P->AXPY( tmp, V );            /*   p <- p - w v       */
    ierr = P->AYPX( beta, R ); CHKERRQ(ierr);      /*   p <- r + p beta    */
    /*   v <- K p           */
    ierr = PromPCApplyBAorAB( Amat, pc, pc_side, P, V, T ); CHKERRQ(ierr);
    ierr = V->Dot(RP,&d1); CHKERRQ(ierr);
    alpha = rho / d1; tmp = -alpha;                 /*   a <- rho / (v,rp)  */
    ierr = S->WAXPY( tmp, V, R ); CHKERRQ(ierr);    /*   s <- r - a v       */
    /*   t <- K s    */
    ierr = PromPCApplyBAorAB( Amat, pc, pc_side, S, T, R ); CHKERRQ(ierr);
    
    ierr = T->Dot( T, &d2 ); CHKERRQ(ierr);
    if (d2 == 0.0) {
      /* t is 0.  if s is 0, then alpha v == r, and hence alpha p
	 may be our solution.  Give it a try? */
      ierr = S->Dot(S,&d1);CHKERRQ(ierr);
      if (d1 != 0.0){ SETERRQ(PETSC_COMM_SELF,-1,"Breakdown, da=s . s != 0");}
      ierr = X->AXPY(alpha,P);CHKERRQ(ierr);   /*   x <- x + a p       */
      if (!avoidnorms) Monitor(it+1,0.0,TRUE);
      conv = -1;
      break;
    }
    ierr = S->Dot( T, &d1 ); CHKERRQ(ierr);
    omega = d1 / d2;                               /*   w <- (t's) / (t't) */
    ierr = X->AXPY( alpha, P );CHKERRQ(ierr);      /*   x <- x + a p       */
    ierr = X->AXPY( omega, S );CHKERRQ(ierr);      /*   x <- x + w s       */
    tmp = -omega; 
    ierr = R->WAXPY( tmp, T, S );CHKERRQ(ierr);    /*   r <- s - w t       */
    if (!avoidnorms) {
      ierr = R->Norm2( &dp );CHKERRQ(ierr);
      conv = converged(r0,dp);
      Monitor(it+1,dp,conv!=0);
      if (conv) { it++;  break; } 
    }
    rhoold   = rho;
    omegaold = omega;
  }
  if( !avoidnorms && (conv == -1 || conv == 0) ) *its = -it;
  else *its = it;

  // Unwind Preconditioner
  if( pc_side == PROM_RIGHT ) {
    ierr = pc->Apply( X, T );CHKERRQ( ierr );
    ierr = X->Copy( T );CHKERRQ(ierr);
  } 

  PetscFunctionReturn(0); 
}

#undef __FUNCT__  
#define __FUNCT__ "cSolveBCGS"
/* PromSolver::cSolveBCGS ***************************************************
 *  
 */
int PromSolver::cSolveBCGS( const PromVector_base * const bb,
			   PromVector_base * const xx, int *its,
			   const bool zerox /*=TRUE*/, 
			   const bool avoidnorms /*=FALSE*/, 
			   const PromPCSide pc_side /*= PROM_RIGHT*/) 
{
  PetscFunctionBegin;
#if defined(PROM_HAVE_COMPLEX)
  int         it,maxit,ierr,conv;
  PromComplex rho,rhoold,alpha,beta,omega,omegaold,d1,d2,tmp;
  PromVector_base *X_b,*V_b,*P_b,*R_b,*RP_b,*T_b,*S_b,*BINVF;
  double      dp = 0.0, r0;
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const       pc = alloced_ ? pc_ : alias_->pc_;
  assert((alloced_ ? work_ : alias_->work_) != NULL);

  maxit     = maxits_;
  X_b       = xx;
  const PromVector_base * const B = bb;
  R_b       = alloced_ ? work_[0] : alias_->work_[0];
  RP_b      = alloced_ ? work_[1] : alias_->work_[1];
  V_b       = alloced_ ? work_[2] : alias_->work_[2];
  T_b       = alloced_ ? work_[3] : alias_->work_[3];
  S_b       = alloced_ ? work_[4] : alias_->work_[4];
  P_b       = alloced_ ? work_[5] : alias_->work_[5];
  BINVF   = alloced_ ? work_[6] : alias_->work_[6];

  PromComplexVector *V = dynamic_cast<PromComplexVector*>(V_b);
  assert(V != NULL);
  PromComplexVector *T = dynamic_cast<PromComplexVector*>(T_b);
  assert(T != NULL);
  PromComplexVector *S = dynamic_cast<PromComplexVector*>(S_b);
  assert(S != NULL);
  PromComplexVector *RP = dynamic_cast<PromComplexVector*>(RP_b);
  assert(RP != NULL);
  PromComplexVector *R = dynamic_cast<PromComplexVector*>(R_b);
  assert(R != NULL);
  PromComplexVector *X = dynamic_cast<PromComplexVector*>(X_b);
  assert(X != NULL);
  PromComplexVector *P = dynamic_cast<PromComplexVector*>(P_b);
  assert(P != NULL);

  /* Compute initial preconditioned residual */
  assert(pc_side == PROM_RIGHT);
  ierr = PromResidual( pc_side, zerox, X, V, T, R, BINVF, B );
  CHKERRQ(ierr);

  /* Test for nothing to do */
  if (!avoidnorms) {
    ierr = R->Norm2( &dp ); CHKERRQ(ierr);
    r0 = dp;
    ierr = Monitor(0,dp); CHKERRQ(ierr);
    conv = converged( r0, dp );
    if (conv != 0){ *its = 0; PetscFunctionReturn(0); }
  }

  /* Make the initial Rp == R */
  ierr = RP->Copy(R); CHKERRQ(ierr);
  rhoold   = 1.0;
  alpha    = 1.0;
  omegaold = 1.0;
  ierr = P->Set( 0.0 );   CHKERRQ(ierr);
  ierr = V->Set( 0.0 );   CHKERRQ(ierr);
  for ( it = 0 ; it < maxit ; it++ ) {
    ierr = R->cDot( RP, rho );  CHKERRQ(ierr);   /*   rho <- (r,rp)      */
    if(rho == 0.0)SETERRQ(PETSC_COMM_SELF,-1,"Breakdown, rho = r . rp = 0");
    beta = (rho/rhoold) * (alpha/omegaold);
    tmp = -omegaold; P->cAXPY( tmp, V );         /*   p <- p - w v       */
    ierr = P->cAYPX( beta, R ); CHKERRQ(ierr);   /*   p <- r + p beta    */
    /*   v <- K p           */
    ierr = PromPCApplyBAorAB( Amat, pc, pc_side, P, V, T ); CHKERRQ(ierr);
    ierr = V->cDot( RP, d1 ); CHKERRQ(ierr);
    alpha = rho / d1; tmp = -alpha;               /*   a <- rho / (v,rp)  */
    ierr = S->cWAXPY( tmp, V, R ); CHKERRQ(ierr); /*   s <- r - a v       */
    /*   t <- K s    */
    ierr = PromPCApplyBAorAB( Amat, pc, pc_side, S, T, R ); CHKERRQ(ierr);
    
    ierr = T->cDot( T, d2 ); CHKERRQ(ierr);
    if (d2 == 0.0) {
      /* t is 0.  if s is 0, then alpha v == r, and hence alpha p
	 may be our solution.  Give it a try? */
      ierr = S->cDot( S, d1 );      CHKERRQ(ierr);
      if (d1 != 0.0){ SETERRQ(PETSC_COMM_SELF,-1,"Breakdown, da=s . s != 0");}
      ierr = X->cAXPY(alpha,P);CHKERRQ(ierr);   /*   x <- x + a p       */
      if (!avoidnorms) Monitor(it+1,0.0,TRUE);
      conv = -1;
      break;
    }
    ierr = S->cDot( T, d1 ); CHKERRQ(ierr);
    omega = d1 / d2;                               /*   w <- (t's) / (t't) */
    ierr = X->cAXPY( alpha, P );CHKERRQ(ierr);      /*   x <- x + a p       */
    ierr = X->cAXPY( omega, S );CHKERRQ(ierr);      /*   x <- x + w s       */
    tmp = -omega; 
    ierr = R->cWAXPY( tmp, T, S );CHKERRQ(ierr);    /*   r <- s - w t       */
    if (!avoidnorms) {
      ierr = R->Norm2( &dp );CHKERRQ(ierr);
      conv = converged(r0,dp);
      Monitor(it+1,dp,conv!=0);
      if (conv) { it++;  break; } 
    }
    rhoold   = rho;
    omegaold = omega;
  }
  if( !avoidnorms && (conv == -1 || conv == 0) ) *its = -it;
  else *its = it;

  // Unwind Preconditioner
  if( pc_side == PROM_RIGHT ) {
    ierr = pc->Apply( X, T );CHKERRQ( ierr );
    ierr = X->Copy( T );CHKERRQ(ierr);
  }
#endif
  PetscFunctionReturn(0); 
}
 
////////////////////////////////////////////////////////////////////////
// GMRES
///////////////////////////////////////////////////////////////////////
#define HH(a,b)  ((alloced_?origin_.e:alias_->origin_.e) + (b)*((alloced_?nEig_:alias_->nEig_)+2) + (a))
#define HES(a,b) ((alloced_?hes_.e:alias_->hes_.e)       + (b)*((alloced_?nEig_:alias_->nEig_)+1) + (a))
#define CC(a)    ((alloced_?ee_.e:alias_->ee_.e)         + (a))
#define SS(a)    ((alloced_?dd_.e:alias_->dd_.e)         + (a))
#define GRS(a)   ((alloced_?d_.e:alias_->d_.e)           + (a))

/* vector names */
#define VEC_WORK(i)    (alloced_ ? work_[i] : alias_->work_[i])
#define VEC_TEMP       VEC_WORK(0)
#define VEC_TEMP_MATOP VEC_WORK(1)
#define VEC_BINVF      VEC_WORK(2)
#define VEC_VV(i) VEC_WORK(VEC_OFFSET + i)
#define VEC_ZZ(i) VEC_WORK(VEC_OFFSET + i + (nEig_+1) )

#if defined(PROM_HAVE_COMPLEX)
#define cHH(a,b)  ((alloced_?origin_.c:alias_->origin_.c) + (b)*((alloced_?nEig_:alias_->nEig_)+2) + (a))
#define cHES(a,b) ((alloced_?hes_.c:alias_->hes_.c)       + (b)*((alloced_?nEig_:alias_->nEig_)+1) + (a))
#define cCC(a)    ((alloced_?ee_.c : alias_->ee_.c)       + (a))
#define cSS(a)    ((alloced_?dd_.c : alias_->dd_.c)       + (a))
#define cGRS(a)   ((alloced_?d_.c  : alias_->d_.c)        + (a))
#endif

#undef __FUNCT__
#define __FUNCT__ "SolveGMRES"
/* PromSolver::SolveGMRES ***************************************************
 *  
 *  - right preconditioned GMRES
 */
int PromSolver::SolveGMRES( const PromVector_base * const bb,
			    PromVector_base * const xx, int *its,
			    const bool zerox /*=TRUE*/,
			    const bool avoidnorms /*=FALSE*/,
			    const PromPCSide pc_side /*= PROM_RIGHT*/ )
{
  int         ierr,it,reason; double rnorm0 = 0.0;
  assert((alloced_ ? work_ : alias_->work_) != NULL);
  PromVector_base *vv_0 = VEC_VV(0);  // a global RHS (yuck!)
  PromVector_base *binvf = VEC_BINVF;
  PromPC_base * const pc = alloced_ ? pc_ : alias_->pc_;
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;

  PetscFunctionBegin;
  /* Save binv*f -- assumes PROM_RIGHT */
  if( pc_side == PROM_LEFT ) {
    ierr = pc->Apply( bb, binvf ); CHKERRQ(ierr);
  }
  else { //
    ierr = binvf->Copy( bb ); CHKERRQ(ierr); // binvf <- bb
  }
  /* Compute the initial (preconditioned) residual (into 'vv_0')*/
  if ( !zerox ) {
    ierr = ResidualGMRES( pc_side, xx, vv_0, binvf, VEC_TEMP ); CHKERRQ(ierr);
  }
  else {
    ierr = vv_0->Copy( binvf ); CHKERRQ(ierr);
    ierr = xx->Set( 0.0 );    CHKERRQ(ierr);
  }
  // doit
  it  = 0; reason = 0;  
  ierr = CycleGMRES( pc_side, bb, xx, &reason, &it, &rnorm0, avoidnorms);
  CHKERRQ(ierr);
  while( reason==0 && it < maxits_ ) {
    ierr = ResidualGMRES( pc_side, xx, vv_0, binvf, VEC_TEMP ); CHKERRQ(ierr);
    if( !avoidnorms && verbose_>2 ) PetscPrintf(Amat->MPIComm(),"*");
    ierr = CycleGMRES( pc_side, bb, xx, &reason, &it, &rnorm0, avoidnorms);
    CHKERRQ(ierr);
    if ( it >= maxits_ ) break; 
  }

  /* mark lack of convergence  */
  if( it >= maxits_ || reason == 0 ) *its = -it;
  else *its = it;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "CycleGMRES"
/* PromSolver::CycleGMRES ***************************************************
 *  
 */
int PromSolver::CycleGMRES( const PromPCSide pc_side, 
			    const PromVector_base * const,
			    PromVector_base * const xx,
			    int *reason, int *itcount, double *rnorm0, 
	 		    const bool avoidnorms )
{
  double      res_norm,res,hapbnd,tt;
  int         ierr,it,conv, max_it = maxits_; assert(max_it>0);
  int         max_k = (alloced_ ? nEig_ : alias_->nEig_); assert(max_k>0);
  bool        hapend = FALSE;
  PromVector_base *vv_0 = VEC_VV(0);
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const pc = alloced_ ? pc_ : alias_->pc_; 

  PetscFunctionBegin;
  /* scale VEC_VV (the initial residual) */
  ierr = vv_0->Normalize( &res_norm ); CHKERRQ(ierr);
  res  = res_norm;

  if( !Amat->isComplex() ) *GRS(0) = res_norm;
  else {
#if defined(PROM_HAVE_COMPLEX)
    *cGRS(0) = res_norm;
#else
    SETERRQ(PETSC_COMM_SELF,0,"????");
#endif
  }
  /* check for the convergence */
  if( res == 0. ) {
    *reason = 1; // 1 == converged, 
    return 0;
  } 
  if( *itcount == 0 ) *rnorm0 = res;
  conv = converged( *rnorm0, res ); CHKERRQ(ierr);
  *reason = conv;
  it = 0;
  while( (*reason == 0 || avoidnorms) && it < max_k && *itcount < maxits_ ) {
    if( !avoidnorms ) Monitor( *itcount, res ); 
    const PromVector_base *const vv_it = VEC_VV(it);  
    PromVector_base *vv_it1 = VEC_VV(it+1);

    // apply AB
    {
      PromVector_base *const Mb = (pc_side==PROM_FLEX) ? VEC_ZZ(it) : VEC_TEMP;
      ierr = PromPCApplyBAorAB( Amat, pc, pc_side, vv_it, vv_it1, Mb );
      CHKERRQ(ierr);
    }
    /* update hessenberg matrix and do Gram-Schmidt */
    if( !Amat->isComplex() ) { 
      ierr = TwoUnmodifiedGramSchmidtOrthogonalization(it);CHKERRQ(ierr);
    }
#if defined(PROM_HAVE_COMPLEX)
    else {
      ierr = TwoUnmodifiedGramSchmidtOrthogonalization_complex(it);CHKERRQ(ierr);
    }
#endif
    /* vv(i+1) . vv(i+1) */
    ierr = vv_it1->Normalize( &tt ); CHKERRQ(ierr);
    /* save the magnitude */
    if( !Amat->isComplex() ) { *HH(it+1,it) = tt; *HES(it+1,it) = tt; }
#if defined(PROM_HAVE_COMPLEX)
    else{ *cHH(it+1,it) = tt; *cHES(it+1,it) = tt; }
#endif
    /* check for the happy breakdown */
    if( !Amat->isComplex() ) {
      if( *GRS(it) == 0.0 ) SETERRQ1(PETSC_COMM_SELF,0,"*GRS(%d)==0.0",it);
      hapbnd  = fabs( tt / *GRS(it) );
    }
#if defined(PROM_HAVE_COMPLEX)
    else{
      if( *cGRS(it) == 0.0 ) SETERRQ1(PETSC_COMM_SELF,0,"*cGRS(%d)==0.0",it);
      hapbnd = std::abs( (double)tt / *cGRS(it) );
    }
#endif
    if (hapbnd > 1.e-99) hapbnd = 1.e-99; // hard wired happy tol!!!
    if (tt < hapbnd) {
      if( verbose_>2  || tt == 0. ) 
	PetscPrintf(Amat->MPIComm(),
		    "Detected happy breakdown, %d) current hapbnd = %g vv(i+1).vv(i+1)=%g N=%d\n",
		    it,hapbnd,tt,Amat->getM());
      hapend = TRUE;
    }

    if( !Amat->isComplex() ) ierr = UpdateGMRESHessenberg( it, hapend, &res );
#if defined(PROM_HAVE_COMPLEX)
    else ierr = UpdateGMRESHessenberg_complex( it, hapend, &res );
#endif
    CHKERRQ(ierr);

    // increment
    it++; (*itcount)++;
    conv = converged( *rnorm0, res ); 
    *reason = conv;
    /* Catch error in happy breakdown and signal convergence and break from loop */
    if( hapend ) {
      if ( conv == 0 ) {
	SETERRQ1(PETSC_COMM_SELF,0,"You reached the happy break down, but convergence was not indicated. Residual norm=%e",res);
      }
      break;
    }
  }
  /* Monitor if we know that we will not return for a restart */
  if(!avoidnorms && (*reason!=0 || *itcount>=max_it)) Monitor(*itcount,res,TRUE);
  else if ( avoidnorms ) *reason = 1; // converged by def
  /*
    Down here we have to solve for the "best" coefficients of the Krylov
    columns, add the solution values together, and possibly unwind the
    preconditioning from the solution
  */
  /* Form the solution (or the solution so far) */
  const PromVector_base *const*vv= &((pc_side==PROM_FLEX)?VEC_ZZ(0) : VEC_VV(0));
  if( !Amat->isComplex() ) {
    ierr = BuildGMRESSoln( pc_side, GRS(0), xx, it-1, vv, VEC_TEMP );
    CHKERRQ(ierr);
  }
#if defined(PROM_HAVE_COMPLEX)
  else {
    ierr = BuildGMRESSoln_complex( pc_side, cGRS(0), xx, it-1, vv, VEC_TEMP );
    CHKERRQ(ierr);
  }
#endif
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "ResidualGMRES"
/* PromSolver::ResidualGMRES ***************************************************
 *  
 * - B is in 'binv'
 * - puts B - A * xx into vv(0)
 * 
 *  pc_side: flag for side of PC (left,right,symm,flex)
 *  xx: init guess (!= 0)
 *  vv: (out)
 *  binvf: RHS
 *  temp: scratch space
 *
 */
int PromSolver::ResidualGMRES( const PromPCSide pc_side, 
			       const PromVector_base * const xx,
			       PromVector_base * const vv,
			       const PromVector_base * const binvf,
			       PromVector_base * const temp )
{
  int ierr;
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;

  PetscFunctionBegin;  
  if( pc_side == PROM_LEFT ) {
    PromPC_base * const       pc = alloced_ ? pc_ : alias_->pc_;
    // t <- BAx
    ierr = PromPCApplyBAorAB(Amat, pc, PROM_LEFT, xx, temp, vv); CHKERRQ(ierr);
  }
  else {
    ierr = KSPMult( Amat, xx, temp ); CHKERRQ(ierr);   /* t <- Ax */
  }
  ierr = vv->Copy( binvf );CHKERRQ(ierr);     
  ierr = vv->AXPY( -1.0, temp );CHKERRQ(ierr); /* binvf - A(B)x */

  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "PromResidual"
/* PromResidual ***************************************************
 *  
 */
int PromSolver::PromResidual( const PromPCSide pc_side, const bool zerox,
			      PromVector_base *vsoln, PromVector_base *vt1,
			      PromVector_base *vt2, PromVector_base *vres,
			      PromVector_base *vbinvf, 
			      const PromVector_base * const vb )
{
  int           ierr;
  const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;
  PromPC_base * const       pc = alloced_ ? pc_ : alias_->pc_;

  PetscFunctionBegin;
  if( vb != vbinvf ) {
    if( pc_side == PROM_RIGHT ) { ierr = vbinvf->Copy( vb ); CHKERRQ(ierr); }
    else { ierr = pc->Apply( vb, vbinvf ); CHKERRQ(ierr); } 
  }
  if ( !zerox ) {
    /* compute initial residual: f - M*x */
    /* (inv(b)*a)*x or (a*inv(b)*b)*x into dest */
    if( pc_side == PROM_RIGHT ) {
      /* we want a * binv * b * x, or just a * x for the first step */
      /* a*x into temp */
      ierr = KSPMult( Amat,  vsoln, vt1 );CHKERRQ(ierr);
    } else {
      /* else we do binv * a * x */
      ierr = KSPMult( Amat, vsoln, vt2 );CHKERRQ(ierr);
      ierr = pc->Apply( vt2, vt1 ); CHKERRQ(ierr);
    }
    /* This is an extra copy for the right-inverse case */
    if( vb != vbinvf ) ierr = vres->Copy( vbinvf );CHKERRQ(ierr);
    ierr = vres->AXPY( -1.0, vt1 );CHKERRQ(ierr); /*inv(b)(f - a*x) into dest*/
  } 
  else if( vb != vbinvf ) {
    ierr = vres->Copy( vbinvf );CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

/* real version of ...
#include <complex>

#define PetscRealPart(a)        (a).real()
#define PetscImaginaryPart(a)   (a).imag()
#define PetscAbsScalar(a)   std::abs(a)
#define PetscConj(a)        std::conj(a)
#define PetscSqrtScalar(a)  std::sqrt(a)
#define PetscPowScalar(a,b) std::pow(a,b)
#define PetscExpScalar(a)   std::exp(a)
#define PetscSinScalar(a)   std::sin(a)
#define PetscCosScalar(a)   std::cos(a)

typedef std::complex<double> PetscScalar;
*/

#undef __FUNCT__  
#define __FUNCT__ "BuildGMRESSoln"
/* PromSolver::BuildGMRESSoln ******************************************
 *  
 */
int PromSolver::BuildGMRESSoln( const PromPCSide pc_side, double* nrs,
				PromVector_base *xx, const int it,
				const PromVector_base *const*vv_0,
				PromVector_base * const temp )
{
  double    tt;
  int       ierr,ii,k,j;
  PromPC_base * const pc = alloced_ ? pc_ : alias_->pc_;
  //const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;

  PetscFunctionBegin;
  /* Solve for solution vector that minimizes the residual */

  /* If it is < 0, no gmres steps have been performed */
  if (it < 0) {
    PetscFunctionReturn(0);
  }
  if (*HH(it,it) == 0.0) SETERRQ1(PETSC_COMM_SELF,-1,"HH(it,it) is identically zero; GRS(it) = %g",*GRS(it));
  if (*HH(it,it) != 0.0) {
    nrs[it] = *GRS(it) / *HH(it,it);
  } 
  else {
    nrs[it] = 0.0;
  }
  
  for (ii=1; ii<=it; ii++) {
    k   = it - ii;
    tt  = *GRS(k);
    for (j=k+1; j<=it; j++) tt  = tt - *HH(k,j) * nrs[j];
    nrs[k]   = tt / *HH(k,k);
  }
  
  /* Accumulate the correction to the solution of the preconditioned problem in TEMP */
  ierr = temp->Set( 0.0 ); CHKERRQ(ierr);
  ierr = temp->MAXPY( it+1, nrs, vv_0 );CHKERRQ(ierr);
  /* unwind pc */
  if( pc_side == PROM_RIGHT ) {
    /*If we preconditioned on the right, we need to solve for the correction to
      the unpreconditioned problem */
    PromVector_base *temp_matop = VEC_TEMP_MATOP;
    ierr = pc->Apply( temp, temp_matop );CHKERRQ(ierr);
    ierr = xx->AYPX( 1.0, temp_matop );CHKERRQ(ierr);
  }
  else {
    ierr = xx->AYPX( 1.0, temp );CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

/*
  Do the scalar work for the orthogonalization.  Return new residual.
*/
#undef __FUNCT__  
#define __FUNCT__ "UpdateGMRESHessenberg"
/* PromSolver::UpdateGMRESHessenberg *****************************************
 *
 *   INPUT:
 *     - it:
 *     - hapend: good breakdown?
 *     - res: residual (out)
 *
 *   SIDE EFFECTS:
 *     - sets 'nwork_' and allocs 'work_'
 *
 *   RETURN:
 *     - PETSc error code
 */ 
int PromSolver::UpdateGMRESHessenberg( const int it, bool hapend, double *res )
{
  double *hh,*cc,*ss,tt;
  int     j;
  //const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;

  PetscFunctionBegin;
  hh  = HH(0,it);
  cc  = CC(0);
  ss  = SS(0);
  
  /* Apply all the previously computed plane rotations to the new column
     of the Hessenberg matrix */
  for (j=1; j<=it; j++) {
    tt  = *hh;
    *hh = *cc * tt + *ss * *(hh+1);
    hh++;
    *hh = *cc++ * *hh - (*ss++ * tt);
  }

  /*
    compute the new plane rotation, and apply it to:
     1) the right-hand-side of the Hessenberg system
     2) the new column of the Hessenberg matrix
     thus obtaining the updated value of the residual
  */
  if( !hapend ) {
    tt = PetscSqrtScalar( *hh * *hh + *(hh+1) * *(hh+1) );
    if (tt == 0.0) SETERRQ(PETSC_COMM_SELF,-1,"Your matrix or preconditioner is the null operator");
    *cc       = *hh / tt;
    *ss       = *(hh+1) / tt;
    *GRS(it+1) = - (*ss * *GRS(it));
    *GRS(it)   = *cc * *GRS(it);
    *hh       = *cc * *hh + *ss * *(hh+1);
    *res      = fabs( *GRS(it+1) );
  } 
  else {
    /* happy breakdown: HH(it+1, it) = 0, therfore we don't need to apply 
       another rotation matrix (so RH doesn't change).  The new residual is 
       always the new sine term times the residual from last time (GRS(it)), 
       but now the new sine rotation would be zero...so the residual should
       be zero...so we will multiply "zero" by the last residual.  This might
       not be exactly what we want to do here -could just return "zero". */
    
    *res = 0.0;
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "ModifiedGramSchmidtOrthogonalization"
/*
    This is the basic orthogonalization routine using modified Gram-Schmidt.
*/
int PromSolver::ModifiedGramSchmidtOrthogonalization( const int it )
{
  
  int       ierr,j;
  double    *hh,*hes,tmp;
  PromVector_base *vv_1 = VEC_VV(it+1), *vv_j;
  //const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;

  PetscFunctionBegin;
  /* update Hessenberg matrix and do Gram-Schmidt */
  hh  = HH(0,it);
  hes = HES(0,it);
  for (j=0; j<=it; j++) {
    /* (vv(it+1), vv(j)) */
    vv_j = VEC_VV(j); ierr = vv_1->Dot( vv_j, hh );CHKERRQ(ierr);
    *hes++ = *hh;
    /* vv(it+1) <- vv(it+1) - hh[it+1][j] vv(j) */
    tmp    = - (*hh++);  
    ierr   = vv_1->AXPY( tmp, vv_j );CHKERRQ(ierr);
  } 
  PetscFunctionReturn(0);
}

/*
  This version uses UNMODIFIED Gram-Schmidt.  It is NOT always recommended, 
  but it can give MUCH better performance than the default modified form
  when running in a parallel environment.
 */
#undef __FUNCT__  
#define __FUNCT__ "UnmodifiedGramSchmidtOrthogonalization"
int PromSolver::UnmodifiedGramSchmidtOrthogonalization( const int it )
{
  int       j,ierr;
  double    *hh,*hes;
  PromVector_base *vv_1 = VEC_VV(it+1);
  const PromVector_base *const*vv_0 = &(VEC_VV(0));
  //const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;

  PetscFunctionBegin;
  /* update Hessenberg matrix and do unmodified Gram-Schmidt */
  hh  = HH(0,it);
  hes = HES(0,it);
  /* 
     This is really a matrix-vector product, with the matrix stored
     as pointer to rows 
     */
  ierr = vv_1->MDot( it+1, vv_0, hes );CHKERRQ(ierr);

  /*
    This is really a matrix-vector product: 
    [h[0],h[1],...]*[ v[0]; v[1]; ...] subtracted from v[it+1].
  */
  
  for (j=0; j<=it; j++) hh[j] = -hes[j]; 
  ierr = vv_1->MAXPY( it+1, hh, vv_0 );CHKERRQ(ierr);
  for (j=0; j<=it; j++) hh[j] = -hh[j];
  PetscFunctionReturn(0);
}

/*
  uses 1 iteration of iterative refinement of UNMODIFIED Gram-Schmidt.  
  It can give better performance when running in a parallel 
  environment and in some cases even in a sequential environment (because
  MAXPY has more data reuse).
  
  Care is taken to accumulate the updated HH/HES values.
 */
#undef __FUNCT__  
#define __FUNCT__ "TwoUnmodifiedGramSchmidtOrthogonalization"
int PromSolver::TwoUnmodifiedGramSchmidtOrthogonalization( const int it )
{
  int       j,ierr,ncnt;
  double    *hh,*hes,fbuf[300],*lhh = NULL;
  PromVector_base *vv_1 = VEC_VV(it+1);
  const PromVector_base *const* vv_0 = &(VEC_VV(0));
  //const PromMatrix_base * const Amat = alloced_ ? A_ : alias_->A_;

  /* Don't allocate small arrays */
  if (it+1 < 300) lhh = fbuf;
  else {
    ierr = PetscMalloc((it+1) * sizeof(double),&lhh);CHKERRQ(ierr);
  }
  
  /* update Hessenberg matrix and do unmodified Gram-Schmidt */
  hh  = HH(0,it);
  hes = HES(0,it);

  /* Clear hh and hes since we will accumulate values into them */
  for (j=0; j<=it; j++) {
    hh[j]  = 0.0;
    hes[j] = 0.0;
  }

  for( ncnt = 0 ; ncnt < 2 ; ncnt++ ) {
    /* 
       This is really a matrix-vector product, with the matrix stored
       as pointer to rows 
    */
    perf_mon_.EventBegin(PromContext::prom_logs_[LOG_AUX1]);
    ierr = vv_1->MDot( it+1, vv_0, lhh ); CHKERRQ(ierr); /* <v,vnew> */
    perf_mon_.EventEnd(PromContext::prom_logs_[LOG_AUX1]);
    /*
      This is really a matrix vector product: 
      [h[0],h[1],...]*[ v[0]; v[1]; ...] subtracted from v[it+1].
    */
    for (j=0; j<=it; j++) lhh[j] = - lhh[j];
    ierr = vv_1->MAXPY( it+1, lhh, vv_0 );           CHKERRQ(ierr);
    for (j=0; j<=it; j++) {
      hh[j]  -= lhh[j];     /* hh += <v,vnew> */
      hes[j] += lhh[j];     /* hes += - <v,vnew> */
    }
    
    /*double hnrm = 0.0,wnrm;
      for (j=0; j<=it; j++) hnrm += lhh[j] * lhh[j];
      hnrm = sqrt(hnrm);
      ierr = vv_1->Norm2( &wnrm ); CHKERRQ(ierr);
      if (wnrm > 1.0286 * hnrm) break;*/
  }

  if( lhh != fbuf ) {ierr = PetscFree(lhh);CHKERRQ(ierr);}

  return 0;
}

#if defined(PROM_HAVE_COMPLEX)

/* PromSolver::BuildGMRESSoln_complex ******************************
 *  
 */
#undef __FUNCT__  
#define __FUNCT__ "BuildGMRESSoln_complex"

int PromSolver::BuildGMRESSoln_complex( const PromPCSide pc_side, PromComplex* nrs,
					PromVector_base *xx, const int it,
					const PromVector_base *const*vv_0,
					PromVector_base * const temp )
{
  PromComplex tt;
  int       ierr,ii,k,j;
  PromPC_base * const pc = alloced_ ? pc_ : alias_->pc_;

  PetscFunctionBegin;
  /* Solve for solution vector that minimizes the residual */
  
  /* If it is < 0, no gmres steps have been performed */
  if (it < 0) {
    PetscFunctionReturn(0);
  }
  if ( (tt=*cHH(it,it)) == 0.0){
    SETERRQ2(PETSC_COMM_SELF,1,"HH(it,it) is identically zero; cGRS(it) = (%g,%g)",cGRS(it)->real(),cGRS(it)->imag());
  }
  else {
    nrs[it] = *cGRS(it) / tt;
  } 

  for (ii=1; ii<=it; ii++) {
    k   = it - ii;
    tt  = *cGRS(k);
    for (j=k+1; j<=it; j++) tt  = tt - *cHH(k,j) * nrs[j];
    nrs[k]   = tt / *cHH(k,k);
  }
  
  /* Accumulate the correction to the solution of the preconditioned problem in TEMP */
  ierr = temp->Set( 0.0 ); CHKERRQ(ierr);
  PromComplexVector *ctemp = dynamic_cast<PromComplexVector*>(temp);
  assert(ctemp != NULL);
  ierr = ctemp->cMAXPY( it+1, nrs, vv_0 );CHKERRQ(ierr);
  /* unwind pc */
  if( pc_side == PROM_RIGHT ) {
    /*If we preconditioned on the right, we need to solve for the correction to
      the unpreconditioned problem */
    PromVector_base *temp_matop = VEC_TEMP_MATOP;
    ierr = pc->Apply( temp, temp_matop );CHKERRQ(ierr);
    ierr = xx->AYPX( 1.0, temp_matop );CHKERRQ(ierr);
  }
  else {
    ierr = xx->AYPX( 1.0, temp );CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

/*
  Do the scalar work for the orthogonalization.  Return new residual.
*/
#undef __FUNCT__  
#define __FUNCT__ "UpdateGMRESHessenberg_complex"
int PromSolver::UpdateGMRESHessenberg_complex( const int it, bool hapend, double *res )
{
  PromComplex *hh,*cc,*ss,tt;
  int     j;

  PetscFunctionBegin;
  hh  = cHH(0,it);
  cc  = cCC(0);
  ss  = cSS(0);
  
  /* Apply all the previously computed plane rotations to the new column
     of the Hessenberg matrix */
  for (j=1; j<=it; j++) {
    tt  = *hh;
    *hh = std::conj(*cc) * tt + *ss * *(hh+1);
    hh++;
    *hh = *cc++ * *hh - (*ss++ * tt);
  }

  /*
    compute the new plane rotation, and apply it to:
     1) the right-hand-side of the Hessenberg system
     2) the new column of the Hessenberg matrix
     thus obtaining the updated value of the residual
  */
  if( !hapend ) {
    //tt = PetscSqrtScalar(PetscConj(*hh) * *hh + PetscConj(*(hh+1)) * *(hh+1)); -- complex
    tt = std::sqrt( std::conj(*hh) * *hh + std::conj( *(hh+1) ) * *(hh+1) );
    if (tt == 0.0) {SETERRQ(PETSC_COMM_SELF,-1,"Your matrix or preconditioner is the null operator");}
    *cc       = *hh / tt;
    *ss       = *(hh+1) / tt;
    *cGRS(it+1) = - (*ss * *cGRS(it));
    *cGRS(it) = std::conj( *cc ) * *cGRS(it);
    *hh       = std::conj( *cc ) * *hh + *ss * *(hh+1);
    *res      = std::abs( *(cGRS(it+1)) );
  }
  else {
    /* happy breakdown: HH(it+1, it) = 0, therfore we don't need to apply 
       another rotation matrix (so RH doesn't change).  The new residual is 
       always the new sine term times the residual from last time (GRS(it)), 
       but now the new sine rotation would be zero...so the residual should
       be zero...so we will multiply "zero" by the last residual.  This might
       not be exactly what we want to do here -could just return "zero". */
    
    *res = 0.0;
  }
  //if(origin_.c == NULL)SETERRQ(PETSC_COMM_SELF,-1,"origin_.c == NULL");
  //assert(origin_.c[(nEig_+2)*(nEig_+1)+(nEig_+1)*(nEig_+1)+(nEig_+2)+2*(nEig_+1)].real()==-13.0);
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "ModifiedGramSchmidtOrthogonalization_complex"
/*
    This is the basic orthogonalization routine using modified Gram-Schmidt.
*/
int PromSolver::ModifiedGramSchmidtOrthogonalization_complex( const int it )
{
  
  int       ierr,j;
  PromComplex *hh,*hes,tmp;
  PromComplexVector *vv_1; PromVector_base *vv_j;

  PetscFunctionBegin;
  vv_1 = dynamic_cast<PromComplexVector*>(VEC_VV(it+1)); assert(vv_1!=NULL);
  /* update Hessenberg matrix and do Gram-Schmidt */
  hh  = cHH(0,it);
  hes = cHES(0,it);
  for (j=0; j<=it; j++) {
    /* (vv(it+1), vv(j)) */
    vv_j = VEC_VV(j); assert(vv_j!=NULL); 
    ierr = vv_1->cDot( vv_j, *hh );    CHKERRQ(ierr);
    *hes++ = *hh;
    /* vv(it+1) <- vv(it+1) - hh[it+1][j] vv(j) */
    tmp    = - (*hh++);  
    ierr   = vv_1->cAXPY( tmp, vv_j );CHKERRQ(ierr);
  } 
  PetscFunctionReturn(0);
}

/*
  This version uses UNMODIFIED Gram-Schmidt.  It is NOT always recommended, 
  but it can give MUCH better performance than the default modified form
  when running in a parallel environment.
 */
#undef __FUNCT__  
#define __FUNCT__ "UnmodifiedGramSchmidtOrthogonalization_complex"
int PromSolver::UnmodifiedGramSchmidtOrthogonalization_complex( const int it )
{
  int       j,ierr;
  PromComplex    *hh,*hes;
  PromComplexVector *vv_1;
  const PromVector_base *const*vv_0 = &(VEC_VV(0));
  
  PetscFunctionBegin;
  vv_1 = dynamic_cast<PromComplexVector*>(VEC_VV(it+1));assert(vv_1!=NULL);
  /* update Hessenberg matrix and do unmodified Gram-Schmidt */
  hh  = cHH(0,it);
  hes = cHES(0,it);
  /* 
     This is really a matrix-vector product, with the matrix stored
     as pointer to rows 
     */
  ierr = vv_1->cMDot( it+1, vv_0, hes );CHKERRQ(ierr);

  /*
    This is really a matrix-vector product: 
    [h[0],h[1],...]*[ v[0]; v[1]; ...] subtracted from v[it+1].
  */
  
  for (j=0; j<=it; j++) hh[j] = -hes[j]; 
  ierr = vv_1->cMAXPY( it+1, hh, vv_0 );CHKERRQ(ierr);
  for (j=0; j<=it; j++) hh[j] = -hh[j];
  PetscFunctionReturn(0);
}

/*
  uses 1 iteration of iterative refinement of UNMODIFIED Gram-Schmidt.  
  It can give better performance when running in a parallel 
  environment and in some cases even in a sequential environment (because
  MAXPY has more data reuse).

  Care is taken to accumulate the updated HH/HES values.
 */
#undef __FUNCT__  
#define __FUNCT__ "TwoUnmodifiedGramSchmidtOrthogonalization_complex"
int PromSolver::TwoUnmodifiedGramSchmidtOrthogonalization_complex( const int it )
{
  int       j,ierr,ncnt;
  PromComplex    *hh,*hes,fbuf[300],*lhh = NULL,tmp;
  PromComplexVector *vv_1 = dynamic_cast<PromComplexVector*>(VEC_VV(it+1));assert(vv_1!=NULL);
  const PromVector_base *const*vv_0 = &(VEC_VV(0)); 

  /* Don't allocate small arrays */
  if (it+1 < 300) lhh = fbuf;
  else lhh = new PromComplex[it+1];
  
  /* update Hessenberg matrix and do unmodified Gram-Schmidt */
  hh  = cHH(0,it);
  hes = cHES(0,it);
  
  /* Clear hh and hes since we will accumulate values into them */
  for (j=0; j<=it; j++) {
    hh[j]  = 0.0;
    hes[j] = 0.0;
  }
  
  for( ncnt = 0 ; ncnt < 2 ; ncnt++ ) {
    /* 
       This is really a matrix-vector product, with the matrix stored
       as pointer to rows 
    */
    perf_mon_.EventBegin(PromContext::prom_logs_[LOG_AUX1]);
    ierr = vv_1->cMDot( it+1, vv_0, lhh ); CHKERRQ(ierr); /* <v,vnew> */
    perf_mon_.EventEnd(PromContext::prom_logs_[LOG_AUX1]);
    /*
      This is really a matrix vector product: 
      [h[0],h[1],...]*[ v[0]; v[1]; ...] subtracted from v[it+1].
    */
    for (j=0; j<=it; j++) lhh[j] = - lhh[j];
    ierr = vv_1->cMAXPY( it+1, lhh, vv_0 );           CHKERRQ(ierr);
    for (j=0; j<=it; j++) {
      hh[j]  -= lhh[j];     /* hh += <v,vnew> */
      hes[j] += lhh[j];     /* hes += - <v,vnew> */
    }

    /*double hnrm = 0.0,wnrm;
      for (j=0; j<=it; j++) {
      tmp = lhh[j] * std::conj(lhh[j]);
      hnrm += tmp.real(); //PetscRealPart(lhh[j] * PetscConj(lhh[j]));
      }
      hnrm = sqrt(hnrm);
      ierr = vv_1->Norm2( &wnrm ); CHKERRQ(ierr);
      if (wnrm > 1.0286 * hnrm) break;*/
  }

  if( lhh != fbuf ) delete [] lhh;

  return 0;
}

#endif

//  PromPC stuff
#if defined(PROM_USE_PETSC)

#undef __FUNCT__
#define __FUNCT__ "CreateASM"
/* PromPC::CreateASM************************************
 *
 */
int PromPC::CreateASM( const int nblocks, PromIS *blockISs ) 
{
  assert(this!=NULL);
  if( getType() != PROMPCASM ) SETERRQ(PETSC_COMM_SELF,-1,(char*)getPCString());
  if( nblocks < 0 ) SETERRQ1(PETSC_COMM_SELF,-1,"nblocks(%d) < 0",nblocks);
  KSP *subksp; assert(pc_ != NULL);
  int  nlocal, first, ierr; 
  PromIS *bIS = (blockISs==NULL) ? PETSC_NULL : blockISs;
    
#if defined(PROM_TRACE_PC_SETUP)
  PetscPrintf(pc_->comm,"\t[0]PromPC::CreateASM (pc setup) this=%p\n",this);
#endif
  PetscFunctionBegin;
  ierr = PCASMSetOverlap( pc_, 0 );                  CHKERRQ(ierr);
  // PC_ASM_BASIC, PC_ASM_INTERPOLATE, PC_ASM_RESTRICT, PC_ASM_NONE 
  ierr = PCASMSetType( pc_, PC_ASM_RESTRICT ); CHKERRQ(ierr);
  ierr = PCASMSetLocalSubdomains( pc_, nblocks, bIS, PETSC_NULL );  CHKERRQ(ierr);
  ierr = PCSetUp( pc_ ); CHKERRQ(ierr);
 
  // set blocks up
  ierr = PCASMGetSubKSP( pc_, &nlocal, &first, &subksp);CHKERRQ(ierr);
  assert(nlocal==nblocks);
  while( nlocal-- ) {
    PC pc;
    ierr = KSPGetPC( subksp[nlocal], &pc );   	 CHKERRQ(ierr);  
    if( blockISs != NULL ){
      ierr = PCSetType( pc, PCLU ); CHKERRQ(ierr);
      ierr = PCFactorSetMatOrderingType( pc, MATORDERINGND );	 CHKERRQ(ierr);
      int sz;
      ierr = ISGetSize( blockISs[nlocal], &sz );         CHKERRQ(ierr);
      double nnz = 80.0; // guess!!!
#if defined(PARCH_ascired)
      double fill = (sz > 0) ? (nnz + 8.0*sqrt((double)sz))/nnz : 1.0;
#else
      double fill = (sz > 0) ? (nnz + 5.0*pow((double)sz,0.65))/nnz : 1.0;
#endif
      if( fill < 1.0 ) fill = 1.0;
      ierr = PCFactorSetFill( pc, fill ); 	         CHKERRQ(ierr);
    }
    else {
      ierr = PCSetType( pc, PCILU ); CHKERRQ(ierr);
      ierr = PCFactorSetMatOrderingType( pc_ , MATORDERINGQMD ); 
      CHKERRQ(ierr);
      ierr = PCFactorSetLevels( pc_ , 1 ); CHKERRQ(ierr);
    }
  }
  ierr = PCSetUpOnBlocks( pc_ );                CHKERRQ(ierr);

  // need to hack setup here as is done in create of ASM!
  if( kkt_ != NULL ) SETERRQ(PETSC_COMM_SELF,1,"should not kkt pc here");
  setup_called_ = 2;

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "Apply_private"
/* PromPC::Apply_private ************************************
 *
 */
int PromPC::Apply_private( const PromVector_base * const bbB, 
			   PromVector_base * const xxB,
			   const bool zeroguess /*=TRUE*/ )
{ 
  assert(this!=NULL); assert(pc_ != NULL); assert(zeroguess);
  int ierr; 
  // doit
  PetscFunctionBegin;
  const PromVector *const bb=dynamic_cast<const PromVector*>(bbB);
  assert(bb!=NULL);
  PromVector *const xx=dynamic_cast<PromVector*>(xxB);assert(xx!=NULL);
  ierr = PCApply( pc_, bb->vec_, xx->vec_ ); CHKERRQ(ierr); 
  if( numits_ > 1 ) {
    Mat A = pc_->mat; Vec X = xx->vec_, B = bb->vec_, R, T;
    if( work_ == NULL ) {
      ierr = VecDuplicate( X, &T );     CHKERRQ(ierr); work_ = T;
      ierr = VecDuplicate( X, &R );     CHKERRQ(ierr); rwork_ = R;
    }
    else { T = work_; R = rwork_; }
    //
    for( int it = 1 ; it < numits_ ; it++ ) {
      ierr = MatMult( A, X, R );    CHKERRQ(ierr); /* r <- b - A x'     */ 
      ierr = VecAYPX( R, -1., B );  CHKERRQ(ierr);
      ierr = PCApply( pc_, R, T );  CHKERRQ(ierr);/* t <- B * r */
      ierr = VecAXPY( X, 1.0, T );   CHKERRQ(ierr); /* x' <- x' + w * t  */ 
    }
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SetType"
/* PromPC::SetType************************************
 *
 */
int PromPC::SetType( PromPCType type, const PromVector_base * const vB ) 
{
  assert(this!=NULL); 
  if(pc_==NULL)SETERRQ(PETSC_COMM_SELF,-1,"pc_==NULL");
  int ierr;
  switch( type ) {
  case PROMPCILU : 
    { 
      ierr = PCSetType( pc_, PCILU  ); CHKERRQ(ierr);
      bool flag; int ii; double dt;
      ierr = options_.GetOptionInt("-pc_ilu_levels",&ii,&flag); CHKERRQ(ierr);
      if( flag ) {
	ierr = PCFactorSetLevels( pc_ ,ii ); CHKERRQ(ierr);
      }
      else {
	ierr=options_.GetOptionDouble("-pc_ilu_use_drop_tolerance",&dt,&flag);
	CHKERRQ(ierr);
	if( flag ) {
	  // guess for pivot tol. and max-row-count	  
	  //ierr = PCFactorSetUseDropTolerance(pc_,(PetscReal)dt,(PetscReal).1,200);
	  //CHKERRQ(ierr);
	}
      } 
    }
    break; 
  case PROMPCJACOBI : ierr = PCSetType(pc_, PCJACOBI); CHKERRQ(ierr);  break;
  case PROMPCASM :  ierr = PCSetType(pc_, PCASM); CHKERRQ(ierr);       break;
  case PROMPCLU : 
    ierr = PCSetType( pc_, PCLU  ); CHKERRQ(ierr);
    ierr = PCFactorSetFill( pc_, 7.0 );  CHKERRQ(ierr);
    ierr = PCFactorSetMatOrderingType( pc_, MATORDERINGND ); CHKERRQ(ierr);
    break;
  case PROMPCNONE : ierr = PCSetType( pc_, PCNONE  ); break;
  default: SETERRQ1(PETSC_COMM_SELF,1,"ERROR: bad pc type=%d\n",type);
  }
  assert( vB != NULL );
  const PromVector *vec = dynamic_cast<const PromVector*>(vB);
  assert(vec!=NULL && vec->vec_!=NULL);
  //ierr = PCSetVector( pc_, vec->vec_ ); CHKERRQ(ierr);

  return 0;
}

#else // petra PromPC 

#endif
